(* Heaps and heapsort.
   Copyright (C) 1992-1995 Jean Goubault and Bull S.A.
   All rights reserved.

   Redistribution and use in source and binary forms are permitted provided
   that: (1) source distributions retain this entire copyright notice and
   comment, and (2) distributions including binaries display the following
   acknowledgement:  ``This product includes software developed by Jean
   Goubault as a contributor of Bull S.A.'' in the
   documentation or other materials provided with the distribution and in
   all advertising materials mentioning features or use of this software.
   Neither the name of Bull S.A. or of its corporate affiliates,
   nor the names of its contributors may be used to endorse or promote
   products derived from this software without specific prior written permission.
   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*)

datatype 'a heap = EmptyHeap | BranchHeap of 'a * 'a heap * 'a heap

exception HeapEmpty

(* The invariant on heaps is as follows:
 either it is EmptyHeap (empty), or its root node contains the
 least element of the heap, and the two successors are again heaps. *)

(*local*)
    infix 4 less
(*in*)
    fun heap (op less) =
	let fun up (w, EmptyHeap) = BranchHeap (w, EmptyHeap, EmptyHeap)
	      | up (w, BranchHeap (v, p, q)) =
		if w less v
		    then BranchHeap (w, up (v, q), p)
		else BranchHeap (v, up (w, q), p)
	    fun removeleft (BranchHeap (w,EmptyHeap,_)) = (w,EmptyHeap)
	      | removeleft (BranchHeap (w,p,q)) =
		let val (x,p') = removeleft p
		in
		    (x, BranchHeap (w,q,p'))
		end
	    fun siftdown (x,EmptyHeap,_) = BranchHeap (x,EmptyHeap,EmptyHeap)
	      | siftdown (x, p as BranchHeap (y,_,_), EmptyHeap) =
		if x less y
		    then BranchHeap (x,p,EmptyHeap)
		else BranchHeap (y, BranchHeap (x,EmptyHeap,EmptyHeap), EmptyHeap)
	      | siftdown (x, p as BranchHeap (y,p1,q1), q as BranchHeap (z,p2,q2)) =
		if x less y andalso x less z
		    then BranchHeap (x,p,q)
		else if y less z
			 then BranchHeap (y, siftdown (x,p1,q1), q)
		     else BranchHeap (z, p, siftdown (x,p2,q2))
	    fun deletemin EmptyHeap = raise HeapEmpty
	      | deletemin (BranchHeap (w,EmptyHeap,q)) = q
	      | deletemin (BranchHeap (w,p,q)) =
		let val (x,p') = removeleft p
		in
		    siftdown (x,q,p')
		end
	    fun mklist (EmptyHeap, acc) = acc
	      | mklist (p as BranchHeap (w,...), acc) =
		mklist (deletemin p, w::acc)
	in
	    |[insert = up, (* insert an element in a heap, return new heap *)
	      findmin = (fn BranchHeap (w,...) => w | _ => raise HeapEmpty),
	      (* return least element of heap *)
	      deletemin = deletemin, (* build heap without its least element *)
	      mkrevlist = fn p => mklist (p,nil)
	      (* build list of elements of heap, sorted in decreasing order *)
	      ]|
	end

    (* sort a set of elements in ascending order, return list: *)
    fun heapsort (op less) =
	let val greater = fn (x,y) => y less x
	    val |[insert,mkrevlist,...]| = heap greater
	    fun putaway ({},p) = p
	      | putaway ({x => _} U rest, p) =
		putaway (rest, insert (x,p))
	in
	    fn s => mkrevlist (putaway (s,EmptyHeap))
	end
(*end*)
