/* HimML primitives.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include "portable.h"
#include "mem.h"
#include "eval.h"
#include "expr.h"
#include "timing.h"
#include "debug.h"
#include "transfer.h"
#include "ex1proto.h"
#include "ex2proto.h"
#include "exeproto.h"
#include "prtproto.h"
#include "rndproto.h"
#include "errproto.h"
#include "stkproto.h"
#include "dirproto.h"
#include "memproto.h"
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/errno.h>
#include <unistd.h>
#else
#include <errno.h>
#endif
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "hakproto.h"
#include "rexec.h"
#ifdef HAS_PROTOTYPES
#include "sexec.h"
#endif

EXTERN ExprPtr _minf,_mnorm,_mden,_zero,_pden,_pnorm,_pinf;
EXTERN ExprPtr numclassttm;
EXTERN ExprPtr noninvException,nthException,
  asciiException,stringnthException,rangeException,
  subscriptException,ioException;
EXTERN ExprPtr writesym,readsym,flushsym,readlinesym,tellsym,seeksym,
  advancesym,seekendsym,suspendsym,truncatesym,convertsym,killsym;
EXTERN ExprPtr voidString,*charStrings;
EXTERN ExprPtr none,optttm;
PUBLIC MapPtr mapgetCache_e,mapgetCache_m,mapgetCache_r;
PUBLIC Boolean mapgetCache_f=false;
PUBLIC ExprPtr leftMarginRef,rightMarginRef,maxLinesRef;
EXTERN COUNTER crtab_maxlines;
EXTERN ExprPtr theArgs;

/* All primitives are declared by DEF_ACTION.
   Note that primitives should not protect their result, as
   it will be protected anyway by Eval() or CallCC() (actually, by Throw()).
   Moreover, the translator to C---and the evaluator as well---assumes
   that primitives never protect anything, so that they do not have to QUIT periodically
   to avoid the explosion of the KEEP tables. */

#define __SEG__ ACTIONS

PRIVATE ExprPtr nancode1(x)
IN CONST REAL *x;
{
  ExprPtr nancode;

#if FLOAT_FORMAT==IEEE754
  {
    INT i;

    i = NaNcodeD(x);
    if (i<0)
      i = -i;
    nancode = Kint(i);
  }
#else
  nancode = izero;
#endif
  return nancode;
}

PRIVATE ExprPtr classify1(x)
IN CONST REAL *x;
{
  ExprPtr e;

  switch (classifyD(x))
    {
    case -C_SNAN:
      e = Kdata(0,numclassttm,nancode1(x));
      (void) KEEP(e);
      return e;
    case -C_INFINITE:
      return _minf;
    case -C_NORMALNUM:
      return _mnorm;
    case -C_DENORMALNUM:
      return _mden;
    case -C_ZERONUM:
    case C_ZERONUM:
      return _zero;
    case C_DENORMALNUM:
      return _pden;
    case C_NORMALNUM:
      return _pnorm;
    case C_INFINITE:
      return _pinf;
    case C_SNAN:
      e = Kdata(1,numclassttm,nancode1(x));
      (void) KEEP(e);
      return e;
    default:
      e = Kdata(1,numclassttm,izero);
      (void) KEEP(e);
      return e;
    }
}

DEF_ACTION_1ARG(classify)
{
  return (arg->kind==(CHAR)E_REAL)?Ktuple2(classify1(&XVAL(arg)),_zero):
    Ktuple2(classify1(&XVAL(REPART(arg))),classify1(&XVAL(IMPART(arg))));
}
ACTION_END

DEF_ACTION_1ARG(not)
{
  return (arg==falseValue)?trueValue:falseValue;
}
ACTION_END

#if PLDI93_HACK
PRIVATE Boolean equalVec(vp1,vp2)
IN CONST REGISTER VecPtr vp1,vp2;
{
  REGISTER COUNTER n;
  REGISTER ExprPtr *ep1,*ep2;

  if ((n=vp1->n) != vp2->n)
    return false;
  for (ep1=vp1->expr, ep2=vp2->expr; n!=0; n--)
    if (!equal(*ep1++,*ep2++))
      return false;
  return true;
}

/*!!! note: if we want to use this hack again, we should
  first test for STKFULL(), and if true, call a beamer function
  with CallCC(), to avoid overflowing the current thread cache. */

PUBLIC Boolean equal(x,y)
IN CONST REGISTER ExprPtr x,y;
{
again:
  if (x==y)
    return true;
  if (x->kind!=y->kind)
    return false;
  switch (x->kind)
    {
    case E_NIL: case E_EMPTY: case E_TUPLE_0: case E_RECORD_0:
      return true;
    case E_INT:
      return IVAL(x)==IVAL(y);
    case E_REAL:
      return XVAL(x)==XVAL(y);
    case E_COMPLEX: case E_CONS: case E_MAPLET: case E_MAP:
    case E_TUPLE_2: case E_RECORD_1:
      if (equal(CAR(x),CAR(y)))
	{
	  x = CDR(x);
	  y = CDR(y);
	  goto again;
	}
      return false;
    case E_DATA:
      if (x->magic==y->magic)
	{
	  x = CDR(x);
	  y = CDR(y);
	  goto again;
	}
      return false;
    case E_STRING:
      return !StrCmp(SLEN(x),STEXT(x),SLEN(y),STEXT(y));
    case E_TUPLE_1:
      x = REFVAL(x);
      y = REFVAL(y);
      goto again;
    case E_RECORD_GNRL:
      if (ASHAPE(x)!=ASHAPE(y))
	return false;
      /* fallthrough */
    case E_TUPLE_GNRL:
      return equalVec(AVAL(x),AVAL(y));
    case E_REF:
    case E_ARRAY_2:
    case E_ARRAY_GNRL: case E_EXCEPTION:
      return false;
    case E_EXTERNAL:
      if (EXTDESC(x)!=EXTDESC(y))
	return false;
      return (*(EXTDESC(x)->eq))(EXTDATA(x),EXTDATA(y));
    default:
      kinderror(message(72),x->kind);
    }
}

DEF_ACTION_2ARGS(equal)
{
  return equal(arg1,arg2)?trueValue:falseValue;
}
ACTION_END

DEF_ACTION_2ARGS(nequal)
{
  return equal(arg1,arg2)?trueValue:falseValue;
}
ACTION_END
#else
DEF_ACTION_2ARGS(equal)
{
  return (arg1==arg2)?trueValue:falseValue;
}
ACTION_END

DEF_ACTION_2ARGS(nequal)
{
  return (arg1!=arg2)?trueValue:falseValue;
}
ACTION_END
#endif

DEF_PRIVATE_ACTION(q_compose)
{ /* ENV_UNDERLYING(env) is the pair (f,g) of functions to compose. */
  va_list _args;
  ExprPtr f,g, uenv, arg;
  ENTER;

  uenv = ENV_UNDERLYING(env);
  f = CAR(uenv);
  g = CDR(uenv);
  M_VARCALL(arg,(*ACTION(g)),QENV(g),ENV_ARITY(env));
  /* call g with varying number of arguments. */
  (void) KEEP(arg);
  arg = (*ACTION(f))(ENV_CONS(QENV(f),1),arg); /* call f as a unary function (always) */
  QUIT;
  return arg;
}

DEF_ACTION_COLLECT_WEAK(compose)
{ /* arg will get the tuple2(f,g) of the two arguments f, g. */
  res = MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_compose),arg);
  /* need to build result in res, which will then be returned
     by ACTION_END_COLLECT. */
}
ACTION_END_COLLECT

DEF_ACTION_1ARG(inv)
{
  return MapInverse0(arg);
}
ACTION_END

DEF_ACTION_2ARGS(mapcompose)
{
  ExprPtr res;
  ENTER;

  res = MapCompose(arg1,arg2);
  QUIT;
  return res;
}
ACTION_END

DEF_ACTION_1ARG(rev)
{
  return Reverse(arg);
}
ACTION_END

DEF_ACTION_2ARGS(rev1)
{
  return Reverse1(arg1,arg2);
}
ACTION_END

DEF_ACTION_2ARGS(append)
{
  return Append0(arg1,arg2);
}
ACTION_END

DEF_ACTION_2ARGS(concat)
{
  ExprPtr str;
  INT len1,len2,len;

  len1 = SLEN(arg1);
  len2 = SLEN(arg2);
  len = len1+len2;
  {
#if STOP_AND_COPY_VECTORS
    CHAR *text;

    text = (CHAR *)alloctmp(len);
#else
    CHAR *text;

    VECALLOC(text,len);
#endif
    safememcpy(text,(char *)STEXT(arg1),len1);
    safememcpy(text+len1,(char *)STEXT(arg2),len2);
#if STOP_AND_COPY_VECTORS
    str = Kstring1(len,text,true);
#else
    str = Kstring1(len,text,false);
#endif
  }
  return str;
}
ACTION_END

DEF_ACTION_1ARG(concatof)
{
  ExprPtr ll,str;
  INT len,len1;
  CHAR *s;

  for (ll=arg, len=0; ll!=NIL; ll=CDR(ll))
    len += SLEN(CAR(ll));
  {
#if STOP_AND_COPY_VECTORS
    CHAR *text;

    text = (CHAR *)alloctmp(len);
#else
    CHAR *text;

    VECALLOC(text,len);
#endif
    for (ll=arg, s=text; ll!=NIL; ll=CDR(ll))
      {
	len1 = SLEN(CAR(ll));
	safememcpy(s,(char *)STEXT(CAR(ll)),len1);
	s += len1;
      }
#if STOP_AND_COPY_VECTORS
    str = Kstring1(len,text,true);
#else
    str = Kstring1(len,text,false);
#endif
  }
  return str;
}
ACTION_END

DEF_ACTION_2ARGS(domrestrto)
{
  return MapDRT(arg2,arg1);
}
ACTION_END

DEF_ACTION_2ARGS(domrestrby)
{
  return MapDRB(arg2,arg1);
}
ACTION_END

DEF_ACTION_2ARGS(rngrestrto)
{
  return MapRRT(arg1,arg2);
}
ACTION_END

DEF_ACTION_2ARGS(rngrestrby)
{
  return MapRRB(arg1,arg2);
}
ACTION_END

DEF_ACTION_2ARGS(union)
{
  return MapOverwrite(arg1,arg2);
}
ACTION_END

DEF_ACTION_2ARGS(inter)
{
  return MapDRT(arg1,arg2);
}
ACTION_END

DEF_ACTION_2ARGS(diff)
{
  return MapDRB(arg1,arg2);
}
ACTION_END

DEF_ACTION_2ARGS(delta)
{
  return MapD(arg1,arg2);
}
ACTION_END

DEF_ACTION_2ARGS(submap)
{
  return MapIncludedp(arg1,arg2)?trueValue:falseValue;
}
ACTION_END

DEF_ACTION_2ARGS(subset)
{
  return SetIncludedp(arg1,arg2)?trueValue:falseValue;
}
ACTION_END

DEF_ACTION_2ARGS(inset)
{
  ExprPtr e,m,aux;
  Boolean found;

  e = arg1;
  m = arg2;
  if (e==mapgetCache_e && m==mapgetCache_m)
    found = mapgetCache_f;
  else
    {
      SETQ(mapgetCache_e,e);
      SETQ(mapgetCache_m,m);
      aux = MapGet(e,m,&found);
      SETQ(mapgetCache_r,aux);
      mapgetCache_f = found;
    }
  return found?trueValue:falseValue;
}
ACTION_END

DEF_ACTION_2ARGS(inmap)
{
  ExprPtr e,m,x,y,aux;
  Boolean found;

  e = arg1;
  x = CAR(e);
  y = CDR(e);
  m = arg2;
  if (x==mapgetCache_e && m==mapgetCache_m)
    found = mapgetCache_f && CDR(e)==mapgetCache_r;
  else
    {
      SETQ(mapgetCache_e,x);
      SETQ(mapgetCache_m,m);
      aux = MapGet(x,m,&found);
      SETQ(mapgetCache_r,aux);
      mapgetCache_f = found;
      if (y!=mapgetCache_r)
	found = false;
    }
  return found?trueValue:falseValue;
}
ACTION_END

DEF_ACTION_1ARG(card)
{
  INT x;

  x = MapCard(arg);
  return Kint(x);
}
ACTION_END

DEF_ACTION_1ARG(len)
{
  INT x;

  x = ListLength(arg);
  return Kint(x);
}
ACTION_END

DEF_ACTION_2ARGS(interp)
{
  return MapInterp(arg1,arg2)?trueValue:falseValue;
}
ACTION_END

PUBLIC ExprPtr nth(ll,x)
IN CONST ExprPtr ll,x;
{
  REGISTER INT n;

  n = IVAL(x);
  if (n<0)
    return Raise(nthException);
  while (n--!=0)
    {
      if (ll==NIL)
	return Raise(nthException);
      ll = CDR(ll);
    }
  if (ll==NIL)
    return Raise(nthException);
  return CAR(ll);
}

DEF_ACTION_2ARGS(nth)
{
  return nth(arg1,arg2);
}
ACTION_END

PUBLIC ExprPtr nthtail(ll,x)
IN CONST ExprPtr ll,x;
{
  REGISTER INT n;

  n = IVAL(x);
  if (n<0)
    return Raise(nthException);
  while (n--!=0)
    {
      if (ll==NIL)
	return Raise(nthException);
      ll = CDR(ll);
    }
  return ll;
}

DEF_ACTION_2ARGS(nthtail)
{
  return nthtail(arg1,arg2);
}
ACTION_END

PRIVATE MapPtr range1(x,y)
IN CONST REGISTER INT x,y;
{
  ExprPtr res;
  INT z;
  ENTER;

  if (x>y)
    return NewMap();
  if (x==y)
    return Singleton(KINT(x));
  z = (x+y)>>1;
  res = MapMerge(range1(x,z),range1(z+1,y));
  KEEPONLY(res);
  return res;
}

PUBLIC ExprPtr range(x,y)
IN CONST ExprPtr x,y;
{
  MapPtr range;
  ENTER;

  range = range1(IVAL(x),IVAL(y));
  QUIT;
  return range;
}

DEF_ACTION_2ARGS(range)
{
  return range(arg1,arg2);
}
ACTION_END

DEF_ACTION_1ARG(size)
{
  return Kint(SLEN(arg));
}
ACTION_END

DEF_ACTION_1ARG(chr)
{
  CHAR c;
  REGISTER INT n;

  n = IVAL(arg);
  if (n<0 || n>255)
    return Raise(asciiException);
  c = n;
  arg = Kstring(1L,&c);
  return arg; /* there was no need to keep it anyway, it was already in a special table */
}
ACTION_END

DEF_ACTION_1ARG(ord)
{
  if (SLEN(arg)==0)
    return Raise(stringnthException);
  return Kint((INT)STEXT(arg)[0]);
}
ACTION_END

PUBLIC ExprPtr ordnth(str,i)
IN CONST ExprPtr str,i;
{
  REGISTER INT n;

  n = IVAL(i);
  if (n<0 || n>=SLEN(str))
    return Raise(stringnthException);
  return Kint((INT)STEXT(str)[n]);
}

DEF_ACTION_2ARGS(ordnth)
{
  return ordnth(arg1,arg2);
}
ACTION_END

DEF_ACTION_1ARG(explode)
{
  ExprPtr list,str;
  CHAR *s;
  INT len;

  list = NIL;
  len = SLEN(arg);
#if !STOP_AND_COPY_VECTORS
  s = STEXT(arg)+len;
#endif
  BEGINQUICKKEEP();
  while (len!=0)
    {
      len--;
#if STOP_AND_COPY_VECTORS
      s = STEXT(arg)+len;
#else
      s--;
#endif
      str = Kstring((INT)1,(CHAR *)s);
      /* this is correct, even in STOP_AND_COPY_VECTORS,
	 since 1-char strings are preallocated */
      list = Kons(KEEP(str),list);
      QUICKKEEP(list);
    }
  ENDQUICKKEEP();
  return list;
}
ACTION_END

DEF_ACTION_2ARGS(assign)
{
  SETREFVAL(arg1,arg2);
  return tuple0;
}
ACTION_END

DEF_ACTION_1ARG(deref)
{
  return REFVAL(arg);
}
ACTION_END

PUBLIC ExprPtr substr(str,i,j)
IN CONST ExprPtr str,i,j;
{
  REGISTER INT m,n;
  ExprPtr val;

  m = IVAL(i);
  if (m<0)
    return Raise(stringnthException);
  n = IVAL(j);
  if (n>SLEN(str))
    return Raise(stringnthException);
  if (n<m)
    n = m;
#if STOP_AND_COPY_VECTORS
  {
    REGISTER INT size = n-m;
    CHAR *s;

    s = (CHAR *)alloctmp(size);
    safememcpy(s,(char *)&STEXT(str)[m],size);
    val = Kstring(size,s);
  }
#else
  val = Kstring(n-m,&STEXT(str)[m]);
#endif
  return val;
}

DEF_ACTION_3ARGS(substr)
{
  return substr(arg1,arg2,arg3);
}
ACTION_END

DEF_ACTION_1ARG(chooseindom)
{
  if (arg==NewMap())
    return Raise(EMPTYEXC);
  return ChooseInDom(arg);
}
ACTION_END

DEF_ACTION_1ARG(chooseinrng)
{
  if (arg==NewMap())
    return Raise(EMPTYEXC);
  return ChooseInRng(arg);
}
ACTION_END

DEF_ACTION_2ARGS(mapremove)
{
  return MapKill(arg1,arg2);
}
ACTION_END

DEF_ACTION_1ARG(mapsplit)
{
  MapPtr map1,map2;
  ENTER;

  MapSplit(arg,&map1,&map2);
  arg = Ktuple2(map1,map2);
  QUIT;
  return arg;
}
ACTION_END

DEF_ACTION_2ARGS(mapadd)
{
  ExprPtr x,y;

  x = CAR(arg1);
  y = CDR(arg1);
  return MapAdd(x,y,arg2);
}
ACTION_END

DEF_ACTION_2ARGS(mapaddunder)
{
  ExprPtr x,y,arg2;

  x = CAR(arg1);
  y = CDR(arg1);
  return MapSlip(x,y,arg2);
}
ACTION_END

DEF_ACTION_2ARGS(cons)
{
  return Kons(arg1,arg2);
}
ACTION_END

DEF_ACTION_1ARG(ref)
{
  ExprPtr e;

  MAKEREF(arg,e);
  return e;
}
ACTION_END

PRIVATE ExprPtr appendrev(ll)
IN CONST REGISTER ExprPtr ll;
{
    REGISTER ExprPtr res;
    ENTER;

    res = CAR(ll);
    while (ll=CDR(ll), ll!=NIL)
      {
	res = Append0(CAR(ll),res);
	KEEPONLY(res);
      }
    QUIT;
    return res;
  }

DEF_ACTION_1ARG(appendof)
{
  if (arg==NIL)
    return NIL;
  {
    ExprPtr res;
    ENTER;

    arg = Nreverse(arg);
    res = appendrev(arg);
    QUIT;
    return res;
  }
}
ACTION_END

PRIVATE ExprPtr do_getx(x,y,data)
IN CONST ExprPtr x,y;
IN CONST CHAR *data;
{
  return x;
}

DEF_ACTION_1ARG(unionof)
{
  ExprPtr set;
  ENTER;

  set = MapCollect0(arg,do_getx,(char *)0L);
  QUIT;
  return set;
}
ACTION_END

DEF_ACTION_1ARG(mergeof)
{
  ExprPtr set;
  ENTER;

  set = NewMap();
  for (; arg!=NIL; arg=CDR(arg))
    {
      set = MapMerge(set,CAR(arg));
      KEEPONLY(set);
    }
  QUIT;
  return set;
}
ACTION_END

DEF_ACTION_1ARG(mergeunderof)
{
  ExprPtr set;
  ENTER;

  set = NewMap();
  for (; arg!=NIL; arg=CDR(arg))
    {
      set = MapMerge(CAR(arg),set);
      KEEPONLY(set);
    }
  QUIT;
  return set;
}
ACTION_END

PRIVATE MapPtr inter_op(x,y)
IN CONST MapPtr x,y;
{
  if (x==NIL)
    return y;
  if (y==NIL)
    return x;
  return MapDRT(x,y);
}

DEF_ACTION_1ARG(interof)
{
  ExprPtr set;
  ENTER;

  if (arg==NewMap())
    return Raise(EMPTYEXC);
  set = MapFold(arg,do_getx,(char *)0L,NIL,inter_op);
  QUIT;
  return set;
}
ACTION_END

DEF_ACTION_COLLECT(id)
{
  res = arg;
}
ACTION_END_COLLECT

DEF_ACTION_2ARGS(overwrite)
{
  return MapOverwrite(arg1,arg2);
}
ACTION_END

DEF_ACTION_1ARG(q_throw)
{ /* env is the continuation */
  Throw(ENV_UNDERLYING(env),arg);
  /*NOTREACHED*/
  return tuple0;
}
ACTION_END

DEF_ACTION_1ARG(throw)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_throw),arg);
}
ACTION_END

DEF_ACTION_COLLECT(q_mapapp)
{ /* abuse: ENV_UNDERLYING(env) is not an environment, but directly the map */
  ExprPtr y;
  ExprPtr uenv;
  Boolean found;

  uenv = ENV_UNDERLYING(env);
  if (arg==mapgetCache_e && uenv==mapgetCache_m)
    {
      y = mapgetCache_r;
      found = mapgetCache_f;
    }
  else
    {
      SETQ(mapgetCache_e,arg);
      SETQ(mapgetCache_m,uenv);
      y = MapGet(arg,uenv,&found);
      SETQ(mapgetCache_r,y);
      mapgetCache_f = found;
    }
  if (!found)
    y = Raise(MAPGETEXC);
  res = y;
}
ACTION_END_COLLECT

DEF_ACTION_1ARG(mapapp)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_mapapp),arg);
}
ACTION_END

DEF_ACTION_0ARG(gc)
{
  GCgrow(0, NULL);
  return tuple0;
}
ACTION_END

DEF_ACTION_1ARG(gcfull)
{
  GCgrow(10, arg);
  return tuple0;
}
ACTION_END

DEF_ACTION_2ARGS(array)
{
  REGISTER COUNTER m;

  m = IVAL(arg1);
  if (m<0)
    return Raise(subscriptException);
  return MakeArray(m,arg2);
}
ACTION_END

DEF_ACTION_2ARGS(aref)
{
  REGISTER COUNTER m;
  REGISTER VecPtr vp;

  m = IVAL(arg2);
  if (m<0)
    return Raise(subscriptException);
  vp = AVAL(arg1);
  if (m>=vp->n)
    return Raise(subscriptException);
  return vp->expr[m];
}
ACTION_END

DEF_ACTION_3ARGS(aset)
{
  REGISTER COUNTER m;
  VecPtr vp;

  m = IVAL(arg2);
  if (m<0)
    return Raise(subscriptException);
  vp = AVAL(arg1);
  if (m>=vp->n)
    return Raise(subscriptException);
  vp->expr[m] = arg3;
  HITBARRIER(arg1);
  return tuple0;
}
ACTION_END

DEF_ACTION_1ARG(length)
{
  return Kint(AVAL(arg)->n);
}
ACTION_END

DEF_ACTION_1ARG(listarray)
{
  return ListToArray0(arg);
}
ACTION_END

PRIVATE void iarraySweep (extdata, doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  switch (doext->xtag)
    {
    case X_SAVE:
      {
	INT len, i, *ip;

	len = ((INT *)extdata)[0];
	outInt (len);
	if (errno) return;
	for (i=0, ip = (INT *)extdata; i<len; i++)
	  {
	    outInt (*++ip);
	    if (errno) return;
	  }
	break;
      }
    case X_MARK:
    case X_MARKDATA:
      break;
    case X_FREE:
      MARKFREE (extdata);
      break;
    }
}

PRIVATE ExprPtr iarrayLoad (desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  INT len;
  INT *ip, *ipp, i;
  ExprPtr e;

  len = inInt ();
  if (errno) return NIL;
  ipp = ip = (INT *)MALLOC ((len+1) * sizeof(INT));
  ipp[0] = len;
  for (i=0; i<len; i++)
    {
      *++ipp = inInt ();
      if (errno)
	{
	  MARKFREE (ip);
	  return NIL;
	}
    }
  e = MakeExternal (desc, (char *)ip);
  (void) KEEP (e);
  return e;
}

PUBLIC extDesc iarrayExtDesc = {
  (CHAR *)"intarray",
  { 'i','a','r','r' }, NULL,
  NULL, /* no hash function, integer arrays are not shared */
  NULL, /* no comparison function, integer arrays are not shared */
  NULL, /* no print function */
  iarraySweep,
  iarrayLoad
};

DEF_ACTION_2ARGS(iarray)
{
  REGISTER COUNTER m;
  INT *ip, *ipp, i;
  INT ival;
  ExprPtr e;

  m = IVAL (arg1);
  if (m<0)
    return Raise(subscriptException);
  ival = IVAL (arg2);
  ipp = ip = (INT *)MALLOC((m+1) * sizeof (INT));
  ipp[0] = (INT)m;
  for (i=0; i<m; i++)
    *++ipp = ival;
  e = MakeExternal (&iarrayExtDesc, (char *)ip);
  return e;
}
ACTION_END

DEF_ACTION_2ARGS(iaref)
{
  REGISTER INT m;
  REGISTER INT *ip;

  m = IVAL(arg2);
  if (m<0)
    return Raise(subscriptException);
  ip = (INT *)EXTDATA (arg1);
  if (m>=ip[0])
    return Raise(subscriptException);
  ip++;
  return Kint (ip [m]);
}
ACTION_END

DEF_ACTION_3ARGS(iaset)
{
  REGISTER INT m, ival;
  REGISTER INT *ip;

  m = IVAL(arg2);
  ival = IVAL(arg3);
  if (m<0)
    return Raise(subscriptException);
  ip = (INT *)EXTDATA(arg1);
  if (m>=ip[0])
    return Raise(subscriptException);
  ip++;
  ip[m] = ival;
  return tuple0;
}
ACTION_END

DEF_ACTION_1ARG(ilength)
{
  return Kint (((INT *)EXTDATA(arg))[0]);
}
ACTION_END

DEF_ACTION_1ARG(ilistarray)
{
  REGISTER INT len;
  REGISTER INT *ip, *ipp;
  ExprPtr e;

  len = ListLength (arg);
  ipp = ip = (INT *)MALLOC ((len+1) * sizeof(INT));
  ip[0] = len;
  for (; arg!=NIL; arg=CDR(arg))
    {
      *++ipp = IVAL (CAR (arg));
    }
  e = MakeExternal (&iarrayExtDesc, (char *)ip);
  return e;
}
ACTION_END

PUBLIC int StrCmp(len1,text1,len2,text2)
IN REGISTER INT len1,len2;
IN CONST REGISTER CHAR *text1,*text2;
{
  REGISTER INT len;

  len = (len1<len2)?len1:len2;
  while (len--)
    {
      if (*text1<*text2)
	return -1;
      if (*text1>*text2)
	return 1;
      text1++;
      text2++;
    }
  if (len1<len2)
    return -1;
  if (len1>len2)
    return 1;
  return 0;
}

DEF_ACTION_2ARGS(strless)
{
  return (StrCmp(SLEN(arg1),STEXT(arg1),SLEN(arg2),STEXT(arg2))<0)?
    trueValue:falseValue;
}
ACTION_END

DEF_ACTION_2ARGS(before)
{
  return arg1;
}
ACTION_END

DEF_ACTION_1ARG(null)
{
  return (arg==NIL)?trueValue:falseValue;
}
ACTION_END

DEF_ACTION_1ARG(empty)
{
  return (arg==NewMap())?trueValue:falseValue;
}
ACTION_END

PUBLIC NORETURN ACTION_HEADER_0ARG(abort)
{
  Toplevel(izero);
  /*NOTREACHED*/
#ifndef HAS_NORETURN
  return NIL;
#endif
}
ACTION_END

PUBLIC NORETURN ACTION_HEADER_1ARG(quit)
{
  (void) ml_done((int)IVAL(arg));
  /*NOTREACHED*/
#ifndef HAS_NORETURN
  return tuple0;
#endif
}
ACTION_END

PUBLIC ExprPtr Select(label,arg)
IN CONST ExprPtr label,arg;
{
  switch (arg->kind)
    {
    case E_TUPLE_1:
      return REFVAL(arg);
    case E_RECORD_1:
      return CDR(arg);
    case E_TUPLE_2:
      return (label==ione)?CAR(arg):CDR(arg);
    case E_RECORD_GNRL:
      if (label->kind!=(CHAR)E_INT)
	label = MapGetSure(label,ASHAPE(arg));
      /* fallthrough */
    case E_TUPLE_GNRL:
      return AVAL(arg)->expr[IVAL(label)-1];
    default:
      kinderror(message(398),arg->kind);
      return NIL;
    }
}

PRINTFUNC(do_printaction)
{
  ExprPtr stream;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[1]; } _x;

  stream = PRINTDATA(ExprPtr);
  _x.n = 1;
  _x.next = theStackFrames;
  _x.p[0] = Kstring((INT)len,text);
  theStackFrames = (gcStackView *)&_x;
  (void) (*ACTION(stream))(ENV_CONS(QENV(stream),1), _x.p[0]);
  theStackFrames = _x.next;
}

DEF_PRIVATE_ACTION_1ARG(q_print)
{ /* ENV_UNDERLYING(env) is the stream on which to print, arg is the object to print */
  valPrint(PACKVAL(arg),PACKTYPE(arg),do_printaction,(char *)ENV_UNDERLYING(env),0);
  return tuple0;
}
ACTION_END

DEF_ACTION_1ARG(print)
{
  ExprPtr stream;

  stream = Select(writesym,arg);
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_print),stream);
}
ACTION_END

DEF_PRIVATE_ACTION_1ARG(q_pretty)
{ /* ENV_UNDERLYING(env) is the stream on which to print, arg is the object to print */
  COUNTER save_maxlines;

  save_maxlines = crtab_maxlines;
  crtab_maxlines = IVAL(REFVAL(maxLinesRef));
  if (crtab_maxlines>0)
    {
      if (!valPretty(PACKVAL(arg),PACKTYPE(arg),do_printaction,(char *)ENV_UNDERLYING(env),
		     0,IVAL(REFVAL(leftMarginRef)),
		     IVAL(REFVAL(rightMarginRef))))
        do_printaction((CHAR *)"\n",(INT)1,(char *)ENV_UNDERLYING(env));
      endPretty(do_printaction,(char *)ENV_UNDERLYING(env));
    }
  crtab_maxlines = save_maxlines;
  return tuple0;
}
ACTION_END

DEF_ACTION_1ARG(pretty)
{
  ExprPtr stream;

  stream = Select(writesym,arg);
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_pretty),stream);
}
ACTION_END

PRIVATE void printstring(text,len,stream)
REGISTER IN CONST CHAR *text;
REGISTER IN INT len;
REGISTER INOUT hFILE *stream;
{
  REGISTER CHAR c;

  while (len--!=0)
    {
       c = *text++;
       hputc(c,stream);
    }
}

DEF_ACTION_1ARG(prtstdout)
{
#if STOP_AND_COPY_VECTORS
  INT len;

  len = symNameLen(arg);
  {
    CHAR *s;

    s = (CHAR *)alloctmp(len);
    safememcpy(s,(char *)symName(arg),len);
    printstring(s,len,hstdout);
  }
#else
  printstring(symName(arg),symNameLen(arg),hstdout);
#endif
  return tuple0;
}
ACTION_END

DEF_ACTION_0ARG(flushstdout)
{
  hfflush(hstdout);
  return tuple0;
}
ACTION_END

DEF_ACTION_1ARG(prtstderr)
{
#if STOP_AND_COPY_VECTORS
  INT len;

  len = symNameLen(arg);
  {
    CHAR *s;

    s = (CHAR *)alloctmp(len);
    safememcpy(s,(char *)symName(arg),len);
    printstring(s,len,hstderr);
  }
#else
  printstring(symName(arg),symNameLen(arg),hstderr);
#endif
  return tuple0;
}
ACTION_END

DEF_ACTION_0ARG(flushstderr)
{
  hfflush(hstderr);
  return tuple0;
}
ACTION_END

PRIVATE void fileSweep(extdata,doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  hFILE *f;

  f = (hFILE *)extdata;
  switch (doext->xtag)
    {
    case X_SAVE:
      (void) hfflush(f);
      {
	INT len;
	CHAR c;

	len = (f->dirname==NULL)?0:strlen((char *)f->dirname);
	outInt(len);
	if (errno) return;
	outChars(f->dirname,len);
	if (errno) return;
	len = (f->filename==NULL)?0:strlen((char *)f->filename);
	outInt(len);
	if (errno) return;
	outChars(f->filename,len);
	if (errno) return;
	c = (CHAR)(f->fildesc==-1);
	outChars(&c,1);
	if (errno) return;
	c = (CHAR)(f->pid==NO_PROCESS);
	outChars(&c,1);
	if (errno) return;
	len = f->realpos;
	outInt(len);
	if (errno) return;
	len = f->flags;
	outInt(len);
      }
      break;
    case X_MARK:
    case X_MARKDATA:
      break;
    case X_FREE:
      hfclose(f); /* this is GC closing the file */
      break;
    }
}

PRIVATE ExprPtr fileLoad(desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  INT len,pos,flags;
  CHAR c,cp;
  CHAR *dirname,*olddir,*name;
  hFILE *hf;
  ExprPtr e;
  
  hf = (hFILE *)new_gc(sizeof(hFILE));
  dirname = name = NULL;
  len = inInt();
  if (errno) goto err;
  olddir = (CHAR *)alloctmp(PWD_BUFSIZE+1);
  dirname = (CHAR *)new_gc(PWD_BUFSIZE+1);
  if (dirname!=NULL)
    {
      dirname[len] = 0;
      inChars(dirname,len);
    }
  else inChars(olddir,len); /* to read up dirname into a dummy */
  if (errno) goto err;
  len = inInt();
  if (errno) goto err;
  name = (CHAR *)new_gc(len+1);
  if (name!=NULL)
    {
      inChars(name,len);
      name[len] = 0;
    }
  else inChars(olddir,len); /* to read up filename into a dummy */
  if (errno) goto err;
  inChars(&c,1);
  if (errno) goto err;
  inChars(&cp,1);
  if (errno) goto err;
  pos = inInt();
  if (errno) goto err;
  flags = inInt();
  if (errno) goto err;
  if (hf==NULL || dirname==NULL || name==NULL)
    {
err:
      if (hf!=NULL)
	DISPOSPTR(hf,sizeof(hFILE));
      if (dirname!=NULL)
	DISPOSPTR(dirname,PWD_BUFSIZE+1);
      if (name!=NULL)
	DISPOSPTR(name,len+1);
      hf = NULL;
    }
  else
    {
      (void) getDir(olddir,PWD_BUFSIZE);
      (void) changeDir(dirname);
      hf->next = NULL;
      hf->dirname = dirname;
      hf->filename = name;
      hf->ptr = hf->end = hf->buf;
      hf->fildesc = -1; /* at first, we create the file in suspended state */
      hf->pid = NO_PROCESS;
      hf->pos = hf->realpos = pos;
      hf->flags = flags;
      linkbuf(hf);
      if (c)
	(void) hfrevive(hf);
      (void) changeDir(olddir);
      errno = 0; /* ignore errors while reviving */
    }
  e = MakeExternal(desc,(char *)hf);
  (void) KEEP(e);
  return e;
}

PUBLIC extDesc outfileExtDesc = {
  (CHAR *)"file-as-outstream",
  { 'o','u','t','f' }, NULL,
  NULL, /* no hash function, files are not shared */
  NULL, /* no comparison function, files are not shared */
  NULL, /* no print function */
  fileSweep,
  fileLoad
};

DEF_PRIVATE_ACTION_1ARG(q_outfile)
{ /* ENV_UNDERLYING(env) is the external with the FILE * inside */
#if STOP_AND_COPY_VECTORS
  INT len;

  len = symNameLen(arg);
  {
    CHAR *s;

    s = (CHAR *)alloctmp(len);
    safememcpy(s,(char *)symName(arg),len);
    printstring(s,len,(hFILE *)EXTDATA(ENV_UNDERLYING(env)));
  }
#else
  printstring(symName(arg),symNameLen(arg),(hFILE *)EXTDATA(ENV_UNDERLYING(env)));
#endif
  return tuple0;
}
ACTION_END

DEF_PRIVATE_ACTION_0ARG(q_outfileflush)
{ /* ENV_UNDERLYING(env) is the external with the FILE * inside */
  if (hfflush((hFILE *)EXTDATA(ENV_UNDERLYING(env)))!=0)
    return raiseIO();
  return tuple0;
}
ACTION_END

DEF_PRIVATE_ACTION_0ARG(q_filetell)
{ /* ENV_UNDERLYING(env) is the external with the FILE * inside */
  INT n;

  n = hftell((hFILE *)EXTDATA(ENV_UNDERLYING(env)));
  return Kint(n);
}
ACTION_END

PRIVATE ExprPtr q_fileseek1(arg,env,whence)
IN CONST ExprPtr arg;
IN CONST ExprPtr env;
IN int whence;
{ /* abuse: env is not an environment, but directly the external with the FILE * inside */
  REGISTER INT n;

  n = IVAL(arg);
  (void) hfseek((hFILE *)EXTDATA(env),n,whence);
  return tuple0;
}

DEF_PRIVATE_ACTION_1ARG(q_fileseek)
{ /* ENV_UNDERLYING(env) is the external with the FILE * inside */
  return q_fileseek1(arg,ENV_UNDERLYING(env),0);
}
ACTION_END

DEF_PRIVATE_ACTION_1ARG(q_fileadv)
{ /* ENV_UNDERLYING(env) is the external with the FILE * inside */
  return q_fileseek1(arg,ENV_UNDERLYING(env),1);
}
ACTION_END

DEF_PRIVATE_ACTION_1ARG(q_fileseekend)
{ /* env is the external with the hFILE * inside */
  return q_fileseek1(arg,ENV_UNDERLYING(env),2);
}
ACTION_END

PUBLIC ExprPtr raiseIO()
{
  ExprPtr err,exc;

  err = KINT((INT)errno);
  MAKECONS(ioException,err,exc);
  return Raise(KEEP(exc));
}

DEF_PRIVATE_ACTION_0ARG(q_suspend)
{ /* ENV_UNDERLYING(env) is the external with the FILE * inside */
  if (hfsuspend((hFILE *)EXTDATA(ENV_UNDERLYING(env)))!=0)
    return raiseIO();
  return tuple0;
}
ACTION_END

DEF_PRIVATE_ACTION_0ARG(q_truncate)
{ /* ENV_UNDERLYING(env) is the external with the FILE * inside */
  hFILE *f;

  f = (hFILE *)EXTDATA(ENV_UNDERLYING(env));
  if (hftruncate(f,hftell(f))!=0)
    return raiseIO();
  return tuple0;
}
ACTION_END

PUBLIC hFILE *ffopen(arg,mode)
IN CONST SymPtr arg;
IN CONST char *mode;
{
  hFILE *file;
  INT len;

  len = symNameLen(arg);
  {
    char *name;

    name = (char *)alloctmp(len+1);
    safememcpy(name,(char *)symName(arg),len);
    name[len] = 0;
    file = hfopen(name,mode);
  }
  return file;
}

PUBLIC int ppopen(cmd,args,mode,hff)
     IN CONST SymPtr cmd;
IN CONST ExprPtr args;
IN CONST char *mode;
OUT hFILE **hff;
{
  INT len,argc;
  ExprPtr l;

  len = symNameLen(cmd)+1;
  argc = 1;
  for (l=args; l!=NIL; l=CDR(l), argc++)
    len += symNameLen(CAR(l))+1;
  {
    char *name,*s,**argv,**argvp;

    name = (char *)alloctmp_anchored(len,(char *)&len);
    argv = (char **)alloctmp_anchored((argc+1)*sizeof(char *),(char *)&len);
    s=name;
    argv[0] = s;
    len = symNameLen(cmd);
    safememcpy(s,(char *)symName(cmd),len);
    s += len;
    *s++ = 0;
    argvp = argv;
    for (l=args; l!=NIL; l=CDR(l))
      {
	*++argvp = s;
	len = symNameLen(CAR(l));
	safememcpy(s,(char *)symName(CAR(l)),len);
	s += len;
	*s++ = 0;
      }
    *++argvp = NULL;
    return hpopen(argv,mode,hff);
  }
}

PRIVATE ExprPtr MkCompiled (kind, action, env)
     IN ExprKind kind;
     IN compiledAction action;
     IN CONST ExprPtr env;
{
  ExprPtr e;

  e = MakeCompiled (kind, action, env);
  return KEEP (e);
}

PRIVATE ExprPtr outfileAction1(arg,mode)
IN CONST ExprPtr arg;
IN CONST char *mode;
{
  hFILE *file;
  ExprPtr ext;
#define N 8
  field fields[N];
  ENTER;

  file = ffopen(arg,mode);
  if (file==NULL)
    return raiseIO();
  ext = MakeExternal(&outfileExtDesc,(char *)file);
  (void) KEEP(ext);
  fields[0].label = writesym;
  fields[0].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_outfile),ext);
  fields[1].label = flushsym;
  fields[1].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_outfileflush),ext);
  fields[2].label = tellsym;
  fields[2].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_filetell),ext);
  fields[3].label = seeksym;
  fields[3].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_fileseek),ext);
  fields[4].label = advancesym;
  fields[4].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_fileadv),ext);
  fields[5].label = seekendsym;
  fields[5].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_fileseekend),ext);
  fields[6].label = suspendsym;
  fields[6].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_suspend),ext);
  fields[7].label = truncatesym;
  fields[7].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_truncate),ext);
  sortFieldArray(N,fields); /* we could sort it in advance */
  arg = MakeRecordGnrl(N,fields,fieldsToShape(N,fields));
  QUIT;
  return arg;
#undef N
}

DEF_ACTION_1ARG(outfile)
{
  return outfileAction1(arg,"w");
}
ACTION_END

DEF_ACTION_1ARG(appendfile)
{
  return outfileAction1(arg,"a");
}
ACTION_END

PUBLIC extDesc outprocessExtDesc = {
  (CHAR *)"process-as-outstream",
  { 'o','u','t','p' }, NULL,
  NULL, /* no hash function, files are not shared */
  NULL, /* no comparison function, files are not shared */
  NULL, /* no print function */
  fileSweep,
  fileLoad
};

DEF_ACTION_2ARGS(outprocess)
{
  hFILE *file[2];
  ExprPtr ext, res;
#define N 3
  field fields[N];
  ENTER;

  if (ppopen(arg1,arg2,"w",file))
    return raiseIO();
  ext = MakeExternal(&outprocessExtDesc,(char *)file[1]);
  (void) KEEP(ext);
  fields[0].label = writesym;
  fields[0].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_outfile),ext);
  fields[1].label = flushsym;
  fields[1].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_outfileflush),ext);
  fields[2].label = killsym;
  fields[2].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_suspend),ext);
  sortFieldArray(N,fields); /* we could sort it in advance */
  res = MakeRecordGnrl(N,fields,fieldsToShape(N,fields));
  QUIT;
  return res;
#undef N
}
ACTION_END

PRIVATE void outstringSweep(extdata,doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  struct outstring *f;

  f = (struct outstring *)extdata;
  switch (doext->xtag)
    {
    case X_SAVE:
      outInt(f->pos);
      if (errno) return;
      outInt(f->n);
      if (errno) return;
      {
	struct outchunk *chunk;

	for (chunk = f->chunks; chunk!=NULL; chunk = chunk->next)
	  {
	    outChars(chunk->c,chunk->end-chunk->start);
	  }
      }
      break;
    case X_MARK:
    case X_MARKDATA:
      break;
    case X_FREE:
      {
	struct outchunk *chunk,*next;

	for (chunk = f->chunks; chunk!=NULL; chunk = next)
	  {
	    next = chunk->next;
	    MARKFREE(chunk);
	  }
	MARKFREE(f);
      }
      break;
    }
}

PRIVATE ExprPtr outstringLoad(desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  struct outstring *outs;
  struct outchunk *chunk;
  INT pos,len;
  ExprPtr e;

  pos = inInt();
  if (errno) return NIL;
  len = inInt();
  if (errno) return NIL;
  if (len!=0)
    {
      chunk = (struct outchunk *)MALLOC(OUTCHSIZE(len));
      inChars(chunk->c,len);
      if (errno)
	{
	  MARKFREE(chunk);
	  return NIL;
	}
      chunk->next = NULL;
      chunk->start = 0;
      chunk->end = len;
    }
  else chunk = NULL;
  outs = (struct outstring *)MALLOC(sizeof(struct outstring));
  outs->chunks = outs->current = chunk;
  outs->pos = pos;
  outs->n = len;
  e = MakeExternal(desc,(char *)outs);
  (void) KEEP(e);
  return e;
}

PUBLIC extDesc outstringExtDesc = {
  (CHAR *)"string-as-outstream",
  { 'o','u','t','s' }, NULL,
  NULL, /* no hash function, files are not shared */
  NULL, /* no comparison function, files are not shared */
  NULL, /* no print function */
  outstringSweep,
  outstringLoad
};

#define OUTCHUNK_INC 256

DEF_PRIVATE_ACTION_1ARG(q_outstring)
{ /* ENV_UNDERLYING(env) is the external with the struct outstring * inside */
  INT i,len,pos,size,siz;
  struct outstring *f;
  struct outchunk *chunk,**ckp;

  len = SLEN(arg);
  if (len!=0)
    {
      f = (struct outstring *)EXTDATA(ENV_UNDERLYING(env));
      pos = f->pos;
      i = 0;
      ckp = &f->current;
      while ((chunk = *ckp)!=NULL)
	{
	  size = chunk->end-pos;
	  siz = len-i;
	  if (siz>size)
	    siz = size;
	  safememcpy(chunk->c+pos-chunk->start,STEXT(arg)+i,siz);
	  i += siz;
	  pos += siz;
	  if (pos<chunk->end)
	    break;
	  ckp = &chunk->next;
	}
      if (chunk==NULL && i!=len)
	{
	  siz = len-i;
	  size = siz;
	  if (size<OUTCHUNK_INC)
	    size = OUTCHUNK_INC;
	  chunk = (struct outchunk *)MALLOC(OUTCHSIZE(size));
	  chunk->next = NULL;
	  chunk->start = pos;
	  chunk->end = pos+size;
	  f->n = pos+siz;
	  safememcpy(chunk->c,STEXT(arg)+i,siz);
	  pos += siz;
	  *ckp = chunk;
	}
      if (chunk!=NULL)
	f->current = chunk;
      if (f->chunks==NULL)
	f->chunks = chunk;
      f->pos = pos;
      if (pos>f->n)
	f->n = pos;
    }
  return tuple0;
}
ACTION_END

DEF_PRIVATE_ACTION_0ARG(q_outstringtell)
{
  return Kint(((struct outstring *)EXTDATA(ENV_UNDERLYING(env)))->pos);
}
ACTION_END

PRIVATE void outstringseek(e,pos)
INOUT ExprPtr e;
IN INT pos;
{
  struct outstring *f;
  struct outchunk *chunk,*next;

  f = (struct outstring *)EXTDATA(e);
  if (pos<0)
    pos = 0;
  if (pos>f->n)
    pos = f->n;
  if (pos!=f->pos)
    {
      chunk = f->current;
      if (chunk==NULL || pos<chunk->start)
	chunk = f->chunks;
      while (chunk!=NULL && pos>=chunk->end && (next = chunk->next)!=NULL)
	chunk = next;
      f->pos = pos;
      f->current = chunk;
    }
}

DEF_PRIVATE_ACTION_1ARG(q_outstringseek)
{
  outstringseek(ENV_UNDERLYING(env),IVAL(arg));
  return tuple0;
}
ACTION_END

DEF_PRIVATE_ACTION_1ARG(q_outstringadv)
{
  outstringseek(ENV_UNDERLYING(env),IVAL(arg)+((struct outstring *)EXTDATA(ENV_UNDERLYING(env)))->pos);
  return tuple0;
}
ACTION_END

DEF_PRIVATE_ACTION_1ARG(q_outstringseekend)
{
  outstringseek(ENV_UNDERLYING(env),IVAL(arg)+((struct outstring *)EXTDATA(ENV_UNDERLYING(env)))->n);
  return tuple0;
}
ACTION_END

DEF_PRIVATE_ACTION_0ARG(q_outstringtruncate)
{
  struct outstring *f;
  struct outchunk *chunk,**ckp;

  f = (struct outstring *)EXTDATA(ENV_UNDERLYING(env));
  chunk = f->current;
  if (chunk!=NULL)
    {
      f->n = f->pos;
      ckp = &chunk->next;
      while (*ckp!=NULL)
	{
	  chunk = *ckp;
	  *ckp = chunk->next;
	  MARKFREE(chunk);
	}
    }
  return tuple0;
}
ACTION_END

DEF_PRIVATE_ACTION_0ARG(q_outstringconvert)
{
  struct outstring *f;
  struct outchunk *chunk;
  CHAR *s,*t;
  INT size;

  f = (struct outstring *)EXTDATA(ENV_UNDERLYING(env));
  CHECKHETEROFREE();
#if STOP_AND_COPY_VECTORS
  s = alloctmp (f->n);
  /*scAlloc(*(char **)&s,f->n);*/
#else
  VECALLOC(s,f->n);
#endif
  for (t=s, chunk=f->chunks; chunk!=NULL; chunk=chunk->next)
    {
      size = chunk->end;
      if (size>f->n)
	size = f->n;
      size -= chunk->start;
      safememcpy(t,chunk->c,size);
      t += size;
    }
#if STOP_AND_COPY_VECTORS
  return Kstring1(f->n,s,true);
#else
  return Kstring1(f->n,s,false);
#endif
}
ACTION_END

DEF_ACTION_1ARG(outstring)
{
  struct outstring *outs;
  ExprPtr ext;
#define N 7
  field fields[N];
  ENTER;

  outs = (struct outstring *)MALLOC(sizeof(struct outstring));
  outs->chunks = outs->current = NULL;
  outs->pos = outs->n = 0;
  ext = MakeExternal(&outstringExtDesc,(char *)outs);
  (void) KEEP(ext);
  ACTION_NAME(q_outstring)(ENV_CONS(ext,1),arg);
  fields[0].label = writesym;
  fields[0].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_outstring),ext);
  fields[1].label = tellsym;
  fields[1].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_outstringtell),ext);
  fields[2].label = seeksym;
  fields[2].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_outstringseek),ext);
  fields[3].label = advancesym;
  fields[3].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_outstringadv),ext);
  fields[4].label = seekendsym;
  fields[4].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_outstringseekend),ext);
  fields[5].label = truncatesym;
  fields[5].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_outstringtruncate),ext);
  fields[6].label = convertsym;
  fields[6].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_outstringconvert),ext);
  sortFieldArray(N,fields); /* we could sort it in advance */
  arg = MakeRecordGnrl(N,fields,fieldsToShape(N,fields));
  QUIT;
  return arg;
#undef N
}
ACTION_END

PRIVATE ExprPtr readn(file,n,stopc,finalc)
INOUT hFILE *file;
IN INT n;
IN int stopc;
OUT int *finalc;
{
  ExprPtr res;
  REGISTER int c;
  CHAR *s;
  INT i;

  *finalc = -2;
  if (n<=0)
    {
      return voidString;
    }
  if (n==1)
    {
again:
      c = hgetc(file);
      *finalc = c;
      if (c==-1)
	{
#ifdef EINTR
	  if (errno==EINTR)
	    goto again;
#endif
	  return voidString;
	}
      return charStrings[c];
    }
  {
    CHAR *buf;

    buf = (CHAR *)alloctmp(n);
    c = 0; /* useless, to prevent compiler from warning me that c may
	      not have been initialized when assigning it to *finalc below. */
    for (i=n, s=buf; i!=0; )
      {
	c = hgetcN(file,(stopc==-1)?n:1);
	if (c==-1)
	  {
#ifdef EINTR
	    if (errno==EINTR)
	      continue;
#endif
	    break;
	  }
	i--;
	*s++ = (CHAR)c;
	if (c==stopc)
	  break;
      }
    *finalc = c;
    res = Kstring1(n-i,buf,true);
  }
  return res;
}

PRIVATE ExprPtr readline(file)
INOUT hFILE *file;
{
  int finalc;
  ExprPtr str,list,aux,lastl;
  ENTER;
#define CHUNK_SIZE 256

  str = readn(file,(INT)CHUNK_SIZE,(int)'\n',&finalc);
  if (finalc==-1 || finalc=='\n')
    return str;
  (void) KEEP(str);
  MAKECONS(str,NIL,list);
  KEEPONLY(list);
  lastl = list;
  while (finalc!=-1 && finalc!='\n')
    {
      str = readn(file,(INT)CHUNK_SIZE,(int)'\n',&finalc);
      (void) KEEP(str);
      MAKECONS(str,NIL,aux);
      SETCDR(lastl,aux);
      lastl = aux;
    }
  str = ACTION_NAME(concatof)(ENV_CONS(NIL,1),list); /* NIL is a dummy env */
#undef CHUNK_SIZE
  KEEPONLY(str);
  return str;
}

PUBLIC ExprPtr inAction1(arg,file)
IN CONST ExprPtr arg;
INOUT hFILE *file;
{
  REGISTER INT n;
  int stopc;

  n = IVAL(arg);
  return readn(file,n,(int)-1,&stopc);
}

DEF_ACTION_1ARG(readstdin)
{
  return inAction1(arg,hstdin);
}
ACTION_END

DEF_ACTION_0ARG(lstdin)
{
  return readline(hstdin);
}
ACTION_END

PUBLIC extDesc infileExtDesc = {
  (CHAR *)"file-as-instream",
  { 'i','n','f',' ' }, NULL,
  NULL, /* no hash function, files are not shared */
  NULL, /* no comparison function, files are not shared */
  NULL, /* no print function */
  fileSweep,
  fileLoad
};

DEF_PRIVATE_ACTION_1ARG(q_infile)
{ /* ENV_UNDERLYING(env) is the external with the hFILE * inside */
  return inAction1(arg,(hFILE *)EXTDATA(ENV_UNDERLYING(env)));
}
ACTION_END

DEF_PRIVATE_ACTION_0ARG(q_linfile)
{ /* ENV_UNDERLYING(env) is the external with the hFILE * inside */
  return readline((hFILE *)EXTDATA(ENV_UNDERLYING(env)));
}
ACTION_END

DEF_ACTION_1ARG(infile)
{
  hFILE *file;
  ExprPtr ext;
#define N 7
  field fields[N];
  ENTER;

  file = ffopen(arg,"r");
  if (file==NULL)
    return raiseIO();
  ext = MakeExternal(&infileExtDesc,(char *)file);
  (void) KEEP(ext);
  fields[0].label = readsym;
  fields[0].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_infile),ext);
  fields[1].label = readlinesym;
  fields[1].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_linfile),ext);
  fields[2].label = tellsym;
  fields[2].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_filetell),ext);
  fields[3].label = seeksym;
  fields[3].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_fileseek),ext);
  fields[4].label = advancesym;
  fields[4].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_fileadv),ext);
  fields[5].label = seekendsym;
  fields[5].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_fileseekend),ext);
  fields[6].label = suspendsym;
  fields[6].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_suspend),ext);
  sortFieldArray(N,fields); /* we could sort it in advance */
  arg = MakeRecordGnrl(N,fields,fieldsToShape(N,fields));
  QUIT;
  return arg;
#undef N
}
ACTION_END

PUBLIC extDesc inprocessExtDesc = {
  (CHAR *)"process-as-instream",
  { 'i','n','p',' ' }, NULL,
  NULL, /* no hash function, files are not shared */
  NULL, /* no comparison function, files are not shared */
  NULL, /* no print function */
  fileSweep,
  fileLoad
};

DEF_ACTION_2ARGS(inprocess)
{
  hFILE *file[2];
  ExprPtr ext, res;
#define N 3
  field fields[N];
  ENTER;

  if (ppopen(arg1,arg2,"r",file))
    return raiseIO();
  ext = MakeExternal(&inprocessExtDesc,(char *)file[0]);
  (void) KEEP(ext);
  fields[0].label = readsym;
  fields[0].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_infile),ext);
  fields[1].label = readlinesym;
  fields[1].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_linfile),ext);
  fields[2].label = killsym;
  fields[2].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_suspend),ext);
  sortFieldArray(N,fields); /* we could sort it in advance */
  res = MakeRecordGnrl(N,fields,fieldsToShape(N,fields));
  QUIT;
  return res;
#undef N
}
ACTION_END

PUBLIC extDesc inoutprocessExtDesc = {
  (CHAR *)"process-as-in-and-outstream",
  { 'i','o','p',' ' }, NULL,
  NULL, /* no hash function, files are not shared */
  NULL, /* no comparison function, files are not shared */
  NULL, /* no print function */
  fileSweep,
  fileLoad
};

DEF_PRIVATE_ACTION_1ARG(q_infile2)
{ /* ENV_UNDERLYING(env) is a couple of externals, with hFILE *'s inside */
  return inAction1(arg,(hFILE *)EXTDATA(CAR(ENV_UNDERLYING(env))));
}
ACTION_END

DEF_PRIVATE_ACTION_0ARG(q_linfile2)
{ /* env is a couple of externals, with hFILE *'s inside */
  return readline((hFILE *)EXTDATA(CAR(ENV_UNDERLYING(env))));
}
ACTION_END

DEF_PRIVATE_ACTION_1ARG(q_outfile2)
{ /* ENV_UNDERLYING(env) is a pair of externals, with hFILE *'s inside */
  return ACTION_NAME(q_outfile)(ENV_CONS(CDR(ENV_UNDERLYING(env)),1), arg);
}
ACTION_END

DEF_PRIVATE_ACTION_0ARG(q_outfileflush2)
{ /* env is a couple of externals, with hFILE *'s inside */
  return ACTION_NAME(q_outfileflush)(ENV_CONS(CDR(ENV_UNDERLYING(env)),0));
}
ACTION_END

DEF_PRIVATE_ACTION_0ARG(q_suspend2)
{ /* env is a couple of externals, with hFILE *'s inside */
  return ACTION_NAME(q_suspend)(ENV_CONS(CDR(ENV_UNDERLYING(env)),0));
}
ACTION_END

DEF_ACTION_2ARGS(inoutprocess)
{
  hFILE *file[2];
  ExprPtr ext2[2],ext, res;
#define N 5
  field fields[N];
  ENTER;

  if (ppopen(arg1,arg2,"rw",file))
    return raiseIO();
  ext2[0] = MakeExternal(&inprocessExtDesc,(char *)file[0]);
  (void) KEEP(ext2[0]);
  ext2[1] = MakeExternal(&inprocessExtDesc,(char *)file[1]);
  (void) KEEP(ext2[1]);
  MAKECONS(ext2[0],ext2[1],ext);
  (void) KEEPONLY(ext);
  fields[0].label = readsym;
  fields[0].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_infile2),ext);
  fields[1].label = readlinesym;
  fields[1].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_linfile2),ext);
  fields[2].label = writesym;
  fields[2].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_outfile2),ext);
  fields[3].label = flushsym;
  fields[3].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_outfileflush2),ext);
  fields[4].label = killsym;
  fields[4].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_suspend2),ext);
  sortFieldArray(N,fields); /* we could sort it in advance */
  res = MakeRecordGnrl(N,fields,fieldsToShape(N,fields));
  QUIT;
  return res;
#undef N
}
ACTION_END

PRIVATE void instringSweep(extdata,doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  struct instring *f;

  f = (struct instring *)extdata;
  switch (doext->xtag)
    {
    case X_SAVE:
      outInt(f->pos);
      /*fallthrough*/
    case X_MARK:
    case X_MARKDATA:
      (*doext->action)(f->string);
      break;
    case X_FREE:
      MARKFREE(extdata);
      break;
    }
}

PRIVATE ExprPtr instringLoad(desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  struct instring *f;
  ExprPtr s,e;
  INT pos;
  
  pos = inInt();
  if (errno) return NIL;
  s = (*load)(sharedref);
  if (errno) return NIL;
  f = (struct instring *)MALLOC(sizeof(struct instring));
  f->string = s;
  f->pos = pos;
  f->n = SLEN(s);
  e = MakeExternal(desc,(char *)f);
  (void) KEEP(e);
  return e;
}

PUBLIC extDesc instringExtDesc = {
  (CHAR *)"string-as-instream",
  { 'i','n','s','t' }, NULL,
  NULL, /* no hash function, files are not shared */
  NULL, /* no comparison function, files are not shared */
  NULL, /* no print function */
  instringSweep,
  instringLoad
};

DEF_PRIVATE_ACTION_1ARG(q_instring)
{ /* ENV_UNDERLYING(env) is the external with the struct instring * inside */
  /* this code is based on substr() */
  REGISTER INT m,n;
  ExprPtr val;
  struct instring *f;

  f = (struct instring *)EXTDATA(ENV_UNDERLYING(env));
  m = f->pos;
  n = m+IVAL(arg);
  if (n>f->n)
    n = f->n;
  if (n<m)
    n = m;
#if STOP_AND_COPY_VECTORS
  {
    REGISTER INT size = n-m;
    CHAR *s;

    s = (CHAR *)alloctmp(size);
    safememcpy(s,(char *)&STEXT(f->string)[m],size);
    val = Kstring(size,s);
  }
#else
  val = Kstring(n-m,&STEXT(f->string)[m]);
#endif
  f->pos = n;
  return val;
}
ACTION_END

DEF_PRIVATE_ACTION_0ARG(q_linstring)
{
  struct instring *f;
  REGISTER CHAR *s,*sbeg,*send,c;
  ExprPtr val;

  f = (struct instring *)EXTDATA(ENV_UNDERLYING(env));
  s = sbeg = STEXT(f->string)+f->pos;
  send = STEXT(f->string)+f->n;
  for (; s<send && (c = *s++, c!='\n'); );
  {
    REGISTER INT size = s-sbeg;
#if STOP_AND_COPY_VECTORS
    CHAR *s;

    s = (CHAR *)alloctmp(size);
    safememcpy(s,(char *)&STEXT(f->string)[f->pos],size);
    val = Kstring(size,s);
#else
    val = Kstring(size,sbeg);
#endif
    f->pos += size;
  }
  return val;
}
ACTION_END

DEF_PRIVATE_ACTION_0ARG(q_instringtell)
{
  return Kint(((struct instring *)EXTDATA(ENV_UNDERLYING(env)))->pos);
}
ACTION_END

PRIVATE void instringseek(ins,n)
INOUT ExprPtr ins;
IN INT n;
{
  struct instring *f;

  f = (struct instring *)EXTDATA(ins);
  if (n<0)
    n = 0;
  if (n>f->n)
    n = f->n;
  f->pos = n;
}

DEF_PRIVATE_ACTION_1ARG(q_instringseek)
{
  instringseek(ENV_UNDERLYING(env),IVAL(arg));
  return tuple0;
}
ACTION_END

DEF_PRIVATE_ACTION_1ARG(q_instringadv)
{
  instringseek(ENV_UNDERLYING(env),
	       IVAL(arg)+((struct instring *)EXTDATA(ENV_UNDERLYING(env)))->pos);
  return tuple0;
}
ACTION_END

DEF_PRIVATE_ACTION_1ARG(q_instringseekend)
{
  instringseek(ENV_UNDERLYING(env),
	       IVAL(arg)+((struct instring *)EXTDATA(ENV_UNDERLYING(env)))->n);
  return tuple0;
}
ACTION_END

DEF_ACTION_1ARG(instring)
{
  struct instring *file;
  ExprPtr ext;
#define N 6
  field fields[N];
  ENTER;

  file = (struct instring *)MALLOC(sizeof(struct instring));
  file->string = arg;
  file->pos = 0;
  file->n = SLEN(arg);
  ext = MakeExternal(&instringExtDesc,(char *)file);
  (void) KEEP(ext);
  fields[0].label = readsym;
  fields[0].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_instring),ext);
  fields[1].label = readlinesym;
  fields[1].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_linstring),ext);
  fields[2].label = tellsym;
  fields[2].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_instringtell),ext);
  fields[3].label = seeksym;
  fields[3].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_instringseek),ext);
  fields[4].label = advancesym;
  fields[4].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_instringadv),ext);
  fields[5].label = seekendsym;
  fields[5].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_instringseekend),ext);
  sortFieldArray(N,fields); /* we could sort it in advance */
  arg = MakeRecordGnrl(N,fields,fieldsToShape(N,fields));
  QUIT;
  return arg;
#undef N
}
ACTION_END

DEF_ACTION_1ARG(dir)
{
  char *dirname;

  dirname = (char *)alloctmp(SLEN(arg)+1);
  safememcpy(dirname,STEXT(arg),SLEN(arg));
  dirname[SLEN(arg)] = 0;
  return listDir((CHAR *)dirname);
}
ACTION_END

DEF_ACTION_1ARG(filebits)
{
  char *name;

  name = (char *)alloctmp(SLEN(arg)+1);
  safememcpy(name,STEXT(arg),SLEN(arg));
  name[SLEN(arg)] = 0;
  return fileBits((CHAR *)name);
}
ACTION_END

DEF_ACTION_1ARG(cd)
{
  char *dirname;

  dirname = (char *)alloctmp(SLEN(arg)+1);
  safememcpy(dirname,STEXT(arg),SLEN(arg));
  dirname[SLEN(arg)] = 0;
  if (changeDir((CHAR *)dirname))
    return raiseIO();
  return tuple0;
}
ACTION_END

DEF_ACTION_0ARG(pwd)
{
  char *dirname;

  dirname = (char *)alloctmp(PWD_BUFSIZE+1);
  errno = 0;
  (void) getDir((CHAR *)dirname,PWD_BUFSIZE);
  if (errno)
    return raiseIO();
  return Kstr((CHAR *)dirname);
}
ACTION_END

DEF_ACTION_1ARG(errormsg)
{
  return ioErrorMsg(IVAL(arg));
}
ACTION_END

DEF_ACTION_1ARG(rmdir)
{
  char *name;

  name = (char *)alloctmp(SLEN(arg)+1);
  safememcpy(name,STEXT(arg),SLEN(arg));
  name[SLEN(arg)] = 0;
  if (rmDir((CHAR *)name))
    return raiseIO();
  return tuple0;
}
ACTION_END

DEF_ACTION_1ARG(mkdir)
{
  char *name;

  name = (char *)alloctmp(SLEN(arg)+1);
  safememcpy(name,STEXT(arg),SLEN(arg));
  name[SLEN(arg)] = 0;
  if (makeDir((CHAR *)name))
    return raiseIO();
  return tuple0;
}
ACTION_END

DEF_ACTION_1ARG(unlink)
{
  char *name;

  name = (char *)alloctmp(SLEN(arg)+1);
  safememcpy(name,STEXT(arg),SLEN(arg));
  name[SLEN(arg)] = 0;
  if (unlink(name))
    return raiseIO();
  return tuple0;
}
ACTION_END

DEF_ACTION_2ARGS(rename)
{
  char *old,*new;
  ExprPtr e;
  INT size;

  e = arg1;
  size = SLEN(e)+1;
  old = (char *)alloctmp(2*size);
  safememcpy(old,STEXT(e),SLEN(e));
  old[SLEN(e)] = 0;
  e = arg2;
  new = old+size;
  safememcpy(new,STEXT(e),SLEN(e));
  new[SLEN(e)] = 0;
  if (rename(old,new))
    return raiseIO();
  return tuple0;
}
ACTION_END

DEF_ACTION_1ARG(system)
{
  char *cmd;
  int res;

  cmd = (char *)alloctmp(SLEN(arg)+1);
  safememcpy(cmd,STEXT(arg),SLEN(arg));
  cmd[SLEN(arg)] = 0;
  res = system(cmd);
  if (res<0)
    return raiseIO();
  return Kint(res);
}
ACTION_END

DEF_ACTION_1ARG(getenv)
{
  char *cmd;

  cmd = (char *)alloctmp(SLEN(arg)+1);
  safememcpy(cmd,STEXT(arg),SLEN(arg));
  cmd[SLEN(arg)] = 0;
  cmd = getenv(cmd);
  if (cmd==NULL)
    return none;
  BEGINQUICKKEEP();
  arg = Kstr((CHAR *)cmd);
  QUICKKEEP(arg);
  arg = Kdata(0,optttm,arg); /* SOME arg */
  ENDQUICKKEEP();
  return arg;
}
ACTION_END

DEF_ACTION_0ARG(maybe)
{
  return (randLong() & (unsigned long)MAX_TWO_TO_THE_N_LONG)?trueValue:falseValue;
  /* works correctly only if unsigned longs are an integral number
     of bits, which we may assume quite safely */
}
ACTION_END

DEF_ACTION_0ARG(time)
{
  struct tms tms;
  REAL u,s;
  ExprPtr auxu,auxs,res;
  ENTER;

  (void) times(&tms);
  u = ((EXTENDED)tms.tms_utime)/((EXTENDED)HERTZ);
  s = ((EXTENDED)tms.tms_stime)/((EXTENDED)HERTZ);
  auxu = Kreal(&u);
  (void) KEEP(auxu);
  auxs = Kreal(&s);
  (void) KEEP(auxs);
  res = Ktuple2(auxu,auxs);
  QUIT;
  return res;
}
ACTION_END

DEF_ACTION_1ARG(dom)
{
  return Dom0(arg);
}
ACTION_END

DEF_ACTION_1ARG(rng)
{
  return Rng0(arg);
}
ACTION_END

DEF_ACTION_1ARG(inds)
{
  return Inds0(arg);
}
ACTION_END

DEF_ACTION_1ARG(elems)
{
  return Elems0(arg);
}
ACTION_END

DEF_ACTION_1ARG(listmap)
{
  return ListToMap0(arg);
}
ACTION_END

DEF_ACTION_1ARG(force)
{
  return (*ACTION(arg))(ENV_CONS(QENV(arg),0));
}
ACTION_END

/*!!!*/

#if CONTINUATION_MODEL & (CALLCC|CATCH)
DEF_PRIVATE_ACTION_1ARG(callcc1)
{ /* arg is the current continuation, and ENV_UNDERLYING(env) is a compiled closure;
     we must apply env to arg. */
  ExprPtr uenv = ENV_UNDERLYING(env);

  return (*ACTION(uenv))(ENV_CONS(QENV(uenv),1), arg);
}
ACTION_END
#endif

#if CONTINUATION_MODEL & CALLCC
DEF_ACTION_1ARG(callcc)
{
  return CallCC(ACTION_NAME(callcc1),arg,false);
}
ACTION_END
#endif

#if CONTINUATION_MODEL & CATCH
DEF_ACTION_1ARG(catch)
{
  return CallCC(ACTION_NAME(callcc1),arg,true);
}
ACTION_END
#endif

DEF_ACTION_0ARG(rand)
{
  return Kint(randLong());
}
ACTION_END

DEF_ACTION_1ARG(num)
{
  REAL x;

  x = (REAL)IVAL(arg);
  return Kreal(&x);
}
ACTION_END

DEF_ACTION_1ARG(int)
{
  INT n;

  if (arg->kind!=(CHAR)E_REAL)
    return Raise(INCOMPEXC);
  n = (INT)XVAL(arg);
  return Kint(n);
}
ACTION_END

DEF_ACTION_1ARG(iabs)
{
  REGISTER INT n;

  n = IVAL(arg);
  return (n<0)?Kint(-n):arg;
}
ACTION_END

DEF_ACTION_2ARGS(ilt)
{
  return (IVAL(arg1)<IVAL(arg2))?trueValue:falseValue;
}
ACTION_END

DEF_ACTION_2ARGS(ile)
{
  return (IVAL(arg1)<=IVAL(arg2))?trueValue:falseValue;
}
ACTION_END

DEF_ACTION_2ARGS(igt)
{
  return (IVAL(arg1)>IVAL(arg2))?trueValue:falseValue;
}
ACTION_END

DEF_ACTION_2ARGS(ige)
{
  return (IVAL(arg1)>=IVAL(arg2))?trueValue:falseValue;
}
ACTION_END

DEF_ACTION_2ARGS(imin)
{
  return (IVAL(arg1)>=IVAL(arg2))?arg2:arg1;
}
ACTION_END

DEF_ACTION_2ARGS(imax)
{
  return (IVAL(arg1)>=IVAL(arg2))?arg1:arg2;
}
ACTION_END

/* Integer division:
   a = b.q+r, |r|<|b| and r and b have the same sign.
   */

PUBLIC INT idiv(a,b)
IN REGISTER INT a,b;
{
  return a/b;
}

PUBLIC INT imod(a,b)
IN REGISTER INT a,b;
{
  return a%b;
}

DEF_ACTION_2ARGS(idiv)
{
  REGISTER INT a,b;

  a = IVAL(arg1);
  b = IVAL(arg2);
  if (b==0)
    return Raise(ARITHEXC);
  return Kint(idiv(a,b));
}
ACTION_END

DEF_ACTION_2ARGS(imod)
{
  REGISTER INT a,b;

  a = IVAL(arg1);
  b = IVAL(arg2);
  if (b==0)
    return Raise(ARITHEXC);
  return Kint(imod(a,b));
}
ACTION_END

DEF_ACTION_2ARGS(idivmod)
{
  REGISTER INT a,b,q,r;

  a = IVAL(arg1);
  b = IVAL(arg2);
  if (b==0)
    return Raise(ARITHEXC);
  q = a/b;
  r = a%b;
  {
    ExprPtr res;
    ENTER;

    res = Ktuple2(KINT(q),KINT(r));
    QUIT;
    return res;
  }
}
ACTION_END

DEF_ACTION_2ARGS(iplus)
{
  return Kint(IVAL(arg1)+IVAL(arg2));
}
ACTION_END

DEF_ACTION_2ARGS(iminus)
{
  return Kint(IVAL(arg1)-IVAL(arg2));
}
ACTION_END

DEF_ACTION_2ARGS(itimes)
{
  return Kint(IVAL(arg1)*IVAL(arg2));
}
ACTION_END

DEF_ACTION_1ARG(ineg)
{
  return Kint(-IVAL(arg));
}
ACTION_END

DEF_ACTION_1ARG(isqr)
{
  return Kint(IVAL(arg)*IVAL(arg));
}
ACTION_END

DEF_ACTION_2ARGS(ior)
{
  return Kint(IVAL(arg1) | IVAL(arg2));
}
ACTION_END

DEF_ACTION_2ARGS(ixor)
{
  return Kint(IVAL(arg1) ^ IVAL(arg2));
}
ACTION_END

DEF_ACTION_2ARGS(iand)
{
  return Kint(IVAL(arg1) & IVAL(arg2));
}
ACTION_END

DEF_ACTION_1ARG(inot)
{
  return Kint(~IVAL(arg));
}
ACTION_END

DEF_ACTION_2ARGS(ilsl)
{
  return Kint(IVAL(arg1) << IVAL(arg2));
}
ACTION_END

DEF_ACTION_2ARGS(ilsr)
{
  return Kint((INT)(((unsigned long)IVAL(arg1)) >> IVAL(arg2)));
}
ACTION_END

DEF_ACTION_2ARGS(iasr)
{
  return Kint(IVAL(arg1) >> IVAL(arg2));
}
ACTION_END

DEF_ACTION_1ARG(inc)
{
  ExprPtr i;

  i = Kint(IVAL(REFVAL(arg))+1);
  SETREFVAL(arg,i);
  return tuple0;
}
ACTION_END

DEF_ACTION_1ARG(dec)
{
  ExprPtr i;

  i = Kint(IVAL(REFVAL(arg))-1);
  SETREFVAL(arg,i);
  return tuple0;
}
ACTION_END

DEF_ACTION_1ARG(make_io_exc)
{
  ExprPtr res;
  ExprPtr uenv;

  uenv = ENV_UNDERLYING(env);
  MAKECONS(uenv,arg,res);
  return res;
}
ACTION_END

DEF_PRIVATE_ACTION_1ARG(q_map)
{ /* ENV_UNDERLYING(env) is the function to apply to each element of the list arg */
  ExprPtr aux;
  compiledAction f;
  ExprPtr uenv, fenv;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[2]; } _x;

  _x.n = 2;
  _x.next = theStackFrames;
  _x.p[0] = NIL;
  _x.p[1] = NOEXPR;
  theStackFrames = (gcStackView *)&_x;

  uenv = ENV_UNDERLYING(env);
  f = ACTION(uenv);
  fenv = ENV_CONS(QENV(uenv),1);
  for (; arg!=NIL; arg = CDR(arg))
    {
      _x.p[1] = (*f)(fenv, CAR(arg));
      MAKECONS(_x.p[1],_x.p[0],aux);
      _x.p[0] = aux;
      _x.p[1] = NOEXPR;
    }
  _x.p[0] = Reverse(_x.p[0]);
  theStackFrames = _x.next;
  return _x.p[0];
}
ACTION_END

DEF_ACTION_1ARG(map)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_map),arg);
}
ACTION_END

DEF_PRIVATE_ACTION_1ARG(q_app)
{ /* ENV_UNDERLYING(env) is the function to apply to each element of the list arg */
  compiledAction f;
  ExprPtr uenv, fenv;

  uenv = ENV_UNDERLYING(env);
  f = ACTION(uenv);
  fenv = ENV_CONS(QENV(uenv),1);
  for (; arg!=NIL; arg = CDR(arg))
    {
      (void) (*f)(fenv, CAR(arg));
    }
  return tuple0;
}
ACTION_END

DEF_ACTION_1ARG(app)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_app),arg);
}
ACTION_END

/*
fun fold f =
    let fun f2 nil = (fn b => b)
	  | f2 (e :: r) = (fn b => f(e,f2 r b))
    in
	f2
    end
*/

DEF_PRIVATE_ACTION_1ARG(q_fold1)
{ /* ENV_UNDERLYING(env) is the pair (f,neutral) */
  compiledAction f;
  ExprPtr uenv, fenv;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[2]; } _x;

  _x.n = 2;
  _x.next = theStackFrames;
  uenv = ENV_UNDERLYING(env);
  _x.p[0] = CDR(uenv);
  _x.p[1] = Nreverse0(arg,NIL);
  theStackFrames = (gcStackView *)&_x;

  f = ACTION(CAR(uenv));
#if MAX_ARITY>=2
  fenv = ENV_CONS(QENV(CAR(uenv)),2);
#else
  fenv = ENV_CONS(QENV(CAR(uenv)),1);
#endif
  for (; _x.p[1]!=NIL; _x.p[1] = CDR(_x.p[1]))
    {
#if MAX_ARITY>=2
      _x.p[0] = (*f)(fenv, CAR(_x.p[1]), _x.p[0]);
#else
      _x.p[0] = Ktuple2(CAR(_x.p[1]),_x.p[0]);
      _x.p[0] = (*f)(fenv, _x.p[0]);
#endif
    }
  theStackFrames = _x.next;
  return _x.p[0];
}
ACTION_END

DEF_PRIVATE_ACTION_COLLECT(q_fold)
{ /* ENV_UNDERLYING(env) is the function to apply, arg is the neutral element */
  ExprPtr e;

  env = ENV_UNDERLYING(env);
  BEGINQUICKKEEP();
  MAKECONS(env,arg,e);
  QUICKKEEP(e);
  e = MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_fold1),e);
  ENDQUICKKEEP();
  res = e;
}
ACTION_END_COLLECT

DEF_ACTION_1ARG(fold)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_fold),arg);
}
ACTION_END

/*
fun revfold f =
    let fun f2 nil = (fn b => b)
	  | f2 (e::r) = (fn b => f2 r (f (e,b)))
    in
	f2
    end
*/

DEF_PRIVATE_ACTION_1ARG(q_revfold1)
{ /* ENV_UNDERLYING(env) is the pair (f,neutral) */
  compiledAction f;
  ExprPtr uenv, fenv;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[1]; } _x;

  _x.n = 1;
  _x.next = theStackFrames;
  uenv = ENV_UNDERLYING(env);
  _x.p[0] = CDR(uenv);
  theStackFrames = (gcStackView *)&_x;

  f = ACTION(CAR(uenv));
#if MAX_ARITY>=2
  fenv = ENV_CONS(QENV(CAR(uenv)),2);
#else
  fenv = ENV_CONS(QENV(CAR(uenv)),1);
#endif
  for (; arg!=NIL; arg = CDR(arg))
    {
#if MAX_ARITY>=2
      _x.p[0] = (*f)(fenv, CAR(arg), _x.p[0]);
#else
      _x.p[0] = Ktuple2(CAR(arg),_x.p[0]);
      _x.p[0] = (*f)(_x.p[0],fenv);
#endif
    }
  theStackFrames = _x.next;
  return _x.p[0];
}
ACTION_END

DEF_PRIVATE_ACTION_COLLECT(q_revfold)
{ /* ENV_UNDERLYING(env) is the function to apply, arg is the neutral element */
  ExprPtr e;

  env = ENV_UNDERLYING(env);
  BEGINQUICKKEEP();
  MAKECONS(env,arg,e);
  QUICKKEEP(e);
  e = MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_revfold1),e);
  ENDQUICKKEEP();
  res = e;
}
ACTION_END_COLLECT

DEF_ACTION_1ARG(revfold)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_revfold),arg);
}
ACTION_END

DEF_ACTION_1ARG(hd)
{
  if (arg==NIL)
    return Raise(MATCHEXC);
  return CAR(arg);
}
ACTION_END

DEF_ACTION_1ARG(tl)
{
  if (arg==NIL)
    return Raise(MATCHEXC);
  return CDR(arg);
}
ACTION_END

DEF_ACTION_2ARGS(system_less)
{
  return (arg1<arg2)?trueValue:falseValue;
}
ACTION_END

DEF_ACTION_0ARG(getargs)
{
  return theArgs;
}
ACTION_END

DEF_ACTION_1ARG(atoi)
{
  CHAR *s, *send;
  Boolean neg;
  INT i;

  s = STEXT(arg);
  send = s + SLEN(arg);
  while (s<send && isspace(*s)) s++;
  neg = false;
  if (s<send && (*s=='~' || *s=='-'))
    { neg = true; s++; }
  i = 0;
  while (s<send && isdigit(*s))
    {
      i = 10*i + (INT) (*s++ - '0');
    }
  return Kint (i);
}
ACTION_END

DEF_ACTION_1ARG(atof)
{
  CHAR *tmp = (CHAR *)alloctmp(SLEN(arg)+1);
  CHAR *s, *t, *send, c;
  double d;
  EXTENDED x;

  for (s=STEXT(arg), send = s+SLEN(arg), t=tmp; s<send; )
    if ((c = *s++)=='~')
      *t++ = '-';
    else *t++ = c;
  *t++ = 0;
  sscanf ((char *)tmp, "%lf", &d);
  x = (EXTENDED)d;
  return Kext(&x);
}
ACTION_END

/*
abstype
  (''a,'ext) color = COLOR of ''a * (num * 'ext)
with
  eqtype color
  datatype (''a,'ext) ufind =
    UFIND of |[find : ''a -> (''a,'ext) color, (* raises UnColored if not found *)
	       oneof : (''a,'ext) color -> ''a, (* gets canonical element of color *)
	       info :  (''a,'ext) color -> 'ext, (* gets info associated with color *)
	       setinfo : (''a,'ext) color * 'ext -> unit,
	       merge : (''a * ''a * 'ext * 'ext -> 'ext) -> ''a * ''a -> ''a color,
	       colored : ''a -> bool,
	       copy : ('ext -> 'ext) -> (''a,'ext) ufind,
	       newcolor : ''a * 'info -> ''a color (* add a new color; unspecified if already there *)
	       ]|
  fun ufindNew : unit -> (''a,'ext) ufind
end
*/
