/* Stdio replacement.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* Replacing stdio was done initially so as to provide safe file functions
   on AIX when SPECIAL_MALLOC was set on a Unix box. SPECIAL_MALLOC redefines
   malloc() and free() to use a home-made library of functions, which are
   faster than run-of-the-mill implementations. On Unix, this home-made
   library uses sbrk() and manages heap space on its own; on the other hand,
   you must not use the standard malloc(), then, since it also wants to manage
   the heap. Unfortunately, most uses of malloc() in the AIX standard C
   library are *inlined*, which defeats the whole scheme of redefining them.
     The best solution is in fact to use the standard malloc()/free() functions
   (let SPECIAL_MALLOC be 0 in file OPTIONS), all the more so as modern
   implementations are fast enough. Moreover, the parts of HimML that were
   slowed down by inefficient implementations of malloc() have been optimized
   independently.
     Then, one good point about this library is that it's standard, and will
   behave mostly in the same way on all machines. But mostly, it provides
   a suspension feature for files (namely, releasing the file descriptor,
   while being able to reopen the file transparently, as though it had
   never been closed) in a mostly portable manner. This is useful in a
   garbage collected language as HimML.
*/

#include <stdio.h>
#include <math.h>
#include "portable.h"
#include "mem.h"
#include "timing.h"
#include "memproto.h"
#ifdef THINK_C
#include <unix.h>
#endif
#if OS==UNIX_SYS_V || OS==UNIX_BSD
/*#include <sys/types.h> already included in timing.h*/
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <signal.h>
#include <unistd.h>
#else
#include <errno.h>
#endif
#include <fcntl.h>
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#ifdef HAS_STDARG
#include <stdarg.h>
#else
#ifdef HAS_VARARGS
#include <varargs.h>
#endif
#endif
#if OS==AMIGA_OS
#include <dos.h>
#include <ios1.h>
#endif
#include "filproto.h"
#include "dirproto.h"
#ifdef HAS_PROTOTYPES
#include "sfile.h"
#endif

#if OS==AMIGA_OS
#include <dos/dos.h>
#include <proto/dos.h>
#include <proto/exec.h>
#include <exec/execbase.h>
EXTERN struct ExecBase *SysBase;
EXTERN unsigned long __defstksz;
#endif

#ifdef OBSOLETE
IMPORT long lseek P((int fildes,long offset,int whence));
#endif

#ifndef SEEK_SET
#define SEEK_SET 0
#endif
#ifndef SEEK_CUR
#define SEEK_CUR 1
#endif
#ifndef SEEK_END
#define SEEK_END 2
#endif

PRIVATE hFILE *fileList;

#define BUFSIZE 32
static char cvt_buf[BUFSIZE];

EXTERN Boolean initializing;

#if OS==UNIX_SYS_V || OS==UNIX_BSD
#ifndef WIFEXITED
#define	WIFEXITED(x)	( !((x) & 0xff) )
#endif
#ifndef WEXITSTATUS
#define	WEXITSTATUS(x)	(int)(WIFEXITED(x) ? (((x) >> 8) & 0xff) : -1)
#endif

int wait_pid(pid)
pid_type pid;
{
  int stat_loc,got_pid;

  stat_loc = 0;
  if (pid!=NO_PROCESS)
    {
      while (got_pid = wait(&stat_loc), pid!=got_pid);
    }
  return stat_loc;
}
#endif

#if OS==AMIGA_OS
#ifdef OBSOLETE
PRIVATE struct ProcId *subtasks = NO_PROCESS;
#endif

int wait_pid(pid)
pid_type pid;
{
  int exitcode;

  exitcode = 0;
  if (pid!=NO_PROCESS)
    {
      exitcode = wait(pid);
      DISPOSPTR(pid,sizeof(struct ProcID));
    }
  return exitcode;
}
#define WEXITSTATUS(n) n
#endif

PUBLIC char *ltoasc(a)
IN unsigned long a;
{ /* an itoa() function that we are sure does not allocate anything */
  /* this may not be the case with itoa(), because some compilers
     implement it by a call to sprintf(), which may in turn allocate
     some memory */
  char *s;

  s = &cvt_buf[BUFSIZE];
  *--s = 0;
  if (a==0L)
    *--s = '0';
  else do {
    *--s = '0'+(char)(a % 10L);
    a = a/10L;
  } while (a!=0L);
  return s;
}

PUBLIC char *rtoasc(xp)
IN CONST EXTENDED *xp;
#define DPT 3
#define MUL 1000.0
/* DPT is the number of figures after the comma,
   MUL is 10**DPT */
{ /* an ftoa() function that we are sure does not allocate anything */
  /* this may not be the case with ftoa(), because some compilers
     implement it by a call to sprintf(), which may in turn allocate
     some memory */
  char *s;
  COUNTER i;
  unsigned long a;

  s = &cvt_buf[BUFSIZE];
  *--s = 0;
  a = (unsigned long)((*xp)*MUL);
  for (i=0; i<DPT; i++)
    {
      *--s = '0'+(char)(a % 10L);
      a = a/10L;
    }
  *--s = '.';
  if (a==0L)
    *--s = '0';
  else do {
    *--s = '0'+(char)(a % 10L);
    a = a/10L;
  } while (a!=0L);
  return s;
}
#undef MUL
#undef DPT

PUBLIC char *timetoasc(t)
unsigned long t;
{
  EXTENDED x;

  x = ((EXTENDED)t)/((EXTENDED)HERTZ);
  return rtoasc(&x);
}

PRIVATE int getflags(mode,flagsp)
IN CONST char *mode;
OUT int *flagsp;
{
  int flags,oflag;
  char *s,c;

  flags = H_TEXTFILE;
  oflag = 0;
  s = mode;
  while ((c = *s++))
    switch (c)
      {
      case 'r': flags |= H_READ; break;
      case 'b': flags &= ~H_TEXTFILE; break;
      case '+': flags |= H_READ | H_WRITE; break;
      case 'w': flags |= H_WRITE; oflag |= O_CREAT | O_TRUNC; break;
      case 'a': flags |= H_WRITE; oflag |= O_CREAT | O_APPEND; break;
      default: errno = EINVAL; return 0;
      }
  *flagsp = flags;
  switch (flags & (H_READ | H_WRITE))
    {
    case H_READ: oflag |= O_RDONLY; break;
    case H_WRITE: oflag |= O_WRONLY; break;
    case H_READ | H_WRITE: oflag |= O_RDWR; break;
    default: errno = EINVAL; return -1;
    }
  return oflag;
}

PRIVATE hFILE *hfopen1(hf,name,oflag,flags,pos)
IN hFILE *hf;
IN CONST char *name;
IN int oflag,flags;
IN long pos;
{
  int fd;

  fd = open(name,oflag,0666);
  if (fd<0)
    return NULL;
#if OS==UNIX_SYS_V || OS==UNIX_BSD
  fcntl(fd,F_SETFD,FD_CLOEXEC);
#endif
  hf->ptr = hf->end = hf->buf;
  hf->fildesc = fd;
  pos = (long)lseek(fd,(off_t)pos,SEEK_CUR);
  hf->pos = hf->realpos = (pos<0)?0:pos;
  hf->flags = flags;
  return hf;
}

PUBLIC void unlinkbuf(hf)
hFILE *hf;
{
  REGISTER hFILE **ff,*fp;
  
  for (ff = &fileList; (fp = *ff) != NULL; ff = &fp->next)
    if (fp==hf)
      {
	*ff = fp->next;
	/* extract it from the file list */
	break;
      }
}

PUBLIC void linkbuf(hf)
hFILE *hf;
{
  hf->next = fileList;
  fileList = hf;
}

PRIVATE void disposebuf(hf)
hFILE *hf;
{
  if (hf->dirname!=NULL)
    {
      DISPOSPTR(hf->dirname,PWD_BUFSIZE+1);
      hf->dirname = NULL;
    }
  if (hf->filename!=NULL)
    {
      DISPOSPTR(hf->filename,strlen(hf->filename)+1);
      hf->filename = NULL;
    }
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#if SPECIAL_MALLOC
  if (theMZone!=(MZone *)0L &&
      ((char *)hf)>=(char *)&theMZone->MZ_block &&
      ((char *)hf)<(char *)MBnext(&theMZone->MZ_block))
    /* some files were allocated before the special
       malloc was installed; we must make sure we
       don't try to free them up */
#endif
#endif
  DISPOSPTR(hf,sizeof(hFILE));
}

PUBLIC hFILE *hfopen(name,mode)
IN CONST char *name,*mode;
{
  hFILE *hf;
  int oflag,flags;
  int len;

  oflag = getflags(mode,&flags);
  if (oflag==-1)
    return NULL;
  hf = (hFILE *)new_gc(sizeof(hFILE));
  if (hf==NULL)
    {
      errno = ENOMEM;
      return NULL;
    }
  hf->filename = NULL;
  hf->dirname = (CHAR *)new_gc(PWD_BUFSIZE+1);
  if (hf->dirname==NULL)
    {
      errno = ENOMEM;
err1:
      DISPOSPTR(hf,sizeof(hFILE));
      return NULL;
    }
  errno = 0;
  (void) getDir(hf->dirname,PWD_BUFSIZE);
  if (errno!=0)
    {
err2:
      DISPOSPTR(hf->dirname,PWD_BUFSIZE+1);
      goto err1;
    }
  len = strlen(name);
  hf->filename = (CHAR *)new_gc(len+1);
  if (hf->filename==NULL)
    {
      errno = ENOMEM;
      goto err2;
    }
  strcpy((char *)hf->filename,name);
  if (hfopen1(hf,name,oflag,flags,0L)==NULL)
    {
      disposebuf(hf);
      hf = NULL;
    }
  else linkbuf(hf);
  return hf;
}

PUBLIC int hfrevive(hf)
IN hFILE *hf;
{
  int oflag,res;
  long pos;
  CHAR *cwd;

  if (hf->fildesc!=-1 || hf->dirname==NULL || hf->filename==NULL)
    {
      errno = EINVAL;
      return -1;
    }
  oflag = 0;
  switch (hf->flags & (H_READ | H_WRITE))
    {
    case H_READ: oflag |= O_RDONLY; break;
    case H_WRITE: oflag |= O_WRONLY | O_APPEND; break;
    case H_READ | H_WRITE: oflag |= O_RDWR | O_APPEND; break;
    default: errno = EINVAL; return -1;
    }
  pos = hf->realpos;
  errno = 0;
  cwd = (CHAR *)alloctmp(PWD_BUFSIZE+1);
  (void) getDir(cwd,PWD_BUFSIZE);
  if (errno!=0)
    return -1;
  res = changeDir(hf->dirname);
  if (res!=0)
    return res;
  if (hfopen1(hf,(char *)hf->filename,oflag,hf->flags,pos)==NULL)
    {
      hf->fildesc = -1;
      hf->pos = hf->realpos = pos;
      hf->pid = NO_PROCESS;
      res = -1;
    }
  else res = 0;
  (void) changeDir(cwd);
  return res;
}

#if OS==UNIX_SYS_V || OS==UNIX_BSD
PRIVATE int pipebroken=0;
PRIVATE void sigpipe(code)
IN int code;
{
  pipebroken = 1;
}

PRIVATE int pipewrite(fd,s,len)
IN int fd;
IN char *s;
IN int len;
{
  void (*old)();
  int res;

  old = (void (*)())signal(SIGPIPE,sigpipe);
  pipebroken = 0;
  res = write(fd,s,len);
  (void) signal(SIGPIPE,old);
  if (pipebroken)
    errno = EPIPE;
  return res;
}
#else
#define pipewrite(fd,s,len) write(fd,s,len)
#endif

PUBLIC int hfflush(f)
IN hFILE *f;
{
  long off;
  int res,len;

  res = 0;
  if (f->fildesc!=-1)
    {
      len = f->ptr-f->buf;
      if ((f->flags & H_WRITE) && len!=0)
	{
	  if (f->pos!=f->realpos && !(f->flags & H_PROCESS))
	    {
	      off = (long)lseek(f->fildesc,(off_t)f->pos,SEEK_SET);
	      if (off==-1)
		return -1;
	    }
	  else off = f->realpos;
	  res = (f->flags & H_PROCESS)?pipewrite(f->fildesc,(char *)f->buf,len):
	    write(f->fildesc,f->buf,len);
	  res = (res==len)?0:(-1);
	  f->realpos = off+len;
	}
      f->ptr = f->end = f->buf;
      f->pos += len;
      if (res!=-1 && f->pos!=f->realpos)
	{
	  if (!(f->flags & H_PROCESS))
	    (void) lseek(f->fildesc,(off_t)f->pos,SEEK_SET);
	  /* there may be an error with interactive files,
	     on which we cannot lseek(), and
	     for which the real position is normally always 0 */
	  f->realpos = f->pos;
	}
    }
  return res;
}

PUBLIC long hftell(f)
IN hFILE *f;
{
  return (f->fildesc==-1)?f->realpos:(f->pos + (f->ptr - f->buf));
}

PUBLIC int hfseek(f,offset,whence)
IN hFILE *f;
IN long offset;
IN int whence;
{
  int res;
  CHAR *here;
  long off;

  if (f->fildesc==-1)
    {
      res = hfrevive(f);
      if (res!=0)
	return res;
    }
  if (f->flags & H_PROCESS)
    {
      errno = ESPIPE;
      return -1;
    }
  switch (whence)
    {
    case SEEK_SET: here = f->buf + (offset-f->pos); break;
    case SEEK_CUR: here = f->ptr + offset; break;
    case SEEK_END:
      off = (long)lseek(f->fildesc,(off_t)0L,SEEK_END);
      if (off==-1)
	return -1;
      here = f->buf + (off+offset-f->pos);
      break;
    default: errno = EINVAL; return -1;
    }
  if (here>=f->buf && here<=f->end &&
      (!(f->flags & H_WRITE) || here>=f->ptr))
    /* if the file is writable, we ensure that f->ptr will
       only increase, or else hfflush would not flush the
       right amount of data to the file */
    {
      f->ptr = here;
      res = 0;
    }
  else
    {
      offset = f->pos + (here - f->buf);
      hfflush(f);
      f->pos = f->realpos = offset;
      res = (long)lseek(f->fildesc,(off_t)offset,SEEK_SET);
    }
  if (res!=-1)
    f->flags &= ~H_EOF;
  return res;
}

PUBLIC int hfsuspend(f)
IN hFILE *f;
{
  int res;

  res = 0;
  if (f->flags & H_PROCESS) /* if process, kill it */
    {
      if (f->pid!=NO_PROCESS)
	{
#if OS==AMIGA_OS
	  struct Message *msg;

	  Forbid();
	  msg = GetMsg(f->pid->parent);
	  if (msg==NULL)
	    Signal((struct Task *)f->pid->child->mp_SigTask,
		   SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_D |
		   SIGBREAKF_CTRL_E | SIGBREAKF_CTRL_F);
	  /* send C-c, C-d, C-e, C-f, hoping that the task will exit politely */
	  else
	    PutMsg(f->pid->parent,msg);
	  Permit();
	  (void) close(f->fildesc);
	  f->fildesc = -1;
	  errno = wait(f->pid);
#else
	  res = kill(f->pid,9); /* kill child process */
	  (void) close(f->fildesc);
	  f->fildesc = -1;
	  if (res==0)
	    {
	      int n;

	      n = wait_pid(f->pid);
	      n = WEXITSTATUS(n);
	      errno = (n<0)?0:n;
	      if (errno!=0)
		res = -1;
	    }
#endif
	  f->pid = NO_PROCESS;
	}
      else
	{
	  errno = ESRCH;
	  res = -1;
	}
    }
  else
    {
      if (f->fildesc!=-1)
	{
	  res = hfflush(f);
	  if (res!=0)
	    return res;
	  res = close(f->fildesc);
	  f->fildesc = -1; /* to signal that file was suspended */
	}
    }
  return res;
}

PUBLIC int hfclose(f)
IN hFILE *f;
{
  int res;

  res = hfsuspend(f);
  unlinkbuf(f);
  disposebuf(f);
  return res;
}

PRIVATE int maxpipeRead = 1;

PRIVATE int hReadNewBuf(f)
IN hFILE *f;
{
  int res;

  if (f->fildesc==-1)
    {
      res = hfrevive(f);
      if (res!=0)
	return res;
    }
  if (!(f->flags & H_READ))
    {
      errno = EBADF;
      return -1;
    }
  if (f->flags & H_EOF)
    {
      errno = 0;
      return -1; /* end of file */
    }
  res = hfflush(f);
  if (res==-1)
    return res;
  if (f->flags & H_PROCESS)
    res = read(f->fildesc,f->buf,maxpipeRead);
  else
    res = read(f->fildesc,f->buf,hFILE_BUFSIZE);
  if (res==-1)
    return res;
  if (res==0)
    {
      f->flags |= H_EOF;
      errno = 0;
      return -1; /* end of file */
    }
  f->realpos += res;
  f->end = f->buf+res;
  return 0;
}

PUBLIC int hReadNewBufGetcN(f,n)
IN hFILE *f;
IN int n;
{
  int c;

  maxpipeRead = (n<hFILE_BUFSIZE)?n:hFILE_BUFSIZE;
  c = hReadNewBuf(f);
  if (c<0)
    return c;
  return (int)(*f->ptr++);
}

PUBLIC int hReadNewBufGetc(f)
IN hFILE *f;
{
  int c;

  maxpipeRead = 1;
  c = hReadNewBuf(f);
  if (c<0)
    return c;
  return (int)(*f->ptr++);
}

PUBLIC int hfgetc(f)
IN hFILE *f;
{
  return hgetc(f);
}

PRIVATE int hWriteNewBuf(f)
IN hFILE *f;
{
  int res,fd;

  fd = f->fildesc;
  if (fd==-1)
    {
      res = hfrevive(f);
      if (res!=0)
	return res;
    }
  if (!(f->flags & H_WRITE))
    {
      errno = EBADF;
      return -1;
    }
  res = hfflush(f);
  return res;
}

PUBLIC int hWriteNewBufPutc(c,f)
IN int c;
IN hFILE *f;
{
  int res;

  res = hWriteNewBuf(f);
  if (res<0)
    return res;
  *f->ptr++ = (CHAR)c;
  return c;
}

PUBLIC int hfputc(c,f)
IN int c;
IN hFILE *f;
{
  return hputc(c,f);
}

PUBLIC int hfputs(s,f)
IN CONST char *s;
IN hFILE *f;
{
  INT len;
  int res,fd;

  len = strlen(s);
  if (f->ptr+len<=f->buf+hFILE_BUFSIZE)
    {
      safememcpy(f->ptr,s,len);
      f->ptr += len;
      res = 0;
    }
  else
    {
      fd = f->fildesc;
      if (fd==-1)
	{
	  res = hfrevive(f);
	  if (res!=0)
	    return res;
	}
      res = hfflush(f);
      if (res>=0)
	{
	  if (len<=hFILE_BUFSIZE)
	    {
	      safememcpy(f->buf,s,len);
	      f->ptr = f->buf+len;
	    }
	  else
	    {
	      res = (f->flags & H_PROCESS)?pipewrite(fd,(char *)f->buf,len):write(fd,s,len);
	      f->realpos += len;
	      f->pos = f->realpos;
	    }
	}
    }
  return res;
}

#define PRINTF_BUFSIZE 4096
PRIVATE char *printfBuf;

#undef HAS_ARGS
#ifdef HAS_STDARG
#define HAS_ARGS
#endif
#ifdef HAS_VARARGS
#define HAS_ARGS
#endif

#ifdef HAS_ARGS
PUBLIC int hvfprintf(f,msg,ap)
IN hFILE *f;
IN CONST char *msg;
IN va_list ap;
{
  vsprintf(printfBuf,msg,ap);
  return hfputs(printfBuf,f);
}
#endif

#ifdef HAS_STDARG
PUBLIC int hfprintf(IN hFILE *f,IN CONST char *msg,...)
{
  va_list args;
  int res;

  va_start(args,msg);
  res = hvfprintf(f,msg,args);
  va_end(args);
  return res;
}
#else
#ifdef HAS_VARARGS
/*VARARGS*/
PUBLIC int hfprintf(f,msg,va_alist)
IN hFILE *f;
IN CONST char *msg;
va_dcl
{
  va_list args;
  int res;

  va_start(args);
  res = hvfprintf(f,msg,args);
  va_end(args);
  return res;
}
#else
PUBLIC int hfprintf(f,msg,a,b,c,d,e,f,g,h)
IN hFILE *f;
IN CONST char *msg;
IN long a,b,c,d,e,f,g,h;
{
  sprintf(printfBuf,msg,a,b,c,d,e,f,g,h);
  return hfputs(printfBuf,f);
}
#endif
#endif

PUBLIC hFILE *hfreopen(filename,mode,f)
IN CONST char *filename,*mode;
IN hFILE *f;
{
  int oflag,flags,res,len;

  res = hfsuspend(f);
  if (res!=0)
    {
err:
      unlinkbuf(f);
      disposebuf(f);
      return NULL;
    }
  errno = 0;
  (void) getDir(f->dirname,PWD_BUFSIZE);
  DISPOSPTR(f->filename,strlen(f->filename)+1);
  len = strlen(filename);
  f->filename = (CHAR *)MALLOC(len+1);
  if (f->filename==NULL)
    {
      errno = ENOMEM;
      goto err;
    }
  strcpy((char *)f->filename,filename);
  oflag = getflags(mode,&flags);
  if (oflag==-1 || hfopen1(f,filename,oflag,flags,0L)==NULL)
    goto err;
  linkbuf(f);
  return f;
}

PUBLIC int hftruncate(f,n)
INOUT hFILE *f;
IN INT n;
{
  int res;

  if (f->flags & H_PROCESS)
    {
      errno = EINVAL;
      return -1;
    }
#if OS==AMIGA_OS
  /* with SAS/S 6.51, we don't have ftruncate(), so we do something else */
  if (f->dirname==NULL || f->filename==NULL)
    {
      errno = EINVAL;
      return -1;
    }
  if (SysBase->SoftVer<36)
    {
      errno = EOSERR;
      return -1;
    }
  res = hfsuspend(f);
  if (res!=0)
    return res;
  { /* now, get a lock on the file; we could not do this while the file
       was still open (because of shared locks). */
    CHAR *cwd;
    BPTR fh;

    errno = 0;
    cwd = (CHAR *)alloctmp(PWD_BUFSIZE+1);
    (void) getDir(cwd,PWD_BUFSIZE);
    if (errno!=0)
      return -1;
    res = changeDir(f->dirname);
    if (res!=0)
      return res;
    fh = Open(f->filename,MODE_READWRITE);
    (void) changeDir(cwd);
    if (fh==NULL)
      {
	errno = IoErr();
	return -1;
      }
    (void) SetFileSize(fh,f->realpos,OFFSET_BEGINNING);
    errno = IoErr();
    res = errno?-1:0;
    (void) Close(fh);
  }
#else
  if (f->fildesc==-1)
    {
      res = hfrevive(f);
      if (res!=0)
	return res;
    }
  res = hfflush(f);
  if (res!=0)
    return res;
  res = ftruncate(f->fildesc,n);
#endif
  return res;
}

#if OS==UNIX_SYS_V || OS==UNIX_BSD
PRIVATE int my_execvp(file,argv)
char *file,*argv[];
{
  char *colon,*pathseq,*path,**newargv;
  int len,flen,vlen;

#ifdef DEBUG
  hfprintf (hstderr, "my_execp (%s)\n", file);
  hfflush (hstderr);
#endif
  flen = strlen(file);
  for (vlen=0; argv[vlen++]!=NULL; );
  newargv = (char **)alloctmp_anchored((vlen+1)*sizeof(char *), (char *)&file);
  if (strchr(file,'/')!=NULL)
    pathseq = ":";
  else if ((pathseq=getenv("PATH"))==NULL)
     pathseq = "";
  for (; pathseq[0]!=0; pathseq=colon+1)
    {
      colon = strchr (pathseq, ':');

      if (colon==NULL)
	colon = pathseq + strlen (pathseq);
      len = colon-pathseq;
      /*
	if (len==0)
	continue;
      */
      path = (char *)alloctmp_anchored(len+flen+5, (char *)&file);
      (void) strncpy(path,pathseq,len);
      path[len] = 0;
      if (len>0)
	(void) strcat(path,"/");
      (void) strcat(path,file);
#ifdef DEBUG
      hfprintf(hstderr, "Trying to exec %s\n", path);
      hfflush (hstderr);
#endif
      (void) execv(path,argv);
#ifdef DEBUG
      hfprintf (hstderr, "Failed.\n");
      hfflush (hstderr);
#endif
      if (errno==ENOEXEC)
	{
#ifdef DEBUG
	  hfprintf (hstderr, "Trying to launch /bin/sh\n");
	  hfflush (hstderr);
#endif
	  safememcpy(newargv+1,argv,vlen*sizeof(char *));
	  newargv[0] = "sh";
	  (void) execv("/bin/sh",newargv);
	  /*syserr("execv");*/
#ifdef DEBUG
	  hfprintf (hstderr, "Failed.\n");
	  hfflush (hstderr);
#endif
	}
    }
  return(-1);
}
#endif

#define MAX_QUOTE 100

#ifdef OBSOLETE
PRIVATE int hack_command(cmd,argc)
IN CONST char *cmd;
OUT int *argc;
{ /* hack cmd by inserting \0s at the right places */
  int len,i,level;
  char quote[MAX_QUOTE];
  char c;
  int chop;

  len = strlen(cmd);
  *argc = 1;
  level = 0;
  chop = true;
  for (i=0; i++<len;)
    {
      c = *cmd++;
      switch (c)
	{
	case ' ': case '\t': case '\n':
	  if (level==0)
	    {
	      cmd[-1] = 0;
	      if (!chop)
		(*argc)++;
	    }
	  break;
#if OS==AMIGA_OS
	case '*': /* the escape character on the Amiga */
#else
	case '\\':
#endif
	  if (i<len)
	    {
	      cmd++;
	      i++;
	    }
	  chop = false;
	  break;
	case '\'': case '\"': case '`':
	  if (level==0)
	    {
	      quote[level++] = c;
	    }
	  else if (c==quote[level-1])
	    {
	      level--;
	    }
	  else if (level<MAX_QUOTE)
	    quote[level++] = c;
	  chop = false;
	  break;
	default:
	  chop = false;
	  break;
	}
    }
  return len;
}
#endif

#ifdef OBSOLETE
PRIVATE void fill_argv(argv,argc,cmd)
OUT char **argv;
IN int argc;
IN CONST char *cmd;
{
  while (argc>0)
    {
      while (*cmd++ == 0);
      *argv++ = cmd-1;
      argc--;
      while (*cmd++);
    }
  *argv = NULL;
}
#endif

PRIVATE int new_gc2(flags,pp,size)
IN int flags;
OUT char **pp;
IN long size;
{
  pp[0] = pp[1] = NULL;
  if (flags & H_READ)
    {
      pp[0] = new_gc(size);
      if (pp[0]==NULL)
	{
	  errno = ENOMEM;
	  return -1;
	}
    }
  if (flags & H_WRITE)
    {
      pp[1] = new_gc(size);
      if (pp[1]==NULL)
	{
	  if (pp[0]!=NULL)
	    {
	      DISPOSPTR(pp[0],size);
	      pp[0] = NULL;
	    }
	  errno = ENOMEM;
	  return -1;
	}
    }
  return 0;
}

PRIVATE void dispos2(pp,size)
IN char **pp;
IN long size;
{
  if (pp[0]!=NULL)
    DISPOSPTR(pp[0],size);
  if (pp[1]!=NULL)
    DISPOSPTR(pp[1],size);
}

PUBLIC int hpopen(argv,mode,hff)
IN CONST char **argv;
IN CONST char *mode;
OUT hFILE **hff; /* file for read, file for write */
{
#if OS==UNIX_SYS_V || OS==UNIX_BSD
  pid_type pid;
  int oflag,flags;
  CHAR *cmdname[2];
  int rd[2],wr[2];
  int sem[2];
  long len;

  oflag = getflags(mode,&flags);
  if (oflag==-1 || (oflag & O_APPEND))
    {
      errno = EINVAL;
      return -1;
    }
  flags &= ~H_TEXTFILE;
  if (new_gc2(flags,(char **)hff,sizeof(hFILE)))
    {
      errno = ENOMEM;
      return -1;
    }
  len = strlen(argv[0]);
  if (new_gc2(flags,(char **)cmdname,len+1))
    {
      errno = ENOMEM;
err:
      dispos2((char **)hff,sizeof(hFILE));
      return -1;
    }
  if (flags & H_READ)
    {
      hff[0]->filename = cmdname[0];
      strcpy((char *)cmdname[0],argv[0]); /* register command as file name */
      hff[0]->dirname = NULL; /* dirname=NULL, so you cannot revive a killed process. */
    }
  if (flags & H_WRITE)
    {
      hff[1]->filename = cmdname[1];
      strcpy((char *)cmdname[1],argv[0]); /* register command as file name */
      hff[1]->dirname = NULL; /* dirname=NULL, so you cannot revive a killed process. */
    }
  rd[0] = rd[1] = wr[0] = wr[1] = -1;
  if (flags & H_READ)
    {
      if (pipe(rd))
	{
err2:
	  dispos2((char **)cmdname,len+1);
	  goto err;
	}
      /*
      fcntl(rd[0],F_SETFL,O_NONBLOCK);
      fcntl(rd[1],F_SETFL,O_NONBLOCK);
      */
    }
  if (flags & H_WRITE)
    {
      if (pipe(wr))
	{
	  /*err3:*/
	  if (flags & H_READ)
	    {
	      (void) close(rd[0]);
	      (void) close(rd[1]);
	    }
	  goto err2;
	}
    }
  if (pipe(sem))
    { /* sem is used to report failure of exec() from child to parent. */
err4:
      if (flags & H_WRITE)
	{
	  (void) close(wr[0]);
	  (void) close(wr[1]);
	}
      goto err2;
    }
  switch (pid = fork())
    {
    case -1:
      {
	int saveerrno;

	saveerrno = errno;
	if (flags & H_READ)
	  {
	    (void) close(rd[0]);
	    (void) close(rd[1]);
	  }
	if (flags & H_WRITE)
	  {
	    (void) close(wr[0]);
	    (void) close(wr[1]);
	  }
	(void) close(sem[0]);
	(void) close(sem[1]);
	errno = saveerrno;
      }
      goto err2;
    case 0: /* child */
      (void) close(sem[0]);
      (void) fcntl(sem[1],F_SETFD,FD_CLOEXEC); /* sem[1] should close on successful exec() */
      if (flags & H_READ)
	{
	  if (close(rd[0]) || close(1) || dup2(rd[1],1)==-1 || close(rd[1]))
	    /* redirect output */
	    goto exec_failed;
	}
      if (flags & H_WRITE)
	{
	  if (close(wr[1]) || close(0) || dup2(wr[0],0)==-1 || close(wr[0]))
	    /* redirect input */
	    goto exec_failed;
	}
      errno = 0;
      (void) my_execvp(argv[0],argv);
exec_failed:
      (void) write(sem[1],(char *)&errno,sizeof(int)); /* send errno to parent */
      exit(0);
    default: /* parent */
      (void) close(sem[1]);
      {
	int len,i;

	len = read(sem[0],(char *)&i,sizeof(int)); /* wait for errno from child */
	(void) close(sem[0]);
	if (len==sizeof(int)) /* then exec failed */
	  {
	    errno = i;
	    goto err4;
	  }
	/* otherwise, exec was successful, which we can tell by the fact that read()
	   returned 0; this happened because when exec() succeeded, it closed
	   the sem[] pipe, which made the read() above return 0. */
      }
      if (flags & H_READ)
	{
	  (void) close(rd[1]);
	  hff[0]->fildesc = rd[0];
	  fcntl(rd[0],F_SETFD,FD_CLOEXEC);
	  if (flags & H_WRITE)
	    hff[0]->pid = NO_PROCESS; /* pid is recorded in the write file instead */
	  else
	    hff[0]->pid = pid;
	  hff[0]->ptr = hff[0]->end = hff[0]->buf;
	  hff[0]->pos = hff[0]->realpos = 0;
	  hff[0]->flags = H_READ | H_PROCESS;
	  linkbuf(hff[0]);
	}
      if (flags & H_WRITE)
	{
	  (void) close(wr[0]);
	  hff[1]->fildesc = wr[1];
	  fcntl(wr[1],F_SETFD,FD_CLOEXEC);
	  hff[1]->pid = pid;
	  hff[1]->ptr = hff[1]->end = hff[1]->buf;
	  hff[1]->pos = hff[1]->realpos = 0;
	  hff[1]->flags = H_WRITE | H_PROCESS;
	}
      return 0;
    }
#else
#if OS==AMIGA_OS
  pid_type pid;
  CHAR *cmdname[2];
  int oflag,flags;
  struct FORKENV env;
  long len;

  pid = (pid_type)new_gc(sizeof(struct ProcID));
  if (pid==NULL)
    {
      errno = ENOMEM;
      return -1;
    }
  oflag = getflags(mode,&flags);
  if (oflag==-1 || (oflag & O_APPEND))
    {
      errno = EINVAL;
err0:
      DISPOSPTR(pid,sizeof(struct ProcID));
      return -1;
    }
  flags &= ~H_TEXTFILE;
  if (new_gc2(flags,(char **)hff,sizeof(hFILE)))
    {
      errno = ENOMEM;
      goto err0;
    }
  len = strlen(argv[0]);
  if (new_gc2(flags,(char **)cmdname,len+1))
    {
      errno = ENOMEM;
err:
      dispos2((char **)hff,sizeof(hFILE));
      goto err0;
    }
  env.priority = 0;
  env.stack = __defstksz;
  env.std_in = (BPTR)0L; /* default */
  env.std_out = (BPTR)0L; /* default */
  env.console = (BPTR)0L; /* default */
  env.msgport = NULL;
  if (flags & H_READ)
    {
      char *pr = alloctmp(80);

      hff[0]->filename = cmdname[0];
      strcpy(cmdname[0],argv[0]); /* register command as file name */
      hff[0]->dirname = NULL; /* dirname=NULL, so you cannot revive a killed process. */
      hff[0]->fildesc = -1;
      sprintf(pr,"PIPE:HimML_%lx",(long)hff[0]);
      env.std_out = Open(pr,MODE_NEWFILE); /* open pipe for writing */
      if (env.std_out==(BPTR)0L)
	{
	  errno = IoErr();
	  goto err;
	}
      hff[0]->fildesc = open(pr,O_RDONLY,0666);
      if (hff[0]->fildesc==-1)
	{
err1:
	  Close(env.std_out);
	  env.std_out = (BPTR)0L;
	  goto err;
	}
    }
  if (flags & H_WRITE)
    {
      char *pw = alloctmp(80);

      hff[1]->filename = cmdname[1];
      strcpy(cmdname[1],argv[0]); /* register command as file name */
      hff[1]->dirname = NULL; /* dirname=NULL, so you cannot revive a killed process. */
      hff[1]->fildesc = -1;
      sprintf(pw,"PIPE:HimML_%lx",(long)hff[1]);
      hff[1]->fildesc = open(pw,O_WRONLY,0666);
      if (hff[1]->fildesc==-1)
	{
err2:
	  if (flags & H_READ)
	    {
	      (void) close(hff[0]->fildesc);
	      goto err1;
	    }
	  goto err;
	}
      env.std_in = Open(pw,MODE_OLDFILE);
      if (env.std_in==(BPTR)0L)
	{
err3:
	  (void) close(hff[1]->fildesc);
	  errno = IoErr();
	  goto err2;
	}
    }
  if (forkv(argv[0],argv,&env,pid)==-1)
    {
err4:
      if (flags & H_WRITE)
	{
	  Close(env.std_in);
	  goto err3;
	}
      goto err2;
    }
  pid->nextID = NULL;
  if (flags & H_READ)
    {
      if (flags & H_WRITE)
	hff[0]->pid = NO_PROCESS; /* pid is recorded in the write file instead */
      else
	hff[0]->pid = pid;
      hff[0]->ptr = hff[0]->end = hff[0]->buf;
      hff[0]->pos = hff[0]->realpos = 0;
      hff[0]->flags = H_READ | H_PROCESS;
      linkbuf(hff[0]);
    }
  if (flags & H_WRITE)
    {
      hff[1]->pid = pid;
      hff[1]->ptr = hff[1]->end = hff[1]->buf;
      hff[1]->pos = hff[1]->realpos = 1;
      hff[1]->flags = H_WRITE | H_PROCESS;
      linkbuf(hff[1]);
    }
  return 0;
#else
  errno = EINVAL;
  hff[0] = hff[1] = NULL;
  return -1;
#endif
#endif
}

PUBLIC void EndHIO()
{
  hFILE *f,*next;

  for (f=fileList; f!=NULL; f=next)
    {
      next = f->next;
      (void) hfclose(f);
    }
  if (printfBuf!=NULL)
    DISPOSPTR(printfBuf,PRINTF_BUFSIZE);
  if (hstdin!=NULL)
    DISPOSPTR(hstdin,sizeof(hFILE));
  if (hstdout!=NULL)
    DISPOSPTR(hstdout,sizeof(hFILE));
  if (hstderr!=NULL)
    DISPOSPTR(hstderr,sizeof(hFILE));
}

#if OS==AMIGA_OS
void times(tms)
OUT struct tms *tms;
{
  (tms)->tms_stime = 0L;
  /* unfortunately, clock() is buggy under SAS/C 6.3,
     and returns sometimes random values. We then use
     timer() instead, unless it is not available. */
#ifndef _STRICT_ANSI
  {
    unsigned int clock[2];

    (void) timer(clock);
    tms->tms_utime = CLOCKS_PER_SEC*clock[0]+clock[1]/(1000000/CLOCKS_PER_SEC);
  }
#else
  tms->tms_utime = clock();
#endif
}
#endif

#define __SEG__ INIT

PUBLIC hFILE *hstdin,*hstdout,*hstderr;

PUBLIC void InitHIO()
{
  fileList = NULL;
  printfBuf = NULL;
  hstdin = hstdout = hstderr = NULL;
  printfBuf = (char *)new_gc(PRINTF_BUFSIZE);
  if (printfBuf==NULL)
    nomem();
  hstdin = (hFILE *)new_gc(sizeof(hFILE));
  if (hstdin==NULL)
    nomem();
  hstdin->dirname = hstdin->filename = NULL;
  hstdin->ptr = hstdin->end = hstdin->buf;
  hstdin->fildesc = fileno(stdin);
  hstdin->pid = NO_PROCESS;
  hstdin->pos = hstdin->realpos = 0;
  hstdin->flags = H_READ | H_TEXTFILE;
  hstdout = (hFILE *)new_gc(sizeof(hFILE));
  if (hstdout==NULL)
    nomem();
  hstdout->dirname = hstdout->filename = NULL;
  hstdout->ptr = hstdout->end = hstdout->buf;
  hstdout->fildesc = fileno(stdout);
  hstdout->pid = NO_PROCESS;
  hstdout->pos = hstdout->realpos = 0;
  hstdout->flags = H_WRITE | H_TEXTFILE;
  hstderr = (hFILE *)new_gc(sizeof(hFILE));
  if (hstderr==NULL)
    nomem();
  hstderr->dirname = hstderr->filename = NULL;
  hstderr->ptr = hstderr->end = hstderr->buf;
  hstderr->fildesc = fileno(stderr);
  hstderr->pid = NO_PROCESS;
  hstderr->pos = hstderr->realpos = 0;
  hstderr->flags = H_WRITE | H_TEXTFILE;
}
