/* Bytecode compiler.
   Copyright (C) 1995-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005,2008,2011 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "eval.h"
#if TRANSLATOR
#include "expr.h"
#include "refcount.h"
#include "mem.h"
#include "bc.h"
#include "bcaux.h"
#include "timing.h"
#include "debug.h"
#include "transfer.h"
#include "errproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "iter.h"
#include "rbc.h"
#ifdef HAS_PROTOTYPES
#include "sbc.h"
#endif
#ifdef CX
#include "cxaux.h"
#define BCFILE cxFile
#else

#define BCFILE bcFile
#endif

#define __SEG__ BYTECODE

#define DEBUG

EXTERN ExprPtr none,optttm;
EXTERN ExprPtr primCompileMap;

#if DEBUGGING!=0
PRIVATE ExprPtr currentDebugZone;
EXTERN Boolean debugEnabled;
#endif
#if PROFILING
EXTERN ExprPtr defaultProf;
#else
#define defaultProf NIL
#endif

PRIVATE void compileStmt P((BCFILE *, compEnv *, ExprPtr));
PRIVATE void compileStmtExpr P((BCFILE *, compEnv *, ExprPtr));
PRIVATE void compileExpr P((BCFILE *, compEnv *, ExprPtr));

struct argList {
  ExprPtr next; /* points to an extdesc with a struct argList inside. */
  compEnv *env;
  ExprPtr arg;
};

PRIVATE void arglistSweep (extdata, doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  switch (doext->xtag)
    {
    case X_MARKDATA:
    case X_SAVE: /* no save option; arg lists are private to this file. */
      break;
    case X_MARK:
      (*doext->action) (((struct argList *)extdata)->arg);
      (*doext->action) (((struct argList *)extdata)->next);
      /* don't mark env, which should be reachable from somewhere else. */
      break;
    case X_FREE:
      MARKFREE (extdata);
      break;
    }
}

PRIVATE ExprPtr arglistLoad (desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  return NIL; /* no way to load arg lists; they are private to this file. */
}

/* By the way, since we never load or save arglists, it is not
 necessary to registerExtDesc(&arglistExtDesc). */

PUBLIC extDesc arglistExtDesc = {
  (CHAR *)"arglist",
  { 'a','r','g','l' }, NULL,
  NULL, /* no hash function, integer arrays are not shared */
  NULL, /* no comparison function, integer arrays are not shared */
  NULL, /* no print function */
  arglistSweep,
  arglistLoad
};

PRIVATE ExprPtr arglistCons (arg, env, next)
IN CONST ExprPtr arg;
IN CONST compEnv *env;
IN CONST ExprPtr next;
{
  struct argList *argl;
  ExprPtr e;

  argl = (struct argList *)MALLOC (sizeof (struct argList));
  argl->arg = arg;
  argl->env = env;
  argl->next = next;
  e = MakeExternal (&arglistExtDesc, (char *)argl);
  (void) KEEP (e);
  return e;
}

PRIVATE ExprPtr arglistNext (al)
IN CONST ExprPtr al;
{
  return ((struct argList *)EXTDATA(al))->next;
}

PRIVATE compEnv *arglistEnv (al)
IN CONST ExprPtr al;
{
  return ((struct argList *)EXTDATA(al))->env;
}

PRIVATE ExprPtr arglistArg (al)
IN CONST ExprPtr al;
{
  return ((struct argList *)EXTDATA(al))->arg;
}

/* compilation to bytecode: */

PRIVATE void compileCGlobal(f,idx)
IN BCFILE *f;
IN INT idx;
{
#ifdef CX
  CHAR buf[64];

  cxLet(f);
  sprintf (buf, "%s = EVARS(GLOBENV)->expr[%d];", cxVarName(f,f->stackDepth-1), (int)idx);
  cxOutAction(f,buf);
#else
  bcPut(BC_CGLOBAL,f);
  bcPut((OPCODE)idx,f);
  bcPush(1,f);
#endif
}

/*!!!*/

PRIVATE INT findConstant(f,e)
IN BCFILE *f;
IN CONST ExprPtr e;
{
  REGISTER bccBuffer *buf;
  REGISTER INT i;
  REGISTER ExprPtr *ep,*epend;

  for (buf = f->cbuf, i = 0; buf!=NULL; buf = buf->next)
    {
      ep = buf->data;
      epend = ep+buf->n;
      for (; ep<epend; i++)
	if (e == *ep++)
	  return i;
    }
  return -1;
}

PUBLIC INT getConstant(f,e)
IN BCFILE *f;
IN CONST ExprPtr e;
{
  INT i;

  i = findConstant(f,e);
  if (i<0)
    {
      i = bccTell(f);
      bccPut(e,f);
    }
  return i;
}

PRIVATE void compileConstant(f,e)
IN BCFILE *f;
IN CONST ExprPtr e;
{
  INT i;

  bcPut(BC_CONST,f);
  i = getConstant(f,e);
  bcPut((OPCODE)i,f);
  bcPush(1,f);
}

PRIVATE void compileClosure(f,code)
IN BCFILE *f;
IN CONST ExprPtr code;
{
  INT i;

  i = getConstant(f,code);
  bcPut(BC_CLOSURE,f);
  bcPut((OPCODE)i,f);
}

#ifdef HAS_PROTOTYPES
PRIVATE void compileFetchOp(IN BCFILE *f, IN OPCODE op, IN OPCODE specop,
			    IN INT n)
#else
PRIVATE void compileFetchOp(f,op,specop,n)
IN BCFILE *f;
IN OPCODE op,specop;
IN INT n;
#endif
{
  n -= f->stackDepth;
  if (specop!=0 && n>=-8 && n<0)
    {
      bcPut(specop+8+n,f);
    }
  else
    {
      bcPut(op,f);
      bcPut((OPCODE)n,f);
    }
  bcPush(1,f);
}

PRIVATE void compileFetch(f,n)
IN BCFILE *f;
IN INT n;
{
  compileFetchOp(f,BC_FETCH,BC_FETCH_m8,n);
}

PRIVATE void compileFetchCar(f,n)
IN BCFILE *f;
IN INT n;
{
  compileFetchOp(f,BC_FETCHCAR,BC_FETCHCAR_m8,n);
}

PRIVATE void compileFetchCdr(f,n)
IN BCFILE *f;
IN INT n;
{
  compileFetchOp(f,BC_FETCHCDR,BC_FETCHCDR_m8,n);
}

PRIVATE void compileFetchRefVal(f,n)
IN BCFILE *f;
IN INT n;
{
  compileFetchOp(f,BC_FETCHREFVAL,BC_FETCHREFVAL_m8,n);
}

PRIVATE void compileNth(f,idx)
IN BCFILE *f;
IN INT idx;
{
  if (idx>=0 && idx<8)
    bcPut(BC_NTH_0+idx,f);
  else
    {
      bcPut(BC_NTH,f);
      bcPut((OPCODE)idx,f);
    }
}

PRIVATE void bcdPutInt(i,f)
IN INT i;
IN BCFILE *f;
{
  bcdPut((OPCODE)(i>>(8*sizeof(OPCODE))),f);
  bcdPut((OPCODE)(i & ((1 << (8*sizeof(OPCODE))) - 1)),f);
}

#if DEBUGGING!=0
PRIVATE void compileCodePoint(f)
IN BCFILE *f;
{
  INT pc;

  if (!debugEnabled)
    return;
  pc = bcTell(f);
  if ((INT)(OPCODE)pc == pc)
    {
      bcdPut(DBG_WHERE,f);
      bcdPut((OPCODE)pc,f);
    }
  else
    {
      bcdPut(DBG_WHERE_BIG,f);
      bcdPutInt(pc,f);
    }
  bcdPut((OPCODE)f->stackDepth,f);
  bcdPut((OPCODE)f->mapStackDepth,f);
  bcdPut ((OPCODE)0,f); /* was: bcdPut((OPCODE)f->excStackDepth,f);*/
}

#ifdef HAS_PROTOTYPES
PRIVATE void compileDebugZone(IN Boolean startp, IN BCFILE *f)
#else
PRIVATE void compileDebugZone(startp,f)
IN Boolean startp;
IN BCFILE *f;
#endif
{
  if (currentDebugZone==NULL)
    return;
  if (!debugEnabled)
    return;
  {
    ExprPtr newfile;

    newfile = CAR(CDR(ARGS(currentDebugZone)));
    if (f->lastDebugZone==NULL ||
	CAR(CDR(ARGS(f->lastDebugZone)))!=newfile)
    { /* File name has changed: */
      INT i;

      i = getConstant (f, newfile);
      bcdPut(DBG_FILE,f);
      bcdPut((OPCODE)i,f);
    }
    f->lastDebugZone = currentDebugZone;
  }
  {
    COUNTER oldline, newline;
    COUNTER oldpos, newpos;

    oldline = f->lastLine;
    oldpos = f->lastPos;
    if (startp)
      {
	newline = ZONE(currentDebugZone)->line1;
	newpos = ZONE(currentDebugZone)->pos1;
      }
    else
      {
	newline = ZONE(currentDebugZone)->line2;
	newpos = ZONE(currentDebugZone)->pos2;
      }
    /* We add 1 to line2 (when at the end of an instruction),
       because we assume that every codepoint is on a different line.
       This means that in, say:
       let val x=3
       in
         x+1
       end
       the codepoint at the end of val x=3 will be shown on the 'in' line. */
    if (newline!=oldline || newpos!=oldpos)
      {
	bcdPut(DBG_LINE,f);
	bcdPut((OPCODE)newline,f);
	bcdPut((OPCODE)newpos, f);
      }
    f->lastLine = newline;
    f->lastPos = newpos;
  }
}

#ifdef HAS_PROTOTYPES
PRIVATE void compileCodePointOp(IN OPCODE op,IN BCFILE *f)
#else
PRIVATE void compileCodePointOp(op,f)
IN OPCODE op;
IN BCFILE *f;
#endif
{
  compileCodePoint(f);
  bcPut(op,f);
}
#else
#define compileCodePoint(f)
#define compileCodePoint(op,f) bcPut(op,f)
#define compileDebugZone(startp,f)
#endif

PRIVATE INT getCurIEnv(env)
IN CONST compEnv *env;
{
  ExprPtr i_l,index;
  ENTER;

  i_l = KONS(NIL,KINT(env->ilevel));
  index = MapGetSure(i_l,CE_VARS(env));
  QUIT;
  return IVAL(index);
}

PRIVATE INT getCurEnv(env)
IN CONST compEnv *env;
{
  ExprPtr i_l,index;
  ENTER;

  i_l = KONS(NIL,KINT(env->level));
  index = MapGetSure(i_l,CE_VARS(env));
  QUIT;
  return IVAL(index);
}

PRIVATE void newCurEnv(env,n)
INOUT compEnv *env;
IN COUNTER n;
{
  ExprPtr i_l;
  ENTER;

  i_l = KONS(NIL,KINT(env->level));
#ifdef DEBUG
  {
    ExprPtr v;
    Boolean found;

    v = MapGet(i_l,CE_VARS(env),&found);
    if (found && IVAL(v)!=n)
      berror(message(661));
  }
#endif
  setMapPut(i_l,KINT(n),&CE_VARS(env)); /* record that we've built the new env */
  QUIT;
}

PRIVATE void compileCurEnv(f,env)
IN BCFILE *f;
IN CONST compEnv *env;
{
  compileFetch(f,getCurEnv(env));
  /*
  bcPut(BC_FETCH,f);
  bcPut((OPCODE)(getCurEnv(env)-f->stackDepth),f);
  bcPush(1,f);
  */
}

PRIVATE void compileCurIEnv(f,env)
IN BCFILE *f;
IN CONST compEnv *env;
{
  compileFetch(f,getCurIEnv(env));
  /*
  bcPut(BC_FETCH,f);
  bcPut((OPCODE)(getCurIEnv(env)-f->stackDepth),f);
  bcPush(1,f);
  */
}

PRIVATE void compileClosureEnv(f,code,env)
IN BCFILE *f;
IN CONST ExprPtr code;
IN CONST compEnv *env;
{
  if (env!=NULL)
    { /* If func has free variables, then build a full closure: */
      compileCurEnv(f,env);
      compileClosure(f,code);
    }
  else /* otherwise, this is a function without free variables, and we use a shortcut: */
    {
      INT i;

      i = getConstant(f,code);
      bcPut(BC_CLOSURE0,f);
      bcPut((OPCODE)i,f);
      bcPush(1,f);
    }
}

PRIVATE BCFILE *bcOpenD (env)
IN CONST compEnv *env;
{
  BCFILE *f;

  f = bcOpenE (env);
#if DEBUGGING
  if (debugEnabled)
    {
      INT i;

      bcdPut (DBG_ENV, f);
      i = getConstant (f, CE_PARENT_DBGNAMES(env));
      bcdPut ((OPCODE)i, f);
    }
#endif
  return f;
}

PRIVATE ExprPtr bcCloseD(f)
IN BCFILE *f;
{
#if DEBUGGING
  if (debugEnabled)
    {
      bcdPut (DBG_STOP, f);
    }
#endif
  return bcClose (f);
}

PRIVATE void compileDataCon(f,env,flags,lbl)
IN BCFILE *f;
IN CONST compEnv *env;
IN INT flags;
IN CONST ExprPtr lbl;
{
  BCFILE *ff;
  ExprPtr e,saveprof;
  INT i;
  ENTER;

  saveprof = CE_PROF(env);
  CE_PROF(env) = defaultProf;

  if (lbl->kind==(CHAR)E_DATA)
    {
      if (lbl->magic<0)
	compileConstant(f,lbl);
      else
	{
	  ff = bcOpenE(env); /* bcOpenE(), not bcOpenD(), as first position
				in the stack is not an actual environment
				(it is empty) from which we can read variable values. */
	  /* --- env arg */
	  bcPut((lbl->mark & MARK_SHARED)?BC_KDATA:BC_DATA,ff);
	  bcPut((OPCODE)lbl->magic,ff);
	  i = getConstant(ff,CAR(lbl));
	  bcPut((OPCODE)i,ff);
	  /* --- env con(arg) */
	  compileDebugZone(false,ff);
	  compileCodePointOp(BC_RETURN,ff);
	  bcPut((OPCODE)1,ff); /* --- con(arg) */
	  e = bcCloseD (ff);

	  compileClosureEnv(f,e,NULL);
	}
    }
  else /* exceptions */
    {
      if (flags & 2) /* generative */
	compileExpr(f,env,lbl);
      else
	compileConstant(f,lbl);
      /* compile constructor, and install it as fake environment
	 for the closure */

      ff = bcOpenE(env); /* bcOpenE(), not bcOpenD(), as first position
			    in the stack is not an actual environment
			    from which we can read variable values. */
      /* --- con arg */
      bcPut((flags & 1)?BC_CONS:BC_KONS,ff); /* --- con(arg) */
      bcPush(-1,ff);
      compileDebugZone(false,ff);
      compileCodePointOp(BC_RETURN,ff);
      bcPut((OPCODE)0,ff); /* --- con(arg) */
      e = bcCloseD (ff);

      compileClosure(f,e);
    }
  CE_PROF(env) = saveprof;
  QUIT;
}

#ifdef HAS_PROTOTYPES
PRIVATE void compileUnary(IN BCFILE *f, IN CONST compEnv *env,
			  IN OPCODE opcode, IN CONST ExprPtr e)
#else
PRIVATE void compileUnary(f,env,opcode,e)
IN BCFILE *f;
IN CONST compEnv *env;
IN OPCODE opcode;
IN CONST ExprPtr e;
#endif
{
  compileExpr(f,env,e);
  bcPut(opcode,f);
}

#ifdef HAS_PROTOTYPES
PRIVATE void compileRecord1(IN BCFILE *f, IN CONST compEnv *env,
			    IN OPCODE opcode, IN CONST ExprPtr e)
#else
PRIVATE void compileRecord1(f,env,opcode,e)
IN BCFILE *f;
IN CONST compEnv *env;
IN OPCODE opcode;
IN CONST ExprPtr e;
#endif
{
  INT i;

  compileExpr(f,env,CDR(e));
  bcPut(opcode,f);
  i = getConstant(f,CAR(e));
  bcPut((OPCODE)i,f);
}

#ifdef HAS_PROTOTYPES
PRIVATE void compileBinary(IN BCFILE *f, IN CONST compEnv *env,
			   IN OPCODE opcode, IN CONST ExprPtr e1, IN CONST ExprPtr e2)
#else
PRIVATE void compileBinary(f,env,opcode,e1,e2)
IN BCFILE *f;
IN CONST compEnv *env;
IN OPCODE opcode;
IN CONST ExprPtr e1,e2;
#endif
{
  compileExpr(f,env,e1);
  compileExpr(f,env,e2);
  bcPut(opcode,f);
  bcPush(-1,f);
}

#ifdef HAS_PROTOTYPES
PRIVATE void compileTupleGnrl(IN BCFILE *f, IN CONST compEnv *env,
			      IN OPCODE opcode, IN CONST ExprPtr e)
#else
PRIVATE void compileTupleGnrl(f,env,opcode,e)
IN BCFILE *f;
IN CONST compEnv *env;
IN OPCODE opcode;
IN CONST ExprPtr e;
#endif
{
  INT i,n;

  n = AVAL(e)->n;
  for (i=0; i<n; i++)
    compileExpr(f,env,AVAL(e)->expr[i]);
  bcPut(opcode,f);
  bcPut((OPCODE)n,f);
  bcPush(1-n,f);
}

#ifdef HAS_PROTOTYPES
PRIVATE void compileRecordGnrl(IN BCFILE *f, IN CONST compEnv *env,
			       IN OPCODE opcode, IN CONST ExprPtr e)
#else
PRIVATE void compileRecordGnrl(f,env,opcode,e)
IN BCFILE *f;
IN CONST compEnv *env;
IN OPCODE opcode;
IN CONST ExprPtr e;
#endif
{
  INT i,n,level0,wlevel,level;
  ExprPtr shape;

  level0 = f->stackDepth;
  shape = ASHAPE(e);
  n = AVAL(e)->n;
  for (i=0; i<n; i++)
    {
      wlevel = level0+IVAL(AVAL(shape)->expr[i])-1;
      if (wlevel>f->stackDepth)
	{
	  bcPut(BC_NOEXPR,f);
	  bcPut((OPCODE)(wlevel-f->stackDepth),f);
	  bcPush(wlevel-f->stackDepth,f);
	}
      level = f->stackDepth; /* here, wlevel<=level */
      compileExpr(f,env,AVAL(e)->expr[i]);
      if (wlevel<level)
	{
	  bcPut(BC_TUCK,f);
	  bcPut((OPCODE)(wlevel-level),f);
	  bcPush(-1,f);
	}
    }
  i = getConstant(f,ASHAPE(shape));
  bcPut(opcode,f);
  bcPut((OPCODE)n,f);
  bcPut((OPCODE)i,f);
  bcPush(1-n,f);
}

PRIVATE void compileVarref(f,env,level,rank)
IN BCFILE *f;
IN CONST compEnv *env;
IN COUNTER level,rank;
{
  level = env->level-level;
  if (env->globenvs & (1L << level)) /* level is GLOBENV (or a copy of it) */
    {
#ifdef RISKY
      if (rank<EVARS(GLOBENV)->n)
	compileConstant(f,EVARS(GLOBENV)->expr[rank]);
      /* known constants are compiled right away */
      else
#endif
	{
	  bcPut(BC_GLOBVARREF,f);
	  bcPut((OPCODE)rank,f);
	  bcPush(1,f);
	}
    }
  else
    {
      ExprPtr i_l,index;
      Boolean found;
      ENTER;

      i_l = KINT(rank);
      if (level!=0)
	i_l = KONS(i_l,KINT(level));
      index = MapGet(i_l,CE_VARS(env),&found);
      if (found) /* variable already on stack */
	{
	  compileFetch(f,IVAL(index));
	  /*
	  bcPut(BC_FETCH,f);
	  bcPut((OPCODE)(IVAL(index)-f->stackDepth),f);
	  */
	}
      else
	{
	  COUNTER i;

	  i = level;
	  do {
	    i_l = KONS(NIL,KINT(i));
	    index = MapGet(i_l,CE_VARS(env),&found);
	    if (found)
	      break;
	  } while (i++!=env->level);
	  /* level should always be >=0 here! */
	  switch (i-level)
	    {
	    case 0:
	      bcPut(BC_VARREF0,f);
	      bcPut((OPCODE)(IVAL(index)-f->stackDepth),f);
	      break;
	    case 1:
	      bcPut(BC_VARREF1,f);
	      bcPut((OPCODE)(IVAL(index)-f->stackDepth),f);
	      break;
	    case 2:
	      bcPut(BC_VARREF2,f);
	      bcPut((OPCODE)(IVAL(index)-f->stackDepth),f);
	      break;
	    case 3:
	      bcPut(BC_VARREF3,f);
	      bcPut((OPCODE)(IVAL(index)-f->stackDepth),f);
	      break;
	    default:
	      bcPut(BC_VARREF,f);
	      bcPut((OPCODE)(IVAL(index)-f->stackDepth),f);
	      bcPut((OPCODE)(i-level),f);
	      break;
	    }
	  bcPut((OPCODE)rank,f);
	  bcPush(1,f);
	}
      QUIT;
    }
}

#if DEBUGGING!=0
PRIVATE Boolean debugVarref(f,env,level,rank)
IN BCFILE *f;
IN CONST compEnv *env;
IN COUNTER level,rank;
{
  if (!debugEnabled)
    return false;
  /*level = env->level-level;*/
  if (env->globenvs & (1L << level)) /* level is GLOBENV (or a copy of it) */
    {
      return false;
      /* do not record any debugging information for global variables */
    }
  else
    {
      ExprPtr i_l,index;
      Boolean found;
      ENTER;

      i_l = KINT(rank);
      if (level!=0)
	i_l = KONS(i_l,KINT(level));
      index = MapGet(i_l,CE_VARS(env),&found);
      if (found) /* variable already on stack */
	{
	  bcdPut(DBG_ACCESS,f);
	  bcdPut((OPCODE)(IVAL(index)),f);
	}
      else
	{
	  COUNTER i;

	  i = level;
	  do {
	    i_l = KONS(NIL,KINT(i));
	    index = MapGet(i_l,CE_VARS(env),&found);
	    if (found)
	      break;
	  } while (i++!=env->level);
	  /* level should always be >=0 here! */
	  bcdPut(DBG_VARREF,f);
	  bcdPut((OPCODE)(IVAL(index)),f);
	  bcdPut((OPCODE)(i-level),f);
	  bcdPut((OPCODE)rank,f);
	}
      QUIT;
      return true;
    }
}
#endif

PRIVATE void compileSwDrop(f,excess)
IN BCFILE *f;
IN INT excess;
{
  if (excess>1)
    {
      bcPut(BC_SWDROP,f);
      bcPut((OPCODE)-excess,f);
      bcPush(1-excess,f);
    }
}

#ifdef HAS_PROTOTYPES
PRIVATE void compileGoto(IN BCFILE *f, IN OPCODE opcode,
			 INOUT ExprPtr gotop, IN COUNTER drop)
#else
PRIVATE void compileGoto(f,opcode,gotop,drop)
IN BCFILE *f;
IN OPCODE opcode;
INOUT ExprPtr gotop;
IN COUNTER drop;
#endif
{
  ExprPtr g;
  INT excess,addr;
  Boolean patch = false;
  OPCODE offset;

  bcPush(-drop,f);
  if (f->dead)
    return;
  g = CAR(gotop);
  switch (g->kind)
    {
    case E_VARREF: /* backward jump */
      excess = f->stackDepth-VARLEVEL(g);
      if (excess<0)
	berror(message(609));
      if (excess==0)
	patch = gotoOp0(&opcode); /* special case for some opcodes, when they do not drop
				     any element from the stack. */
      bcPut(opcode,f);
      bcPut((OPCODE)excess,f);
      addr = bcTell(f);
      offset = (OPCODE)(VARRANK(g)-addr);
      if (addr+offset!=VARRANK(g))
	berror(message(94));
      if (patch)
	offset++; /* because pc will be 1 too low when taking the branch in that case */
      bcPut(offset,f);
      break;
    case E_NIL: case E_CONS: /*case E_TUPLE_2:*/ /* forward jump */
      {
	ExprPtr aux;
	ENTER;

	bcPut(opcode,f);
	bcPut((OPCODE)-1,f); /* dummy, to be patched later on. */
	aux = Kvarref(bcTell(f),f->stackDepth);
	(void) KEEP(aux);
	g = Kons(aux,g);
	SETCAR(gotop,g);
	QUIT;
	bcPut((OPCODE)-1,f); /* dummy, to be patched later on. */
      }
      break;
    default:
      kinderror(message(616),g->kind);
      /*NOTREACHED*/
      break;
    }
  if (opcode==BC_GOTO || opcode==BC_GOTO_0)
    f->dead = 1;
}

PRIVATE ExprPtr beginForward(f)
IN BCFILE *f;
{
  ExprPtr gotop;

  gotop = MakePair(E_CONS,NIL,(f->dead?trueValue:falseValue));
  (void) KEEP(gotop);
  return gotop;
}

PRIVATE void endForward0(f,gotop)
IN BCFILE *f;
INOUT ExprPtr gotop;
{
  ExprPtr ll,vr;
  INT excess;

  for (ll = CAR(gotop); ll!=NIL; ll=CDR(ll))
    {
      vr = CAR(ll);
      switch (ll->kind)
	{
	case E_CONS: /* normal goto */
	  excess = VARLEVEL(vr)-f->stackDepth;
	  if (excess<0)
	    berror(message(608));
	  bcPatchJump(VARRANK(vr),excess,f);
	  break;
	case E_TUPLE_2: /* switch table entry */
	  bcPatchSwitchEntry(VARRANK(vr),f);
	  break;
	}
    }
}

PRIVATE void endForward(f,gotop)
IN BCFILE *f;
INOUT ExprPtr gotop;
{
  endForward0(f,gotop);
  if (CAR(gotop)!=NIL) /* in which case this is coming from (potentially) live code */
    { /* Indeed, compileGoto's in dead code are disabled, so only goto's coming from
	 live code are recorded. */
      if (f->stackDepth>f->maxStackDepth)
	berror(message(643));
      f->dead = 0;
      SETCAR(gotop,tuple0); /* to signal compileGoto() that it is now illegal to jump to gotop. */
#ifdef INSERT_MERGE_NODES
      bcPut(BC_MERGE,f);
#endif
    }
}

#define isForward(gotop) (CAR(gotop)->kind==(CHAR)E_CONS)

#ifdef HAS_PROTOTYPES
PRIVATE void beginBackward0(IN BCFILE *f, INOUT ExprPtr gotop, IN OPCODE opcode)
#else
PRIVATE void beginBackward0(f,gotop,opcode)
IN BCFILE *f;
INOUT ExprPtr gotop;
IN OPCODE opcode;
#endif
{
  ExprPtr aux;

  aux = Kvarref(bcTell(f),f->stackDepth);
  SETCAR(gotop,aux);
#ifndef INSERT_MERGE_NODES
  if (opcode!=BC_MERGE)
#endif
  bcPut(opcode,f);
}

PRIVATE ExprPtr beginBackward(f)
IN BCFILE *f;
{
  ExprPtr gotop;

  gotop = beginForward(f);
  compileDebugZone(true,f);
  compileCodePoint(f);
  beginBackward0(f,gotop,BC_LOOP);
  return gotop;
}

#ifdef OBSOLETE
PRIVATE void endForwardBeginBackward(f,gotop)
IN BCFILE *f;
INOUT ExprPtr gotop;
{
  endForward0(f,gotop);
  if (CDR(gotop)!=trueValue) /* if code was not dead when beginForward() was called, */
    { /* then we estimate that this code has some chance of being called later, by
	 a jump backward, and we compile a MERGE node. */
      f->dead = 0;
      beginBackward0(f,gotop,BC_MERGE);
    }
}
#endif

#define endBackward(f,gotop)

#if DEBUGGING!=0
PRIVATE INT beginDebug(f,st)
IN BCFILE *f;
OUT bcdState *st;
{
  INT pc;
  INT patch;

  if (!debugEnabled)
    return 0;
  bcdGetState(f,st);
  st->dz = currentDebugZone;
  if (f->dead)
    return -1;
  pc = bcTell(f);
  if ((INT)(OPCODE)pc == pc)
    {
      bcdPut(DBG_BEGIN,f);
      bcdPut((OPCODE)pc,f);
    }
  else
    {
      bcdPut(DBG_BEGIN_BIG,f);
      bcdPutInt(pc,f);
    }
  patch = bcdTell(f);
  bcdPut((OPCODE)-1,f); /* unknown offset */
  return patch;
}

PRIVATE void endDebug(patch,f,st)
IN BCFILE *f;
IN INT patch;
IN CONST bcdState *st;
{
  INT pc;

  if (!debugEnabled)
    return;
  currentDebugZone = st->dz;
  if (patch==-1)
    return;
  if (patch==bcdTell(f)-1) /* no debugging information recorded since beginDebug() */
    { /* then erase the BEGIN or BEGIN_BIG. */
      bcdRewindState(f,st);
    }
  else
    {
      bcdPatch(patch,f);
      pc = bcTell(f);
      if ((INT)(OPCODE)pc == pc)
	{
	  bcdPut(DBG_END,f);
	  bcdPut((OPCODE)pc,f);
	}
      else
	{
	  bcdPut(DBG_END_BIG,f);
	  bcdPutInt(pc,f);
	}
    }
}
#endif

PRIVATE void setStackDepth(f,level)
IN BCFILE *f;
IN COUNTER level;
{
  if (f->dead || level==f->stackDepth)
    f->stackDepth = level;
  else
    berror(message(615));
}

PRIVATE void compileRaise(f,env)
IN BCFILE *f;
IN CONST compEnv *env;
{
  ExprPtr handlers, h;
  COUNTER mlevel;

  compileDebugZone(true,f);
  handlers = CE_HANDLERS(env);
  if (handlers!=NIL)
    { /* compile a goto when the label for the handler is known */
      COUNTER savelevel;

      savelevel = f->stackDepth;
      handlers = CAR(handlers);
      mlevel = COUNTEDN(handlers);
      h = COUNTEDVAL(handlers);
      compileMapDrop(f,mlevel);
      compileSwDrop(f,f->stackDepth-COUNTEDN(h));
      compileGoto(f,BC_GOTO,COUNTEDVAL(h),0);
      f->stackDepth = savelevel;
    }
  else
    {
      compileCodePointOp(BC_RAISE,f);
    }
}

PRIVATE void compileRaiseImm(f,env,idx)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST INT idx;
{
  compileCGlobal(f,idx);
  compileRaise(f,env);
}

PRIVATE void compileBindings P((IN BCFILE *f,IN CONST compEnv *env,
				IN CONST ExprPtr bindings,INOUT ExprPtr errp));

PRIVATE void compileExprFun(f,env,stmt,errp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr stmt;
INOUT ExprPtr errp;
{
  compileExpr(f,env,stmt);
}

MAPCOLLECTFUNC(do_restricttolevel)
{
  INT level;

  level = MAPDATA(INT);
  if ((level==0 && x->kind!=(CHAR)E_CONS) ||
      (level!=0 && x->kind==(CHAR)E_CONS && IVAL(CDR(x))==level))
    return MapletKons(x,y);
  return NewMap();
}

PRIVATE MapPtr restrictToLevel(oldmap,newmap,level)
IN CONST MapPtr oldmap,newmap;
IN INT level;
{
  MapPtr map;
  ENTER;

  map = MapDomRestrBy(newmap,oldmap);
  map = MapCollect(map,do_restricttolevel,(char *)level);
  map = MapMerge(oldmap,map);
  QUIT;
  return map;
}

#if DEBUGGING
MAPSWEEPFUNC(do_resetdbgnames)
{
  compEnv *env;
  ExprPtr name;

  env = MAPDATA(compEnv *);
  name = LETBIND(y);
  y = MakePair ((ExprKind)y->kind, x, LETBODY(y));
  (void) KEEP (y);
  CE_PARENT_DBGNAMES(env) = MapPut (name, y, CE_PARENT_DBGNAMES(env));
  return false;
}
#endif

PRIVATE void resetDbgNames (env)
INOUT compEnv *env;
{
#if DEBUGGING
  if (debugEnabled)
    {
      ENTER;

      (void) MapSweep (CE_DBGNAMES(env), do_resetdbgnames, env);
      KEEPONLY(CE_PARENT_DBGNAMES(env));
    }
#endif
  CE_DBGNAMES(env) = NewMap();
}

PRIVATE void compileLet1(f,env,bindings,body,errp,localp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr bindings;
IN CONST ExprPtr body;
INOUT ExprPtr errp;
IN Boolean localp;
{
  COUNTER n,n0;
  bindCompFun comp;
  Boolean fake;
#if DEBUGGING!=0
  INT dstart;
  bcdState st;
#endif
  compEnv *newenv = (compEnv *)alloctmp(sizeof(compEnv));

  comp = localp?compileBindings:compileExprFun;
  *newenv = *env;
  CE_LETDELAY(newenv) = NewMap();
  resetDbgNames (newenv);
  newenv->_comp.next = theHeapFrames;
  theHeapFrames = (gcStackView *)&newenv->_comp;

  n = n0 = IVAL(CAR(bindings));
  if ((fake = (n<0))) /* then this is an imported module (an interface file, normally),
			 and the evaluator then first allocates a fake copy of the toplevel
			 environment GLOBENV below the current one. */
    {
      newenv->level++;
      newenv->globenvs |= (1L << newenv->level);
      n = -1-n;
    }
  newenv->level++;
  if ((newenv->flags & CEF_NEEDENV) &&
      !(n0<0 ||
	needEnvBindings(CDR(bindings),newenv) ||
	(localp?needEnvBindings(body,newenv):
	 needEnv(body,newenv))))
    newenv->flags &= ~CEF_NEEDENV;
  /* If n0<0, we do not disable the NEEDENV feature, the reason is the following:
     if n0<0, then this is not just a let or a local, but in fact an imported module.
     Such modules may be huge, and the only reasonable way to import them is with
     CEF_NEEDENV | CEF_SPARESTACK: this way, every binding is compiled as:
     <compute value> <get components of value as specified by the pattern>
     <store components in heap-allocated environment> <drop temporary data on stack>.
     Otherwise, every list of bindings is compiled as:
     <compute value> <get components>, possibly followed by <store components>
     (if CEF_NEEDENV is set), and leaving all components on the stack, which
     would be troublesome in the case of modules since the stack has a fixed
     size (BC_STACK_SIZE=512 by default).
     This can still be troublesome in case of big let- or local-expressions,
     or when importing a module inside a let, but I hope this won't happen.
     */
  if (newenv->flags & CEF_NEEDENV)
    {
      compileCurEnv(f,env); /* --- env (if needenv); --- (if !needenv) */
      /* notice that we compileCurEnv() on env, not newenv: we want to get the
	 old environment. */
      if (fake)
	bcPut(BC_FAKEENV,f);
      bcPut(BC_CONSENV,f);
      bcPut((OPCODE)n,f);
      newCurEnv(newenv,f->stackDepth-1);
    }
  /* --- env (if needenv); --- (if !needenv) */
#if DEBUGGING!=0
  dstart = beginDebug(f,&st);
#endif
  newenv->ilevel = newenv->level;
  compileBindings(f,newenv,CDR(bindings),errp);
  newenv->ilevel = env->ilevel;
  newenv->flags = env->flags;
  (*comp)(f,newenv,body,errp); /* --- <...> res */
  if (localp)
    {
      CE_VARS(env) = restrictToLevel(CE_VARS(env),CE_VARS(newenv),env->ilevel);
      CE_INLINES(env) = restrictToLevel(CE_INLINES(env),CE_INLINES(newenv),env->ilevel);
    }
#if DEBUGGING!=0
  endDebug(dstart,f,&st);
#endif

  theHeapFrames = newenv->_comp.next;
}

MAPSWEEPFUNC(do_dumpletstores)
{
  struct dumpletstores *data;

  data = MAPDATA(struct dumpletstores *);
  bcPut((OPCODE)(IVAL(y)-data->f->stackDepth),data->f);
  bcPut((OPCODE)(IVAL(x)),data->f);
  return false;
}

PRIVATE void dumpLetStores(f,envlevel,set)
IN BCFILE *f;
IN INT envlevel;
IN CONST MapPtr set;
{
  struct dumpletstores *data;

  if (set==NewMap())
    return;
  data = (struct dumpletstores *)alloctmp(sizeof(struct dumpletstores));
  if (set->kind==(CHAR)E_MAPLET)
    {
      bcPut(BC_LET1,f);
      bcPut((OPCODE)(envlevel-f->stackDepth),f);
      data->f = f;
      data->envlevel = envlevel;
      (void) do_dumpletstores(MAPLETX(set),MAPLETY(set),(CHAR *)data);
    }
  else
    {
      bcPut(BC_LET,f);
      bcPut((OPCODE)(envlevel-f->stackDepth),f);
      data->f = f;
      data->envlevel = envlevel;
      (void) MapSweep(set,do_dumpletstores,data);
      bcPut((OPCODE)0,f);
    }
}

#if DEBUGGING!=0
MAPSWEEPFUNC(do_flushdbgnames)
{
  struct flushdbgnames *data;
  COUNTER level,rank;

  data = MAPDATA(struct flushdbgnames *);
  if (x->kind==E_INT)
    {
      rank = IVAL(x);
      level = 0;
    }
  else
    {
      rank = IVAL(CAR(x));
      level = IVAL(CDR(x));
    }
  if (debugVarref(data->f,data->env,level,rank))
    {
      INT bcupos,n,i;
      CHAR *namebuf;
      ExprPtr name;
      OPCODE op;

      name = LETBIND(y);
      n = SLEN(name);
      namebuf = (CHAR *)alloctmp(n+1);
      safememcpy(namebuf,STEXT(name),n);
      namebuf[n] = 0;
      bcupos = bcuTell(data->f);
      bcuPut(namebuf,n+1,1,data->f);
      switch (y->kind)
	{
	case E_APPL: op = DBG_DEFCON; break;
	case E_GOTO: op = DBG_DEFEXC; break;
	default: /* case E_LET: */ op = DBG_DEFNAME; break;
	}
      bcdPut(op,data->f);
      bcdPut((OPCODE)bcupos,data->f);
      i = getConstant(data->f,LETBODY(y));
      bcdPut((OPCODE)i,data->f);
    }
  return false;
}

PRIVATE void flushDbgNames(f,env)
IN BCFILE *f;
IN CONST compEnv *env;
{
  struct flushdbgnames *data;
  MapPtr dbgnames;

  if (!debugEnabled)
    return;
  dbgnames = CE_DBGNAMES(env);
  if (dbgnames==NewMap())
    return;
  data = (struct flushdbgnames *)alloctmp(sizeof(struct flushdbgnames));
  data->f = f;
  data->env = env;
  (void) MapSweep(dbgnames,do_flushdbgnames,data);
  resetDbgNames (env);
}
#else
#define flushDbgNames(f,env)
#endif

PRIVATE ExprPtr compilePattern P((IN BCFILE *f,IN CONST compEnv *env,IN CONST ExprPtr pat,
				  INOUT ExprPtr errp,IN CONST ExprPtr choicelist,
				  IN Boolean notagcheck));

PRIVATE void addLetDelay(env,rank,index)
INOUT compEnv *env;
IN INT rank, index;
{
  ExprPtr r,i,i_l;
#ifdef DEBUG
  ExprPtr ii;
  Boolean found;
#endif
  ENTER;

  r = KINT(rank);
  i = KINT(index);
#ifdef DEBUG
  ii = MapGet(r,CE_LETDELAY(env),&found);
  if (found && ii!=i)
    berror(message(610));
#endif
  setMapPut(r,i,&CE_LETDELAY(env));
  i_l = r;
  if (env->ilevel!=0)
    i_l = KONS(i_l,KINT(env->ilevel));
  setMapPut(i_l,i,&CE_VARS(env));
  QUIT;
}

#if DEBUGGING!=0
PRIVATE void addDbgName(env,rank,data)
INOUT compEnv *env;
IN INT rank;
IN CONST ExprPtr data;
{
  ExprPtr i_l;
  ENTER;

  if (debugEnabled)
    {
      i_l = KINT(rank);
      if (env->ilevel!=0)
	i_l = KONS(i_l,KINT(env->ilevel));
      setMapPut(i_l,data,&CE_DBGNAMES(env));
    }
  QUIT;
}
#else
#define addDbgName(env,rank,data)
#endif

PRIVATE ExprPtr compilePatternAt(f,env,pat,level,errp,choicelist,notagcheck)
IN BCFILE *f;
INOUT compEnv *env;
IN CONST ExprPtr pat;
IN COUNTER level;
INOUT ExprPtr errp;
IN CONST ExprPtr choicelist;
IN Boolean notagcheck;
{
 again:
  switch (pat->kind)
    {
    case E_DEBUGNAME:
      addDbgName(env,COUNTEDN(DNINFO(pat)),COUNTEDVAL(DNINFO(pat)));
      pat = DNPAT(pat);
      goto again;
    case E_VARINTRO:
      addLetDelay(env,VARRANK(pat),level);
      return choicelist;
    case E_PAT_ANY:
    case E_TUPLE_0:
    case E_RECORD_0:
      return choicelist; /* because of type-checking, tags always match. */
    default:
      if (level!=f->stackDepth-1)
	compileFetch(f,level);
      return compilePattern(f,env,pat,errp,choicelist,notagcheck);
    }
}

PRIVATE ExprPtr compileVectorMatch P((IN BCFILE *f,IN CONST compEnv *env,
				      IN CONST ExprPtr pat,IN COUNTER idx,
				      IN COUNTER level,INOUT ExprPtr errp,
				      IN CONST ExprPtr choicelist));

DEF_PRIVATE_ACTION_1ARG(compilevectormatch)
{
  struct compilevectormatch *data;

  data = (struct compilevectormatch *)ENV_UNDERLYING(env);
  return compileVectorMatch(data->f,data->env,data->pat,data->idx,data->level,
			    data->errp,data->choicelist);
}
ACTION_END

PRIVATE ExprPtr compileVectorMatch1(f,env,pat,idx,level,errp,choicelist)
IN BCFILE *f;
INOUT compEnv *env;
IN CONST ExprPtr pat;
IN COUNTER idx;
IN COUNTER level;
INOUT ExprPtr errp;
IN CONST ExprPtr choicelist;
{
  struct compilevectormatch *data = (struct compilevectormatch *)
    alloctmp(sizeof(struct compilevectormatch));

  data->f = f;
  data->env = env;
  data->pat = pat;
  data->idx = idx;
  data->level = level;
  data->errp = errp;
  data->choicelist = choicelist;
  return CallCC(ACTION_NAME(compilevectormatch),(ExprPtr)data,true);
}

#define CHOICE_MATCH 0
#define VECTOR_MATCH 1
#define RECORDMAPADD_MATCH 2
#define SWITCH_MATCH 3
#define NFOLD_MATCH 4

PRIVATE ExprPtr newChoice(f,env,errp,patchoice,choicelist,idx,level,matchkind)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr errp;
IN CONST ExprPtr patchoice;
IN CONST ExprPtr choicelist;
IN COUNTER idx,level;
IN COUNTER matchkind;
{
  ExprPtr data[13],e,newerrp;
  ENTER;

  newerrp = beginForward(f);
  data[0] = KINT(f->stackDepth);
#define CH_STACKLEVEL(e) IVAL(AVAL(e)->expr[0])
  data[1] = KINT(level);
#define CH_DATALEVEL(e) IVAL(AVAL(e)->expr[1])
  data[2] = patchoice;
#define CH_PATCHOICE(e) AVAL(e)->expr[2]
  data[3] = newerrp;
#define CH_ERRP(e) AVAL(e)->expr[3]
  data[4] = choicelist;
#define CH_NEXT(e) AVAL(e)->expr[4]
  data[5] = CE_LETDELAY(env);
#define CH_LETDELAY(e) AVAL(e)->expr[5]
  data[6] = KINT(idx);
#define CH_IDX(e) IVAL(AVAL(e)->expr[6])
  data[7] = KINT(matchkind);
#define CH_MATCHKIND(e) IVAL(AVAL(e)->expr[7])
  data[8] = CE_VARS(env);
#define CH_VARS(e) AVAL(e)->expr[8]
  data[9] = CE_INLINES(env);
#define CH_INLINES(e) AVAL(e)->expr[9]
  data[10] = CE_DBGNAMES(env);
#define CH_DBGNAMES(e) AVAL(e)->expr[10]
  data[11] = CE_PARENT_DBGNAMES(env);
#define CH_PARENT_DBGNAMES(e) AVAL(e)->expr[11]
  data[12] = errp;
#define CH_SWITCH_ERRP(e) AVAL(e)->expr[12]
  e = MakeRecord(E_TUPLE_GNRL,NIL,13,data);
  KEEPONLY(e);
  return e;
}

PRIVATE Boolean isTruePat(pat)
IN CONST ExprPtr pat;
{
again:
  switch (pat->kind)
    {
    case E_DEBUGNAME:
      pat = DNPAT(pat);
      goto again;
#ifdef DATACONPAT
    case E_DATACON:
      pat = COUNTEDVAL(pat);
      goto again;
#endif
    case E_DATA:
      return pat==trueValue;
    default:
      return false;
    }
}

PRIVATE Boolean isFalsePat(pat)
IN CONST ExprPtr pat;
{
again:
  switch (pat->kind)
    {
    case E_DEBUGNAME:
      pat = DNPAT(pat);
      goto again;
#ifdef DATACONPAT
    case E_DATACON:
      pat = COUNTEDVAL(pat);
      goto again;
#endif
    case E_DATA:
      return pat==trueValue;
    default:
      return false;
    }
}

PRIVATE Boolean nontrivMatch(pat)
IN CONST ExprPtr pat;
{
  switch (pat->kind)
    {
    case E_PAT_ANY:
    case E_TUPLE_0: /* because of type-checking, () and |[]| patterns always match */
    case E_RECORD_0:
      return false;
    }
  return true;
}

PRIVATE ExprPtr compileVectorMatch(f,env,pat,idx,level,errp,choicelist)
IN BCFILE *f;
INOUT compEnv *env;
IN CONST ExprPtr pat;
IN COUNTER idx;
IN COUNTER level;
INOUT ExprPtr errp;
IN CONST ExprPtr choicelist;
{
  if (STKFULL(pat))
    {
      return compileVectorMatch1(f,env,pat,idx,level,errp,choicelist);
    }
again:
  switch (pat->kind)
    {
    case E_PAT_TUPLE_GNRL_CONS:
    case E_PAT_RECORD_GNRL_CONS:
      if (nontrivMatch(CAR(pat)))
	{
	  compileFetch(f,level);
	  compileNth(f,idx);
	  choicelist = compilePattern(f,env,CAR(pat),errp,choicelist,false);
	}
      idx++;
      pat = CDR(pat);
      goto again;
    case E_PAT_REVTUPLE_GNRL_CONS:
    case E_PAT_REVRECORD_GNRL_CONS:
      choicelist = compileVectorMatch(f,env,CDR(pat),idx+1,level,errp,choicelist);
      if (nontrivMatch(CAR(pat)))
	{
	  compileFetch(f,level);
	  compileNth(f,idx);
	  choicelist = compilePattern(f,env,CAR(pat),errp,choicelist,false);
	}
      break;
    case E_PATCHOICE:
      {
	ENTER;

	choicelist = newChoice(f,env,errp,CDR(pat),choicelist,idx,level,VECTOR_MATCH);
	choicelist = compileVectorMatch(f,env,CAR(pat),idx,level,
					CH_ERRP(choicelist),choicelist);
	KEEPONLY(choicelist);
      }
      break;
    case E_PAT_ANY: /* ellipsis */
      break;
    case E_TUPLE_0:
    case E_RECORD_0: /* end of vector */
      break;
    default:
      kinderror(message(611),pat->kind);
      /*NOTREACHED*/
      break;
    }
  return choicelist;
}

PRIVATE ExprPtr compileNfoldMatch(f,env,pat,idx,level,errp,choicelist)
/* Works as for compileVectorMatch, except the vector is not
   an E_TUPLE_GNRL (or similar) at position level on the stack,
   rather its elements are all on the stack at positions level, level+1, ..., level+n-1.
   (n is not given.)
*/
IN BCFILE *f;
INOUT compEnv *env;
IN CONST ExprPtr pat;
IN COUNTER idx;
IN COUNTER level;
INOUT ExprPtr errp;
IN CONST ExprPtr choicelist;
{
again:
  switch (pat->kind)
    {
    case E_TUPLE_2:
    case E_UNSHARED_TUPLE_2:
      if (nontrivMatch(CAR(pat)))
	{
	  choicelist = compilePatternAt(f,env,CAR(pat),level+idx,errp,choicelist,false);
	}
      if (nontrivMatch(CDR(pat)))
	{
	  choicelist = compilePatternAt(f,env,CDR(pat),level+idx+1,errp,choicelist,false);
	}
      break;
    case E_PAT_REVTUPLE_2:
      if (nontrivMatch(CDR(pat)))
	{
	  choicelist = compilePatternAt(f,env,CDR(pat),level+idx+1,errp,choicelist,false);
	}
      if (nontrivMatch(CAR(pat)))
	{
	  choicelist = compilePatternAt(f,env,CAR(pat),level+idx,errp,choicelist,false);
	}
      break;
    case E_PAT_TUPLE_GNRL_CONS:
    case E_PAT_RECORD_GNRL_CONS:
      if (nontrivMatch(CAR(pat)))
	{
	  choicelist = compilePatternAt(f,env,CAR(pat),level+idx,errp,choicelist,false);
	}
      idx++;
      pat = CDR(pat);
      goto again;
    case E_PAT_REVTUPLE_GNRL_CONS:
    case E_PAT_REVRECORD_GNRL_CONS:
      choicelist = compileNfoldMatch(f,env,CDR(pat),idx+1,level,errp,choicelist);
      if (nontrivMatch(CAR(pat)))
	{
	  choicelist = compilePatternAt(f,env,CAR(pat),level+idx,errp,choicelist,false);
	}
      break;
    case E_PATCHOICE:
      {
	ENTER;

	choicelist = newChoice(f,env,errp,CDR(pat),choicelist,idx,level,NFOLD_MATCH);
	choicelist = compileNfoldMatch(f,env,CAR(pat),idx,level,
					CH_ERRP(choicelist),choicelist);
	KEEPONLY(choicelist);
      }
      break;
    case E_PAT_ANY: /* ellipsis */
      break;
    case E_TUPLE_0:
    case E_RECORD_0: /* end of vector */
      break;
    default:
      kinderror(message(611),pat->kind);
      /*NOTREACHED*/
      break;
    }
  return choicelist;
}

PRIVATE ExprPtr compileTupleMatch(f,env,pat,errp,choicelist)
IN BCFILE *f;
INOUT compEnv *env;
IN CONST ExprPtr pat;
INOUT ExprPtr errp;
IN CONST ExprPtr choicelist;
{
  ExprPtr t2pp,finp,aux;
  COUNTER level;
  /* pat is a list of E_PAT_TUPLE_GNRL_CONS or E_PAT_REVTUPLE_GNRL_CONS.
     It is terminated by E_TUPLE_0 (fixed size tuples), or by
     E_PAT_ANY (ellipsis).
  */
  /* value may be a tuple_gnrl, or a tuple_2; in the last case,
     CDR(pat) must be an ellipsis or a one-element list terminated
     by an ellipsis or tuple0. */

  level = f->stackDepth-1;
  finp = beginForward(f);
  aux = CDR(pat);
  if (aux==patany || ((aux->kind==(CHAR)E_PAT_TUPLE_GNRL_CONS ||
		       aux->kind==(CHAR)E_PAT_REVTUPLE_GNRL_CONS) &&
		      CDR(aux)==patany))
    { /* These are the only cases where value may be a tuple2 */
      t2pp = beginForward(f);
      compileGoto(f,BC_ISTUPLE2,t2pp,0);
      /* if val->kind==(CHAR)E_TUPLE_2: */
      if (aux==patany) /* if pattern has an ellipsis */
	{
	  if (nontrivMatch(CAR(pat)))
	    {
	      compileFetchCar(f,level); /* compute CAR(val) */
	      /* then match CAR(val) */
	      choicelist = compilePattern(f,env,CAR(pat),errp,choicelist,false);
	    }
	}
      else
	if (pat->kind==(CHAR)E_PAT_TUPLE_GNRL_CONS)
	  {
	    aux = CAR(pat);
	    if (nontrivMatch(aux))
	      {
		compileFetchCar(f,level); /* compute CAR(val) */
		/* then match CAR(val) */
		choicelist = compilePattern(f,env,aux,errp,choicelist,false);
	      }
	    aux = CAR(CDR(pat));
	    if (nontrivMatch(aux))
	      {
		compileFetchCdr(f,level); /* then match CDR(val) */
		choicelist = compilePattern(f,env,aux,errp,choicelist,false);
	      }
	  }
	else
	  {
	    aux = CAR(CDR(pat));
	    if (nontrivMatch(aux))
	      {
		compileFetchCdr(f,level); /* compute CDR(val) */
		choicelist = compilePattern(f,env,aux,errp,choicelist,false); /* then match CDR(val) */
	      }
	    aux = CAR(pat);
	    if (nontrivMatch(aux))
	      {
		compileFetchCar(f,level); /* then match CAR(val) */
		choicelist = compilePattern(f,env,aux,errp,choicelist,false);
	      }
	  }
      compileGoto(f,BC_GOTO,finp,0);
      setStackDepth(f,level+1);
      endForward(f,t2pp);
    }
  /* now we produce the code for the case when val is a E_TUPLE_GNRL;
     note that for the information in CE_LETDELAY(env) and CE_VARS(env)
     to be valid at all times, it is necessary that in any case
     (E_TUPLE_2 or E_TUPLE_GNRL) the local variables get bound to
     values on the stack *at the same level* in both branches of the
     alternative.
     This can be done by being sure that whatever the alternative, the
     pattern is swept by the various compile* functions in the same order.
     */
  /* if val->kind!=(CHAR)E_TUPLE_2: */
  choicelist = compileVectorMatch(f,env,pat,0,level,errp,choicelist);
  if (isForward(finp))
    {
      endForward(f,finp);
    }
  return choicelist;
}

PRIVATE ExprPtr compileRecordMatch(f,env,pat,errp,choicelist)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr pat;
INOUT ExprPtr errp;
IN CONST ExprPtr choicelist;
{
  /* pat is a list of E_PAT_RECORD_GNRL_CONS or E_PAT_REVRECORD_GNRL_CONS.
     It is terminated by E_RECORD_0 (fixed size records), or by
     E_PAT_ANY (ellipsis).
  */
  ExprPtr r1pp,finp,aux;
  COUNTER level;

  level = f->stackDepth-1;
  finp = beginForward(f);
  aux = CDR(pat);
  /* val may be a record_gnrl, or a record_1; in the last case,
     CDR(pat) must be an ellipsis. */
  if (aux==patany)
    {
      r1pp = beginForward(f);
      compileGoto(f,BC_ISRECORD1,r1pp,0);
      /* if val->kind==(CHAR)E_RECORD_1: */
      aux = CAR(pat);
      if (nontrivMatch(aux))
	{
	  compileFetchCdr(f,level);
	  choicelist = compilePattern(f,env,aux,errp,choicelist,false);
	}
      compileGoto(f,BC_GOTO,finp,0);
      setStackDepth(f,level+1);
      endForward(f,r1pp);
    }
  /* if val->kind!=(CHAR)E_TUPLE_2: */
  choicelist = compileVectorMatch(f,env,pat,0,level,errp,choicelist);
  if (isForward(finp))
    {
      endForward(f,finp);
    }
  return choicelist;
}

PRIVATE void compileCheckTag(f,patbits,errp)
IN BCFILE *f;
IN CONST ExprPtr patbits;
INOUT ExprPtr errp;
{
  ExprPtr gotop;
  COUNTER i,level;
  ENTER;

  gotop = beginForward(f);
  level = f->stackDepth;
  for (i=1; i<E_UNUSED; i++)
    {
      if (PATCHECKBITS(patbits)[i>>3] & (1 << (i & 7)))
	{ /* there must be at least one bit here */
	  compileGoto(f,(OPCODE)(BC_HASTAG+i),gotop,0);
	}
    }
  compileGoto(f,BC_GOTO,errp,0);
  setStackDepth(f,level);
  endForward(f,gotop);
  QUIT;
}

PRIVATE void compileCheckCardEq(f,n,errp)
IN BCFILE *f;
IN COUNTER n;
IN CONST ExprPtr errp;
{
  switch (n)
    {
    case 0:
      compileGoto(f,BC_ISEMPTY,errp,0);
      break;
    case 1:
      compileGoto(f,BC_ISMAPLET,errp,0);
      break;
    default:
      compileGoto(f,BC_CARDEQ,errp,0);
      bcPut((OPCODE)n,f);
      break;
    }
}

PRIVATE void compileCheckCardGeq(f,n,errp)
IN BCFILE *f;
IN COUNTER n;
IN CONST ExprPtr errp;
{
  switch (n)
    {
    case 0:
      break;
    case 1:
      compileGoto(f,BC_ISNEMAP,errp,0);
      break;
    case 2:
      compileGoto(f,BC_ISMAP,errp,0);
      break;
    default:
      compileGoto(f,BC_CARDGEQ,errp,0);
      bcPut((OPCODE)n,f);
      break;
    }
}

PRIVATE void compileSelectGnrl(f,env,label)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr label;
{
  if (label->kind==(CHAR)E_INT)
    { /* if label is precomputed */
      compileNth(f,IVAL(label)-1);
    }
  else
    {
      bcPut(BC_SELECT,f);
      bcPut((OPCODE)getConstant(f,label),f);
    }
}

PRIVATE ExprPtr compileRecordGnrlMapAddMatch P((IN BCFILE *f,
						IN CONST compEnv *env,
						IN CONST ExprPtr pat,
						IN COUNTER level,
						INOUT ExprPtr errp,
						IN CONST ExprPtr choicelist));

DEF_PRIVATE_ACTION_1ARG(compilerecordgnrlmapaddmatch)
{
  struct compilerecmapaddpattern *data;

  data = (struct compilerecmapaddpattern *)ENV_UNDERLYING(env);
  return compileRecordGnrlMapAddMatch(data->f,data->env,data->pat,data->level,
				      data->errp,data->choicelist);
}
ACTION_END

PRIVATE ExprPtr compileRecordGnrlMapAddMatch1(f,env,pat,level,errp,choicelist)
IN BCFILE *f;
INOUT compEnv *env;
IN CONST ExprPtr pat;
IN COUNTER level;
INOUT ExprPtr errp;
IN CONST ExprPtr choicelist;
{
  struct compilerecmapaddpattern *data =
    (struct compilerecmapaddpattern *)alloctmp(sizeof(struct compilerecmapaddpattern));

  data->f = f;
  data->env = env;
  data->pat = pat;
  data->level = level;
  data->errp = errp;
  data->choicelist = choicelist;
  return CallCC(ACTION_NAME(compilerecordgnrlmapaddmatch),(ExprPtr)data,true);
}

PRIVATE ExprPtr compileRecordGnrlMapAddMatch(f,env,pat,level,errp,choicelist)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr pat;
IN COUNTER level;
INOUT ExprPtr errp;
IN CONST ExprPtr choicelist;
{
  if (STKFULL(pat))
    return compileRecordGnrlMapAddMatch1(f,env,pat,level,errp,choicelist);
again:
  switch (pat->kind)
    {
    case E_PAT_RECORD_MAPADD:
      {
	ExprPtr aux;

	aux = CAR(pat);
	if (nontrivMatch(CDR(aux)))
	  {
	    compileFetch(f,level);
	    compileSelectGnrl(f,env,CAR(aux));
	    choicelist = compilePattern(f,env,CDR(aux),errp,choicelist,false);
	  }
      }
      pat = CDR(pat);
      goto again;
    case E_PATCHOICE:
      {
	ExprPtr aux;
	ENTER;

	choicelist = newChoice(f,env,errp,CDR(pat),choicelist,0,level,RECORDMAPADD_MATCH);
	aux = CAR(pat);
	choicelist = compileRecordGnrlMapAddMatch(f,env,aux,level,CH_ERRP(choicelist),
						  choicelist);
	KEEPONLY(choicelist);
	break;
      }
    default: /* should be an ellipsis */
      break;
    }
  return choicelist;
}

PRIVATE ExprPtr compileRecordMapAddMatch(f,env,pat,errp,choicelist)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr pat;
INOUT ExprPtr errp;
IN CONST ExprPtr choicelist;
{
  /* pat is a E_PAT_RECORD_MAPADD list, terminated by an ellipsis
     (coded as E_PAT_ANY), with possible E_PATCHOICEs inside.
     It cannot be terminated by an end-of-record, since in that
     case the compiler would be able to determine the type and
     assign indexes to all the fields, so that we would have an
     E_PAT_RECORD_GNRL_CONS or E_PAT_REVRECORD_GNRL_CONS pattern.
     This list contains couples (label,pattern) to be matched.
  */
  ExprPtr r1pp,finp,aux;
  COUNTER level;

  level = f->stackDepth-1;
  finp = beginForward(f);
  aux = CDR(pat);
  if (aux==patany) /* pattern |[a = <p>,...]|, may match a E_RECORD_1 */
    {
      r1pp = beginForward(f);
      compileGoto(f,BC_ISRECORD1,r1pp,0);
      aux = CDR(CAR(pat)); /* no need to compare labels, since type-checking ensures
			      they match */
      if (nontrivMatch(aux))
	{
	  compileFetchCdr(f,level);
	  choicelist = compilePattern(f,env,aux,errp,choicelist,false);
	}
      compileGoto(f,BC_GOTO,finp,0);
      setStackDepth(f,level+1);
      endForward(f,r1pp);
    }
  /* val may be a record_gnrl, or a record_1; in the last case,
     CDR(pat) must be an ellipsis, so now we assume val is a record_gnrl. */
  choicelist = compileRecordGnrlMapAddMatch(f,env,pat,level,errp,choicelist);
  if (isForward(finp))
    {
      endForward(f,finp);
    }
  return choicelist;
}

PRIVATE Boolean cstPattern(pat,val)
IN CONST ExprPtr pat;
OUT ExprPtr *val;
{
  switch (pat->kind)
    {
    case E_NIL: case E_TUPLE_0: case E_RECORD_0: case E_EMPTY:
    case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
      *val = pat;
      return true;
    case E_QUOTE:
      *val = REFVAL(pat);
      return true;
    default:
      return false;
    }
}

PRIVATE void compileDrop(f,n)
IN BCFILE *f;
IN COUNTER n;
{
  if (n!=0)
    {
      bcPut(BC_DROP,f);
      bcPut((OPCODE)n,f);
      bcPush(-n,f);
    }
}

PRIVATE void compilePushDrop(f,level,push)
IN BCFILE *f;
IN COUNTER level;
IN Boolean push;
{
  if (push)
    compileSwDrop(f,f->stackDepth-level);
  else
    compileDrop(f,f->stackDepth-level);
}

PRIVATE void compileMapDrop(f,maplevel)
IN BCFILE *f;
IN COUNTER maplevel;
{
  if (f->mapStackDepth!=maplevel)
    {
      bcPut(BC_MAPDROP,f);
      bcPut((OPCODE)(f->mapStackDepth-maplevel),f);
    }
}

PRIVATE INT getDataTag(pat,ttmp)
IN CONST ExprPtr pat;
OUT ExprPtr *ttmp;
{
  INT tag = MAX_SIGNED_LONG;

again:
  switch (pat->kind)
    {
#ifdef DATACONPAT
    case E_DATACON:
      pat = COUNTEDVAL(pat);
      goto again;
#endif
    case E_DATA:
      tag = (INT)pat->magic;
      *ttmp = CAR(pat);
      break;
    case E_PAT_AS:
      pat = PATAS2(pat);
      goto again;
    case E_QUOTE:
      pat = REFVAL(pat);
      goto again;
    }
  return tag;
}

PRIVATE MapPtr addDataTag(body,pat,tags,ttmp)
IN CONST ExprPtr body,pat;
IN CONST MapPtr tags;
OUT ExprPtr *ttmp;
{
  INT tag;
  ExprPtr tagi,match;
  Boolean found;

  tag = getDataTag(pat,ttmp);
  if (tag==MAX_SIGNED_LONG)
    return tags;
  tagi = KINT(tag);
  (void) MapGet(tagi,tags,&found);
  if (found)
    return NIL; /* cannot compile a switch statement */
  match = MakePair(E_CONS,pat,body);
  (void) KEEP(match);
  tags = MapPut (tagi,match,tags);
  return tags;
}

PRIVATE MapPtr collectDataTags(pat,body,ttmp,lastmatchp)
IN CONST ExprPtr pat,body;
OUT ExprPtr *ttmp;
OUT ExprPtr *lastmatchp;
{
  MapPtr tags;

  tags = NewMap();
again:
  switch (pat->kind)
    {
#ifdef DATACONPAT
    case E_DATACON:
      pat = COUNTEDVAL(pat);
      goto again;
#endif
    case E_PATCHOICE:
      tags = addDataTag(body,CAR(pat),tags,ttmp);
      if (tags==NIL)
	break;
      pat = CDR(pat);
      body = CDR(pat);
      pat = CAR(pat);
      goto again;
    default:
      {
	INT tag;

	tag = getDataTag(pat,ttmp);
	if (tag==MAX_SIGNED_LONG)
	  {
	    ExprPtr match;
	    match = MakePair(E_CONS,pat,body);
	    (void) KEEP(match);
	    *lastmatchp = match;
	  }
	else
	  {
	    *lastmatchp = NOEXPR;
	    tags = addDataTag(body,pat,tags,ttmp);
	  }
      }
      break;
    }
  return tags;
}

PRIVATE ExprPtr compileMapletPattern(f,env,pat,failp,choicelist)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr pat;
INOUT ExprPtr failp;
IN CONST ExprPtr choicelist;
{
  COUNTER level;

  level = f->stackDepth;
  if (pat->kind==(CHAR)E_CONS)
    {
      if (nontrivMatch(CAR(pat)))
	{
	  if (nontrivMatch(CDR(pat)))
	    compileFetchCar(f,level-1);
	  else
	    bcPut(BC_CAR,f);
	  choicelist = compilePattern(f,env,CAR(pat),failp,choicelist,false);
	}
      if (nontrivMatch(CDR(pat)))
	{
	  if (level!=f->stackDepth)
	    compileFetchCdr(f,level-1);
	  else
	    bcPut(BC_CDR,f);
	  choicelist = compilePattern(f,env,CDR(pat),failp,choicelist,false);
	}
    }
  else
    {
      if (nontrivMatch(CDR(pat)))
	{
	  if (nontrivMatch(CAR(pat)))
	    compileFetchCdr(f,level-1);
	  else
	    bcPut(BC_CDR,f);
	  choicelist = compilePattern(f,env,CDR(pat),failp,choicelist,false);
	}
      if (nontrivMatch(CAR(pat)))
	{
	  if (level!=f->stackDepth)
	    compileFetchCar(f,level-1);
	  else
	    bcPut(BC_CAR,f);
	  choicelist = compilePattern(f,env,CAR(pat),failp,choicelist,false);
	}
    }
  return choicelist;
}

PRIVATE ExprPtr compileMapAddPatternGnrl1 P((IN BCFILE *f,IN CONST compEnv *env,
					     IN CONST ExprPtr pat,INOUT ExprPtr nomatchp,
					     IN CONST ExprPtr choicelist,
					     INOUT ExprPtr emptyp));

PRIVATE ExprPtr compileMapAddPatternGnrl(f,env,pat,nomatchp,choicelist)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr pat;
INOUT ExprPtr nomatchp;
IN CONST ExprPtr choicelist;
{
  choicelist = compileMapAddPatternGnrl1(f,env,pat,nomatchp,choicelist,nomatchp);
  return choicelist;
}

PRIVATE ExprPtr compilePatSwitch(f,env,pat,errpp,choicelist,defpatp)
IN BCFILE *f;
INOUT compEnv *env;
IN CONST ExprPtr pat;
INOUT ExprPtr *errpp;
IN CONST ExprPtr choicelist;
OUT ExprPtr *defpatp;
{
  INT i,m,n;
  MapPtr tags;
  ExprPtr ttm;
  Boolean found;
  ExprPtr tagi,p,aux,g,gotop,mainp,defaultp,defp,lastmatch,errp;
  ENTER;

  errp = *errpp;
  *defpatp = NOEXPR;
  tags = collectDataTags(pat,NIL,&ttm,&lastmatch);
  if (tags->kind==(CHAR)E_MAP) /* compile switch only if at least two distinct
					   datatype kinds to match */
    {
      bcPut(BC_SWITCH,f);
      m = -ttm->magic;
      bcPut((OPCODE)(m+2), f);
      n = AVAL(CAR(ttm))->n;
      bcPut((OPCODE)(n-m-1), f);
      mainp = beginForward(f);
      defaultp = beginForward(f);
      if (lastmatch!=NOEXPR)
	{
	  choicelist = newChoice(f,env,errp,lastmatch,choicelist,0,f->stackDepth-1,
				 CHOICE_MATCH);
	  defp = *errpp = CH_ERRP(choicelist);
	}
      else defp = errp;
      for (i=0; i<n; i++)
	{
	  tagi = KINT(i-m);
	  p = MapGet(tagi,tags,&found);
	  if (found)
	    {
	      if (CDR(p)==NIL) /* This is the main pattern */
		{
		  *defpatp = CAR(p);
		  gotop = mainp;
		}
	      else
		{
		  choicelist = newChoice(f,env,defp,p,
					 choicelist,0,f->stackDepth-1,SWITCH_MATCH);
		  gotop = CH_ERRP(choicelist);
		}
	    }
	  else /* default case */
	    gotop = defaultp;
	  /* the following copied from the forward jump case in compileGoto(): */
	  if (!f->dead)
	    {
	      aux = Kvarref(bcTell(f),f->stackDepth);
	      (void) KEEP(aux);
	      g = CAR(gotop);
	      g = Ktuple2(aux,g); /* signal that this is not a run-of-the-mill goto
				     instruction, but a switch table entry. */
	      SETCAR(gotop,g);
	      bcPut((OPCODE)-1,f); /* dummy, to be pached later on. */
	    }
	}
      f->dead = 1;
      endForward(f,defaultp);
      compileGoto(f,BC_GOTO,defp,0); /* instead of filling unmatched entries of the table
					with a branch to defp directly, we store a branch
					to defaultp, which does a goto to defp.  This way,
					we can handle decreases in stackDepth, which the
					switch tables cannot, directly. */
      endForward(f,mainp);
    }
  KEEPONLY(choicelist);
  return choicelist;
}

DEF_PRIVATE_ACTION_1ARG(compilepattern)
{
  struct compilepattern *data;

  data = (struct compilepattern *)ENV_UNDERLYING(env);
  return compilePattern(data->f,data->env,data->pat,data->errp,data->choicelist,
			data->notagcheck);
}
ACTION_END

PRIVATE ExprPtr compilePattern1(f,env,pat,errp,choicelist,notagcheck)
IN BCFILE *f;
INOUT compEnv *env;
IN CONST ExprPtr pat;
INOUT ExprPtr errp;
IN CONST ExprPtr choicelist;
IN Boolean notagcheck;
{
  struct compilepattern *data = (struct compilepattern *)alloctmp(sizeof(struct compilepattern));

  data->f = f;
  data->env = env;
  data->pat = pat;
  data->errp = errp;
  data->choicelist = choicelist;
  data->notagcheck = notagcheck;
  return CallCC(ACTION_NAME(compilepattern),(ExprPtr)data,true);
}

PRIVATE ExprPtr compilePattern(f,env,pat,errp,choicelist,notagcheck)
IN BCFILE *f;
INOUT compEnv *env;
IN CONST ExprPtr pat;
INOUT ExprPtr errp;
IN CONST ExprPtr choicelist;
IN Boolean notagcheck;
{
  ENTER;

  if (STKFULL(pat))
    {
      return compilePattern1(f,env,pat,errp,choicelist,notagcheck);
    }
  if (!notagcheck)
    {
      ExprPtr defpat;

      choicelist = compilePatSwitch(f,env,pat,&errp,choicelist,&defpat);
      if (defpat!=NOEXPR)
	{
	  pat = defpat;
	  notagcheck = true;
	}
    }
again:
  switch (pat->kind)
    {
    case E_DEBUGNAME:
      addDbgName(env,COUNTEDN(DNINFO(pat)),COUNTEDVAL(DNINFO(pat)));
      pat = DNPAT(pat);
      goto again;
    case E_VARINTRO:
      addLetDelay(env,VARRANK(pat),f->stackDepth-1);
      break;
    case E_CONS:
    case E_UNSHARED_CONS:
      compileGoto(f,BC_ISCONS,errp,0); /* check that value is a CONS */
      /*fallthrough*/
    case E_MAPLET:
dec_cons:
      {
	COUNTER level;

	level = f->stackDepth-1;
	if (nontrivMatch(CAR(pat)))
	  {
	    compileFetchCar(f,level);
	    choicelist = compilePattern(f,env,CAR(pat),errp,choicelist,notagcheck);
	  }
	if (nontrivMatch(CDR(pat)))
	  {
	    compileFetchCdr(f,level);
	    pat = CDR(pat);
	  tcagain:
	    notagcheck = false;
	    goto again;
	  }
	break;
      }
    case E_PAT_REVCONS:
      compileGoto(f,BC_ISCONS,errp,0); /* check that value is a CONS */
dec_rev_cons:
      {
	COUNTER level;

	level = f->stackDepth-1;
	if (nontrivMatch(CDR(pat)))
	  {
	    compileFetchCdr(f,level);
	    choicelist = compilePattern(f,env,CDR(pat),errp,choicelist,notagcheck);
	  }
	if (nontrivMatch(CAR(pat)))
	  {
	    compileFetchCar(f,level);
	    pat = CAR(pat);
	    goto tcagain;
	  }
	break;
      }
#ifdef DATACONPAT
    case E_DATACON:
      pat = COUNTEDVAL(pat);
      goto tcagain;
#endif
    case E_DATA:
      {
	SHORT magic;

	magic = pat->magic;
	if (!notagcheck)
	  {
	    ExprPtr ttm;

	    ttm = CAR(pat);
	    if (AVAL(CAR(ttm))->n>1) /* don't test tag when there is only one possibility. */
	      {
		compileGoto(f,BC_ISDATA,errp,0);
		bcPut((OPCODE)magic,f);
	      }
	  }
	pat = CDR(pat);
	if (magic>=0 && nontrivMatch(pat))
	  {
	    compileFetchCdr(f,f->stackDepth-1);
	    goto tcagain;
	  }
	break;
      }
    case E_PAT_CONS:
      compileGoto(f,BC_ISCONS,errp,0); /* check that value is a CONS */
      if (nontrivMatch(REFVAL(pat)))
	{
	  COUNTER level;

	  level = f->stackDepth-1;
	  compileFetchCar(f,level);
	  compileFetchCdr(f,level);
	  bcPut(BC_KTUPLE2,f);
	  bcPush(-1,f);
	  pat = REFVAL(pat);
	  goto tcagain;
	}
      break;
    case E_PAT_ANY:
    case E_TUPLE_0:
    case E_RECORD_0:
      break; /* because of type-checking, tags always match. */
    case E_NIL: case E_EMPTY:
      compileGoto(f,BC_HASTAG+pat->kind,errp,0);
      break;
    case E_TUPLE_1:
    case E_UNSHARED_TUPLE_1:
      if (nontrivMatch(REFVAL(pat)))
	{
	  compileFetchRefVal(f,f->stackDepth-1);
	  pat = REFVAL(pat);
	  goto tcagain;
	}
      break;
    case E_TUPLE_2:
    case E_UNSHARED_TUPLE_2:
      goto dec_cons;
    case E_PAT_REVTUPLE_2:
      goto dec_rev_cons;
    case E_PAT_TUPLE_GNRL_CONS:
    case E_PAT_REVTUPLE_GNRL_CONS:
      choicelist = compileTupleMatch(f,env,pat,errp,choicelist);
      break;
    case E_RECORD_1:
    case E_UNSHARED_RECORD_1:
      pat = CDR(pat);
      if (nontrivMatch(pat))
	{
	  compileFetchCdr(f,f->stackDepth-1);
	  goto tcagain;
	}
      break;
    case E_PAT_RECORD_GNRL_CONS: /* records with known indexes */
    case E_PAT_REVRECORD_GNRL_CONS: /* records with known indexes */
      choicelist = compileRecordMatch(f,env,pat,errp,choicelist);
      break;
    case E_STRING:
    case E_INT: case E_REAL: case E_COMPLEX:
    case E_EXCEPTION:
      compileGoto(f,BC_IS,errp,0);
      bcPut((OPCODE)getConstant(f,pat),f);
      break;
    case E_REF:
      pat = REFVAL(pat);
      if (nontrivMatch(pat))
	{
	  compileFetchRefVal(f,f->stackDepth-1);
	  goto tcagain;
	}
      break;
    case E_QUOTE:
      compileGoto(f,BC_IS,errp,0);
      bcPut((OPCODE)getConstant(f,REFVAL(pat)),f);
      break;
    case E_VARREF: /* this is only used in patterns for exceptions */
      compileFetch(f,f->stackDepth-1);
      compileVarref(f,env,VARLEVEL(pat),VARRANK(pat));
      compileGoto(f,BC_ISEQUAL,errp,2);
      break;
    case E_APPL: /* this is only used for exceptions applied to arguments */
      compileGoto(f,BC_ISCONS,errp,0);
      /* now the name of the exception as recorded in the pattern
	 is given by a E_VARREF that points to an entry in the environment
	 where we find a compiled closure with the exception in the
	 QENV part. */
      {
	ExprPtr exc;

	exc = CAR(pat);
	compileFetchCar(f,f->stackDepth-1);
	compileVarref(f,env,VARLEVEL(exc),VARRANK(exc));
	bcPut(BC_QENV,f);
	compileGoto(f,BC_ISEQUAL,errp,2);
      }
      pat = CDR(pat);
      if (nontrivMatch(pat))
	{
	  compileFetchCdr(f,f->stackDepth-1);
	  goto tcagain;
	}
      break;
    case E_PAT_AS: /* PATAS1(pat) must now be a variable */
      choicelist = compilePattern(f,env,PATAS1(pat),errp,choicelist,notagcheck);
      /* now, stack depth should not have moved and val should still be on top of stack */
      pat = PATAS2(pat);
      goto again;
    case E_PAT_CHECK:
      compileCheckTag(f,pat,errp);
      pat = PATCHECKVAL(pat);
      goto tcagain;
    case E_PAT_PACK:
      compileFetchCdr(f,f->stackDepth-1); /* PACKTYPE is CDR */
      compileGoto(f,BC_TYPEMATCH,errp,1);
      bcPut((OPCODE)getConstant(f,PACKTYPE(pat)),f);
      /*dec_car:*/
      pat = CAR(pat); /* PACKVAL is CAR */
      if (nontrivMatch(pat))
	{
	  compileFetchCar(f,f->stackDepth-1); /* PACKVAL is CAR */
	  goto tcagain;
	}
      break;
    case E_PAT_CARD_EQ:
      compileCheckCardEq(f,COUNTEDN(pat),errp);
      pat = COUNTEDVAL(pat);
      goto tcagain;
    case E_PAT_CARD_GEQ:
      compileCheckCardGeq(f,COUNTEDN(pat),errp);
      pat = COUNTEDVAL(pat);
      goto tcagain;
    case E_PATCHOICE:
      choicelist = newChoice(f,env,errp,CDR(pat),choicelist,0,f->stackDepth-1,CHOICE_MATCH);
      errp = CH_ERRP(choicelist);
      pat = CAR(pat);
      goto again;
    case E_PAT_RECORD_MAPADD:
      choicelist = compileRecordMapAddMatch(f,env,pat,errp,choicelist);
      break;
    case E_PAT_MAPADD:
      choicelist = compileMapAddPatternGnrl(f,env,pat,errp,choicelist);
      break;
    default:
      kinderror(message(648),pat->kind);
      break;
    }
  KEEPONLY(choicelist);
  return choicelist;
}

PRIVATE void defineMe(env,pat)
IN CONST compEnv *env;
IN CONST ExprPtr pat;
{
  switch (pat->kind)
    {
    case E_DEBUGNAME:
      pat = DNPAT(pat);
      /*fallthrough*/
    case E_VARINTRO:
      env->me.level = env->ilevel;
      env->me.rank = VARRANK(pat);
      break;
    default:
      env->me.level = -1;
      env->me.rank = -1;
      break;
    }
}

PRIVATE void compileBinding P((IN BCFILE *f,IN CONST compEnv *env,
			       IN CONST ExprPtr binding, INOUT ExprPtr errp));

DEF_PRIVATE_ACTION_1ARG(compilebinding)
{
  struct compilebinding *data;

  data = (struct compilebinding *)ENV_UNDERLYING(env);
  (void) compileBinding(data->f,data->env,data->binding,data->errp);
  return NIL;
}
ACTION_END

PRIVATE void compileBinding1(f,env,binding,errp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr binding;
INOUT ExprPtr errp;
{
  struct compilebinding *data = (struct compilebinding *)alloctmp(sizeof(struct compilebinding));

  data->f = f;
  data->env = env;
  data->binding = binding;
  data->errp = errp;
  (void) CallCC(ACTION_NAME(compilebinding),(ExprPtr)data,true);
}

PRIVATE void compilePolyFun P((IN BCFILE *f,IN CONST compEnv *env,
			       IN CONST ExprPtr func,IN Boolean keepme));
PRIVATE void compileMemoFn P((IN BCFILE *f, IN CONST compEnv *env,
			      IN CONST ExprPtr func, IN CONST ExprPtr rem));

PRIVATE void compileBinding(f,env,binding,errp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr binding;
INOUT ExprPtr errp;
{
  ENTER;

  switch (binding->kind)
    {
    case E_LET:
      /* local bindings */
      if (STKFULL(binding))
	{
	  compileBinding1(f,env,binding,errp);
	  return;
	}
      compileLet1(f,env,LETBIND(binding),LETBODY(binding),errp,true);
      break;
    case E_DEBUGDECL:
      {
#if DEBUGGING!=0
	if (debugEnabled)
	  {
	    currentDebugZone = REFVAL(binding);
	    compileDebugZone(false,f);
	    compileCodePoint(f);
	  }
#endif
	break;
      }
    default:
      {
	struct Var saveme;
	ExprPtr pat,e,savevar = NOEXPR,saveinline = NOEXPR;
	COUNTER savelevel = 0;
	INT saveflags;
	Boolean spare;

	pat = CAR(binding);
	e = CDR(binding);
	if (isTruePat(pat))
	  {
	    compileTest(f,env,e,errp);
	  }
	else if (isFalsePat(pat))
	  {
	    ExprPtr rightp;
	    COUNTER level;

	    level = f->stackDepth;
	    rightp = beginForward(f);
	    compileTest(f,env,e,rightp);
	    compileGoto(f,BC_GOTO,errp,0);
	    setStackDepth(f,level);
	    if (isForward(rightp))
	      {
		endForward(f,rightp);
	      }
	  }
	else if (nontrivMatch(pat))
	  {
	    spare = (env->flags & (CEF_NEEDENV | CEF_SPARESTACK))==(CEF_NEEDENV | CEF_SPARESTACK);
	    if (spare)
	      {
		savevar = CE_VARS(env);
		(void) KEEP(savevar);
		saveinline = CE_INLINES(env);
		(void) KEEP(saveinline);
		savelevel = f->stackDepth;
	      }
	    saveme = env->me;
	    defineMe(env,pat);
	    saveflags = env->flags;
	    if ((env->flags & CEF_NEEDENV) && !needEnv(e,env))
	      env->flags &= ~CEF_NEEDENV;
	    switch (e->kind)
	      {
	      case E_PROFILE:
		{
		  ExprPtr saveprof;

		  saveprof = CE_PROF(env);
		  CE_PROF(env) = CDR(e);
		  e = CAR(e);
		  if (e->kind==(CHAR)E_POLYFUN
		      /*|| e->kind==(CHAR)E_N_LAMBDA*/
		      )
		    compilePolyFun(f,env,e,true);
		  /* evaluate function, but keep the definition of me, so as
		     to be able to detect tail recursions. */
		  else
		    compileExpr(f,env,e); /* evaluate value */
		  CE_PROF(env) = saveprof;
		}
	      break;
	      case E_POLYFUN:
		/*case E_N_LAMBDA:*/
		{
		  ExprPtr saveprof;

		  saveprof = CE_PROF(env);
		  CE_PROF(env) = defaultProf;
		  compilePolyFun(f,env,e,true);
		  /* evaluate function, but keep the definition of me, so as
		     to be able to detect tail recursions. */
		  CE_PROF(env) = saveprof;
		}
		break;
	      default:
		{
		  ExprPtr saveprof;

		  saveprof = CE_PROF(env);
		  CE_PROF(env) = defaultProf;
		  compileExpr(f,env,e); /* evaluate value */
		  CE_PROF(env) = saveprof;
		}
		break;
	      }
	    env->flags = saveflags;
	    env->me = saveme;
	    (void) compilePattern(f,env,pat,errp,NIL,false);
	    /* we can ignore the result choicelist, which should
	       simply be NIL itself, since we do not include choices in
	       Let bindings. */
	    if (env->flags & CEF_NEEDENV)
	      {
		dumpLetStores(f,getCurIEnv(env),CE_LETDELAY(env));
		if (spare)
		  {
		    CE_VARS(env) = savevar;
		    CE_INLINES(env) = saveinline;
		    compileDrop(f,f->stackDepth-savelevel);
		  }
	      }
	    CE_LETDELAY(env) = NewMap();
	    /* there is no letconsenv trickery here, as we assume that when
	       compiling bindings (typically in a BC_LET), the current env
	       has already been allocated by BC_CONSENV. */
	    flushDbgNames(f,env);
	  }
	else /* bindings like val _ = e, where we just want to evaluate e
		for its side-effects. */
	  {
	    compileStmt(f,env,e);
	  }
      }
    }
  QUIT;
}

PRIVATE void compileBindings(f,env,bindings,errp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr bindings;
INOUT ExprPtr errp;
{
#if DEBUGGING!=0
  ExprPtr savedz = currentDebugZone;
#endif

  while (bindings!=NIL)
    {
      compileBinding(f,env,CAR(bindings),errp);
      bindings = CDR(bindings);
    }
#if DEBUGGING!=0
  currentDebugZone = savedz;
#endif
}

PRIVATE void compileLet(f,env,bindings,body)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr bindings;
IN CONST ExprPtr body;
{
  ExprPtr gotop;
  COUNTER level;
  ENTER;

  level = f->stackDepth;
  gotop = beginForward(f);
  compileLet1(f,env,bindings,body,gotop,false);
  compilePushDrop(f,level,true); /* --- res */
  if (isForward(gotop))
    {
      ExprPtr endp;

      endp = beginForward(f);
      compileGoto(f,BC_GOTO,endp,0); /* goto just after the raise instruction */
      setStackDepth(f,level); /* --- */
      endForward(f,gotop);
      compileRaiseImm(f,env,I_BINDEXC); /* raise Bind */
      endForward(f,endp);
    }
  /* --- res */
  QUIT;
}

PRIVATE ExprPtr compilePolyFun1 P((IN BCFILE *f,INOUT compEnv *env,IN CONST ExprPtr func,
				   IN CONST COUNTER nargs,
				   IN CONST compileFunc compile,IN COUNTER push,
				   INOUT ExprPtr wrongp,IN CONST ExprPtr arglist));

PRIVATE void compileExprOrAppl (f, env, arg)
IN BCFILE *f;
INOUT compEnv *env;
IN CONST ExprPtr arg;
{
  if (env==NULL) /* then arg is an arglist, i.e., an external pointing to a struct argList,
		    coding an application. */
    compileApply1 (f, arglistEnv (arg), arglistArg (arg), arglistNext (arg),
		   false, compileExprFun, true, NULL);
  else
    compileExpr (f, env, arg);
}

#ifdef OBSOLETE
PRIVATE COUNTER compileExprList (f, env, arg)
IN BCFILE *f;
INOUT compEnv *env;
IN CONST ExprPtr arg;
{
  COUNTER n;

  for (n=0; arg!=NIL; arg=CDR(arg), n++)
    compileExprOrAppl(f,env,CAR(arg));
  return n;
}
#endif

#ifdef OBSOLETE
PRIVATE Boolean extractCon(e,ttmp,magicp,argp)
     IN CONST ExprPtr e;
     OUT ExprPtr *ttmp;
     OUT SHORT *magicp;
     OUT ExprPtr *argp;
{
  ExprPtr lbl,f;

 again:
  switch (e->kind)
    {
    case E_DEBUGZONE:
      e = DZCODE(e);
      goto again;
    case E_PAT_AS:
      e = CDR(e);
      goto again;
    case E_DATACON:
      lbl = COUNTEDVAL(e);
      if (lbl->magic<0)
	{
	  *magicp = lbl->magic;
	  *ttmp = CAR(lbl);
	  *argp = NIL;
	  return true;
	}
      return false;
    case E_APPL: case E_GOTO:
      f = APPLFUN(e);
    fagain:
      switch (f->kind)
	{
	case E_DEBUGZONE:
	  f = DZCODE(f);
	  goto fagain;
	case E_PAT_AS:
	  f = CDR(f);
	  goto fagain;
	case E_DATACON:
	  lbl = COUNTEDVAL(f);
	  if (lbl->magic>=0)
	    {
	      *magicp = lbl->magic;
	      *ttmp = CAR(lbl);
	      *argp = APPLARG(e);
	      return true;
	    }
	  return false;
	default:
	  return false;
	}
    default:
      return false;
    }
}
#endif

PRIVATE void compileStmtFun(f,env,e,wrongp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr e;
INOUT ExprPtr wrongp;
{
  compileStmt(f,env,e);
}

PRIVATE void compileStmtOrAppl (f, env, arg)
IN BCFILE *f;
INOUT compEnv *env;
IN CONST ExprPtr arg;
{
  if (env==NULL) /* then arg is an arglist, i.e., an external pointing to a struct argList,
		    coding an application. */
    compileApply1 (f, arglistEnv (arg), arglistArg (arg), arglistNext (arg),
		   false, compileStmtFun, false, NULL);
  else
    compileStmt (f, env, arg);
}

PRIVATE void compileExprOrApplNargs(f, env, arg, nargs)
IN BCFILE *f;
INOUT compEnv *env;
IN CONST ExprPtr arg;
IN COUNTER nargs;
{
  ExprPtr farg;

  switch (nargs)
    {
    case 0:
      compileStmtOrAppl(f, env, arg);
      break;
    case 1:
      compileExprOrAppl(f, env, arg);
      break;
    case 2:
      farg = forgetExtra(arg);
      if (env==NULL || (farg->kind!=(CHAR)E_TUPLE_2 && farg->kind!=(CHAR)E_UNSHARED_TUPLE_2))
	{
	  compileExprOrAppl(f, env, arg); /* --- (x,y) */
	  bcPut(BC_CARCDR, f);
	  bcPush(1, f); /* --- x y */
	}
      else
	{
	  compileExpr (f, env, CAR(farg)); /* --- x */
	  compileExpr (f, env, CDR(farg)); /* --- x y */
	}
      break;
    case 3:
      farg = forgetExtra(arg);
      if (env==NULL || (farg->kind!=(CHAR)E_TUPLE_GNRL && farg->kind!=(CHAR)E_UNSHARED_TUPLE_GNRL)
	  || AVAL(farg)->n!=3)
	{
	  compileExprOrAppl(f, env, arg); /* --- (x,y,z) */
	  bcPut(BC_3ARGS, f);
	  bcPush(2, f); /* --- x y z */
	}
      else
	{
	  compileExpr (f, env, AVAL(farg)->expr[0]); /* --- x */
	  compileExpr (f, env, AVAL(farg)->expr[1]); /* --- x y */
	  compileExpr (f, env, AVAL(farg)->expr[2]); /* --- x y z */
	}
      break;
    default:
      farg = forgetExtra(arg);
      if (env==NULL || (farg->kind!=(CHAR)E_TUPLE_GNRL && farg->kind!=(CHAR)E_UNSHARED_TUPLE_GNRL)
	  || AVAL(farg)->n!=nargs)
	{
	  compileExprOrAppl(f, env, arg); /* --- (x1, ..., xn) */
	  bcPut(BC_NARGS, f);
	  bcPut(nargs, f);
	  bcPush(nargs-1, f); /* --- x1 ... xn */
	}
      else
	{
	  COUNTER i;

	  for (i=0; i<nargs; i++)
	    compileExpr (f, env, AVAL(farg)->expr[i]);
	}
      break;
    }
}

PRIVATE COUNTER tryNargs(arg)
IN CONST ExprPtr arg;
{
  ExprPtr farg;

  farg = forgetExtra(arg);
  switch (farg->kind)
    {
    case E_TUPLE_2: case E_UNSHARED_TUPLE_2:
      return 2;
    case E_TUPLE_GNRL: case E_UNSHARED_TUPLE_GNRL:
      return AVAL(farg)->n;
    default:
      return 1;
    }
}

PRIVATE COUNTER tryNformals1 (pat)
     IN CONST ExprPtr pat;
{
  COUNTER nargs;

  nargs = 0;
 again:
  switch (pat->kind)
    {
    case E_TUPLE_2:
    case E_UNSHARED_TUPLE_2:
    case E_PAT_REVTUPLE_2:
      nargs = 2;
      break;
    case E_PAT_TUPLE_GNRL_CONS:
    case E_PAT_RECORD_GNRL_CONS:
    case E_PAT_REVTUPLE_GNRL_CONS:
    case E_PAT_REVRECORD_GNRL_CONS:
      nargs++;
      pat = CDR(pat);
      goto again;
    case E_TUPLE_0:
    case E_RECORD_0:
      break;
    default:
      nargs = -1; /* do not know */
      break;
    }
  return nargs;
}

PRIVATE COUNTER tryNformals (func)
     IN CONST ExprPtr func;
{
  ExprPtr match;
  ExprPtr pat;
  COUNTER nargs, n;

  match = COUNTEDVAL(func);
  pat = CAR(match);
  if (pat->kind==(CHAR)E_PATCHOICE)
    {
      nargs = tryNformals1 (CAR(pat));
      while (nargs>=0)
	{
	  pat = CAR(CDR(pat));
	  if (pat->kind==(CHAR)E_PATCHOICE)
	    {
	      n = tryNformals1 (CAR(pat));
	      if (n!=nargs)
		nargs = -1; /* do not know. */
	    }
	  else
	    {
	      n = tryNformals1 (pat);
	      if (n!=nargs)
		nargs = -1; /* do not know */
	      break; /* and stop. */
	    }
	}
    }
  else nargs = tryNformals1 (pat);
  return nargs;
}

PRIVATE ExprPtr compilePolyFun0(f,env,func,nargs,compile,push,wrongp,globerrp,finp,arglist)
IN BCFILE *f;
INOUT compEnv *env;
IN CONST ExprPtr func;
IN COUNTER nargs;
IN CONST compileFunc compile;
IN COUNTER push;
INOUT ExprPtr wrongp,globerrp,finp;
IN CONST ExprPtr arglist;
{
  ExprPtr choicelist,errp;
  COUNTER matchkind,idx,level,savelevel;
  ExprPtr match,pat,body,e;
  COUNTER nlocals;
  Boolean first;
  compEnv *locenv;
#if DEBUGGING!=0
  bcdState st, st0;
  INT dstart, dstart0 = 0;
  ExprPtr savedz = currentDebugZone;
#endif

  match = COUNTEDVAL(func);
  nlocals = COUNTEDN(func);

  locenv = (compEnv *)alloctmp(sizeof(compEnv));
  *locenv = *env;
  locenv->_comp.next = theHeapFrames;
  theHeapFrames = (gcStackView *)&locenv->_comp;

  savelevel = f->stackDepth;
  /* --- arg1 ... argn
   where n=nargs */
  choicelist = NIL;
  matchkind = CHOICE_MATCH;
  idx = 0;
  level = f->stackDepth-nargs;
  errp = globerrp;
  if (nargs!=1)
    matchkind = NFOLD_MATCH;
  first = true;
#if DEBUGGING!=0
  if (debugEnabled && CE_PROF(env)!=NIL
#if PROFILING
      && CE_PROF(env)!=defaultProf
#endif
      )
    {
      ExprPtr name;
      INT i;

      dstart0 = beginDebug(f,&st0);
      name = CAR(ARGS(CE_PROF(env)));
      i = getConstant (f, name);
      bcdPut(DBG_NAME,f);
      bcdPut((OPCODE)i,f);
    }
#endif
  while (true)
    {
      pat = CAR(match);
#if DEBUGGING!=0
      dstart = beginDebug(f,&st);
#endif
      locenv->ilevel = locenv->level+1;
      switch (matchkind)
	{
	case CHOICE_MATCH:
	  choicelist = compilePattern(f,locenv,pat,errp,choicelist,false);
	  break;
	case VECTOR_MATCH:
	  choicelist = compileVectorMatch(f,locenv,pat,idx,level,
					  errp,choicelist);
	  break;
	case NFOLD_MATCH:
	  choicelist = compileNfoldMatch(f,locenv,pat,idx,level,
					 errp,choicelist);
	  break;
	case RECORDMAPADD_MATCH:
	  choicelist = compileRecordGnrlMapAddMatch(f,locenv,pat,level,
						    errp,choicelist);
	  break;
	case SWITCH_MATCH:
	  choicelist = compilePattern(f,locenv,pat,errp,choicelist,true);
	  break;
	}
      /* --- <...> */
      locenv->level = locenv->ilevel;
      if ((locenv->flags & CEF_NEEDENV) && !needEnv(CDR(match),locenv))
	locenv->flags &= ~CEF_NEEDENV;
      if (locenv->flags & CEF_NEEDENV)
	{
	  compileCurEnv(f,env); /* --- <...> env (if needenv); --- <...> (if !needenv) */
	  /* Notice that we compileCurEnv() on env, not locenv: we want the old
	     environment. */
	  bcPut(BC_CONSENV,f);
	  bcPut((OPCODE)nlocals,f);
	  newCurEnv(locenv,f->stackDepth-1);
	  dumpLetStores(f,f->stackDepth-1,CE_LETDELAY(locenv));
	}
      /* --- <...> env (if needenv); --- <...> (if !needenv) */
      CE_LETDELAY(locenv) = NewMap();
      /* This is what we have called someplace else 'letconsenv trickery':
	 we only allocate the environment after all patterns have been
	 matched, and then we fill it in with the values we have gotten
	 (and only if such an environment is needed). */
      flushDbgNames(f,locenv);

      body = CDR(match);

#if DEBUGGING
      currentDebugZone = savedz;
      if (debugEnabled)
	{
	  ExprPtr e;

	  e=body;
	dbg_again:
	  switch (e->kind)
	    {
	    case E_DEBUGZONE:
	      currentDebugZone = DZINFO(e);
	      /*e = DZCODE(e);*/
	      break;
	    case E_PROFILE:
	      e = CAR(e);
	      goto dbg_again;
	    default:
	      break;
	    }
	}
#endif

      compileDebugZone(true,f);
      compileCodePointOp(BC_FUNCTION,f);


      if (first && arglist!=NIL && choicelist==NIL)
	{
	  ExprPtr saveprof;
	  struct Var saveme;

	  saveprof = CE_PROF(locenv);
	  saveme = locenv->me;
	  e = body;
reapply:
	  switch (e->kind)
	    {
	    case E_DEBUGZONE:
#if DEBUGGING!=0
	      if (debugEnabled)
		currentDebugZone = DZINFO(e);
#endif
	      e = DZCODE(e);
	      goto reapply;
	    case E_PROFILE:
	      CE_PROF(locenv) = CDR(e);
	      e = CAR(e);
	      goto reapply;
	    case E_PAT_AS:
	      e = CDR(e);
	      defineMe (locenv, NIL); /* disallow tail recursions from inlined body
				      if inlined function is not itself in tail position. */
	      goto reapply;
	    case E_MEMO_FN:
	      e = MEMOFUN(e); /* a memo function as immediate value in the code does not
				 memoize (or it would memoize into a closure that would
				 be immediately garbageable) */
	      goto reapply;
	    case E_POLYFUN:
	      {
		COUNTER nargs, nformals;

		nargs = tryNargs (arglistArg (arglist));
		nformals = tryNformals (e);
		if (nformals!=nargs)
		  nargs = 1; /* if function and argument do not agree
				on number of arguments, then allocate tuples. */
		if (nargs==1)
		  {
		    compileExprOrAppl (f,arglistEnv (arglist),arglistArg (arglist));
		    /* we evaluate the argument in its associated environment,
		       not locenv.  Otherwise references to variables
		       run a chance of being wrong. */
		    arglist = arglistNext (arglist);
		    arglist = compilePolyFun1(f,locenv,e,1,compile,push,wrongp,arglist);
		  }
		else
		  {
		    compileExprOrApplNargs (f, arglistEnv (arglist), arglistArg (arglist),
					    nargs);
		    /* we evaluate the argument in its associated environment,
		       not locenv.  Otherwise references to variables
		       run a chance of being wrong. */
		    arglist = arglistNext (arglist);
		    arglist = compilePolyFun1(f,locenv,e,nargs,compile,push,wrongp,arglist);
		  }
		  break;
	      }
	    default:
	      compileExpr(f,locenv,body);
	      break;
	    }
	  CE_PROF(locenv) = saveprof;
	  locenv->me = saveme;
	}
      else
	{
	  (*compile)(f,locenv,body,wrongp);
	}
      if (arglist!=NIL)
	push = 1; /* We have been unable to fully apply the function to all
		     of its arguments, so we give back an ExprPtr on top
		     of the run-time stack, just as though push had been
		     set to 1 in the first place. compileApply1() will then
		     do the rest. */

#if DEBUGGING!=0
      endDebug(dstart,f,&st);
#endif
      first = false;
      switch (push)
	{
	case 1: /* when compile==compileExpr, where we really want to return a result */
	  compileSwDrop(f,f->stackDepth-savelevel+nargs); /* --- res */
	  break;
	case 2: /* when moreover this is a return from a function */
	  compileDebugZone(false,f);
	  compileCodePointOp(BC_RETURN,f);
	  bcPut((OPCODE)(f->stackDepth-1),f); /* --- res */
	  bcPush(savelevel-nargs+1-f->stackDepth,f);
	  f->dead = 1;
	  break;
	default: /* case 0: don't do compileDrop() if push==0 (as compilePushDrop() would do),
		    since it will be done anyway by the next goto, unless there is no element
		    left in choicelist (see below). */
	  break;
	}
      locenv->level--;
      locenv->flags = env->flags;

      if (choicelist==NIL)
	{
	  if (!push)
	    /* happens when compile==compileTest or compileProg, where we don't want
	       to return any result */
	    compileDrop(f,f->stackDepth-savelevel+nargs); /* --- */
	  break;
	}
      compileGoto(f,BC_GOTO,finp,0);

      setStackDepth(f,CH_STACKLEVEL(choicelist));
      endForward(f,CH_ERRP(choicelist));
      level = CH_DATALEVEL(choicelist);
      match = CH_PATCHOICE(choicelist);
      matchkind = CH_MATCHKIND(choicelist);
      idx = CH_IDX(choicelist);
      CE_LETDELAY(locenv) = CH_LETDELAY(choicelist);
      CE_VARS(locenv) = CH_VARS(choicelist);
      CE_INLINES(locenv) = CH_INLINES(choicelist);
      CE_DBGNAMES(locenv) = CH_DBGNAMES(choicelist);
      CE_PARENT_DBGNAMES(locenv) = CH_PARENT_DBGNAMES(choicelist);
      errp = CH_SWITCH_ERRP(choicelist);
      choicelist = CH_NEXT(choicelist);
    }

  theHeapFrames = locenv->_comp.next;
#if DEBUGGING!=0
  if (debugEnabled && CE_PROF(env)!=NIL
#if PROFILING
      && CE_PROF(env)!=defaultProf
#endif
      )
    {
      endDebug(dstart0,f,&st0);
    }
  currentDebugZone = savedz;
#endif
  return arglist;
}

DEF_PRIVATE_ACTION_1ARG(compilepolyfun1)
{
  struct compilepolyfun1 *data;

  data = (struct compilepolyfun1 *)ENV_UNDERLYING(env);
  return compilePolyFun1(data->f,data->env,data->func,data->nargs,
			 data->compile,data->push,
			 data->wrongp,data->arglist);
}
ACTION_END

PRIVATE ExprPtr compilePolyFun1k(f,env,func,nargs,compile,push,wrongp,arglist)
IN BCFILE *f;
INOUT compEnv *env;
IN CONST ExprPtr func;
IN COUNTER nargs;
IN CONST compileFunc compile;
IN COUNTER push;
INOUT ExprPtr wrongp;
IN CONST ExprPtr arglist;
{
  struct compilepolyfun1 *data =
    (struct compilepolyfun1 *)alloctmp(sizeof(struct compilepolyfun1));

  data->f = f;
  data->env = env;
  data->func = func;
  data->compile = compile;
  data->push = push;
  data->wrongp = wrongp;
  data->arglist = arglist;
  data->nargs = nargs;
  return CallCC(ACTION_NAME(compilepolyfun1),(ExprPtr)data,true);
}

PRIVATE ExprPtr compilePolyFun1(f,env,func,nargs,compile,push,wrongp,arglist)
IN BCFILE *f;
INOUT compEnv *env;
IN CONST ExprPtr func;
IN COUNTER nargs;
IN CONST compileFunc compile;
IN COUNTER push;
INOUT ExprPtr wrongp;
IN CONST ExprPtr arglist;
{
  ExprPtr globerrp,finp;
  COUNTER level, savelevel;

  if (STKFULL(level))
    {
      return compilePolyFun1k(f,env,func,nargs,compile,push,wrongp,arglist);
    }
  level = f->stackDepth-nargs;
  finp = beginForward(f);
  globerrp = beginForward(f);
  arglist = compilePolyFun0(f,env,func,nargs,compile,push,wrongp,globerrp,finp,arglist);
  savelevel = f->stackDepth;
  if (arglist!=NIL)
    push = 1;
  if (isForward(globerrp))
    {
      compileGoto(f,BC_GOTO,finp,0);
      setStackDepth(f,level);
      endForward(f,globerrp);
      compileRaiseImm(f,env,I_MATCHEXC);
      if (push==2)
	{
	  compileDebugZone(false,f);
	  compileCodePointOp(BC_RETURN,f);
	  bcPut((OPCODE)(f->stackDepth-1),f); /* --- res */
	}
      else
	compilePushDrop(f,level,push);
    }
  setStackDepth (f, savelevel);
  if (isForward(finp))
    {
      endForward(f,finp);
    }
  return arglist;
}

PRIVATE void compilePolyFun(f,env,func,keepme)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr func;
IN Boolean keepme;
{
  BCFILE *ff;
  compEnv *newenv;
  ExprPtr e;
  Boolean needsenv;
#if DEBUGGING!=0
  ExprPtr savedz = currentDebugZone;
#endif
  COUNTER nargs;
  ENTER;

  newenv = (compEnv *)alloctmp(sizeof(compEnv));
  *newenv = *env;
  newenv->_comp.next = theHeapFrames;
  theHeapFrames = (gcStackView *)&newenv->_comp;

  CE_VARS(newenv) = NewMap();
  newCurEnv(newenv,0); /* record that current environment is at position 0 in the stack */
  CE_HANDLERS(newenv) = NIL;
  CE_LETDELAY(newenv) = NewMap();
  CE_DZ(newenv) = NIL;
  resetDbgNames (newenv);
  if (!keepme)
    defineMe(newenv,NIL);

  newenv->flags |= CEF_NEEDENV;
  /* A priori, any body of a function needs a heap-allocated environment.
     When it does not, compilePolyFun0() will be able to decide it by itself. */
  needsenv = needEnv(func,env);

  ff = needsenv?bcOpenD(env):bcOpenE(env);
  /* if needsenv, then a proper environment will be present as first
     element in the stack, and we use bcOpenD().  Otherwise, there
     is no need to compile a DBG_ENV debugging instruction. */
  ff->ntargs = FUNARITY(func);
  nargs = tryNformals (func);
  if (nargs<0) /* unknown arity */
    nargs = 1;
  ff->nargs = nargs;
  ff->stackDepth = 1+nargs;

#ifdef OBSOLETE
  compileDebugZone(true,ff); /* probably useless: will record just any
				debug zone, not necessarily the one
				at the start of function. */
  compileCodePointOp(BC_FUNCTION,ff);
#endif
  (void) compilePolyFun1(ff,newenv,func,nargs,compileExprFun,2,NULL,NIL);
  /* As push==2, compilePolyFun1 produces code that returns on its own, so the
     following is useless:
     compileDebugZone(false,ff);
     compileCodePointOp(BC_RETURN,ff);
     bcPut((OPCODE)1,ff);
     */
  e = bcCloseD (ff);

  compileClosureEnv(f,e,(needsenv?env:NULL));

#if DEBUGGING!=0
  currentDebugZone = savedz;
#endif
  theHeapFrames = newenv->_comp.next;
  QUIT;
}

PRIVATE void compileMemoFn(f,env,func,rem)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr func,rem;
{
  BCFILE *ff;
  compEnv *newenv;
  ExprPtr e;
  ExprPtr errp;
  Boolean needsenv;
  ENTER;

  /* memofn x => f x ==
     let val r = ref {} in
     fn x => if x inset !r
     		then ? (!r) x
	     else let val y = f x
	          in
		     r := !r ++ {x => y}; y
		  end
      end
      (* this function is called the stub *)
      */
  needsenv = needEnv(func,env);
  ff = needsenv?bcOpenD(env):bcOpenE(env);
  /* Now, this function starts with the closure environment in position 0;
     this env is actually cons(memoizing set, normal env).
     (This cons-cell is mutable in its first position.) */
  ff->ntargs = FUNARITY(func);
#ifdef OBSOLETE
  compileDebugZone(true,ff);
  compileCodePointOp(BC_FUNCTION,ff); /* probably useless: will record just any
					 debug zone, not necessarily the one
					 at the start of function. */
#endif
  /* --- (memo,env) arg */
  errp = beginForward(ff);
  compileGoto(ff,BC_NCARGET,errp,0); /* --- res ; if success */
  compileDebugZone(false,ff);
  compileCodePointOp(BC_RETURN,ff);
  bcPut((OPCODE)0,ff);
  ff->dead = 1;
  /* --- res */
  /* if res was not found: */
  setStackDepth(ff,2);
  endForward(ff,errp); /* --- (memo,env) arg */
  compileFetchCdr(ff,0); /* --- (memo,env) arg env */
  compileFetch(ff,1); /* --- (memo,env) arg env arg */

  newenv = (compEnv *)alloctmp(sizeof(compEnv));
  *newenv = *env;
  newenv->_comp.next = theHeapFrames;
  theHeapFrames = (gcStackView *)&newenv->_comp;

  CE_VARS(newenv) = NewMap();
  newCurEnv(newenv,2); /* record that current environment is at position 2 in the stack */
  CE_HANDLERS(newenv) = NIL;
  CE_LETDELAY(newenv) = NewMap();
  CE_DZ(newenv) = NIL;
  resetDbgNames (newenv);
  defineMe(newenv,NIL); /* cannot do tail recursion on memo functions */

  newenv->flags |= CEF_NEEDENV;
  /* A priori, any body of a function needs a heap-allocated environment.
     When it does not, compilePolyFun0() will be able to decide it by itself. */

  (void) compilePolyFun1(ff,newenv,func,1,compileExprFun,1,NULL,NIL);
  /* --- (memo,env) arg env res */
  compileSwDrop(ff,2);
  /* --- (memo,env) arg res */
  bcPut(BC_NCARPUT,ff);
  bcPush(-2,ff); /* --- res */
  compileDebugZone(false,ff);
  compileCodePointOp(BC_RETURN,ff);
  bcPut((OPCODE)0,ff);
  ff->dead = 1;
  e = bcCloseD (ff);

  compileConstant(f,NIL /*rem*/);
  if (needsenv) /* If func has free variables, then build a full closure: */
    compileCurEnv(f,env);
  else /* Otherwise, if memo function does not have free variables */
    compileConstant(f,NIL); /* compile a dummy environment */
  bcPut(BC_CONSMEMO,f);
  bcPush(-1,f);
  compileClosure(f,e);

  theHeapFrames = newenv->_comp.next;
  QUIT;
}

PRIVATE ExprPtr forgetExtra(e)
IN CONST ExprPtr e;
{
again:
  switch (e->kind)
    {
    case E_DEBUGZONE:
      e = DZCODE(e);
      goto again;
    case E_PROFILE:
      e = CAR(e);
      goto again;
    }
  return e;
}

PRIVATE Boolean isMe(env,fun)
IN CONST compEnv *env;
IN CONST ExprPtr fun;
{
  fun = forgetExtra(fun);
  return fun->kind==(CHAR)E_VARREF
    && env->level-VARLEVEL(fun)==env->me.level
      && VARRANK(fun)==env->me.rank;
}

/***** Compiling well-known primitives: *****/

PRIVATE void compileTwoArgs(f,arg_env,arg,revp)
IN BCFILE *f;
IN CONST compEnv *arg_env;
IN CONST ExprPtr arg;
IN Boolean revp;
{
  arg = forgetExtra(arg);
  if (arg->kind==(CHAR)E_TUPLE_2 || arg->kind==(CHAR)E_UNSHARED_TUPLE_2)
    {
      compileExpr(f,arg_env,CAR(arg));
      compileExpr(f,arg_env,CDR(arg));
      if (revp)
	{
	  bcPut(BC_SWAP,f);
	}
    }
  else
    {
      compileExpr(f,arg_env,arg);
      bcPut(revp?BC_CDRCAR:BC_CARCDR,f);
      bcPush(1,f);
    }
}

PRIVATE void compileThreeArgs(f,arg_env,arg)
IN BCFILE *f;
IN CONST compEnv *arg_env;
IN CONST ExprPtr arg;
{
  arg = forgetExtra(arg);
  if ((arg->kind==(CHAR)E_TUPLE_GNRL || arg->kind==(CHAR)E_UNSHARED_TUPLE_GNRL)
      && AVAL(arg)->n==3)
    {
      compileExpr(f,arg_env,AVAL(arg)->expr[0]);
      compileExpr(f,arg_env,AVAL(arg)->expr[1]);
      compileExpr(f,arg_env,AVAL(arg)->expr[2]);
    }
  else
    {
      compileExpr(f,arg_env,arg);
      bcPut(BC_3ARGS,f);
      bcPush(2,f);
    }
}

PRIVATE Boolean compileOneArgAndMaplet(f,arg_env,arg)
IN BCFILE *f;
IN CONST compEnv *arg_env;
IN CONST ExprPtr arg;
{
  ExprPtr cdr;

  arg = forgetExtra(arg);
  if (arg->kind==(CHAR)E_TUPLE_2 || arg->kind==(CHAR)E_UNSHARED_TUPLE_2)
    {
      compileExpr(f,arg_env,CAR(arg));
      cdr = CDR (arg);
      cdr = forgetExtra(cdr);
      if (cdr->kind==(CHAR)E_MAPLET)
	{
	  compileExpr(f,arg_env,MAPLETX(cdr));
	  compileExpr(f,arg_env,MAPLETY(cdr));
	  return true;
	}
      else
	{
	  compileExpr(f,arg_env,cdr);
	  return false;
	}
    }
  else
    {
      compileExpr(f,arg_env,arg);
      bcPut(BC_CARCDR,f);
      bcPush(1,f);
      return false;
    }
}

#ifdef HAS_PROTOTYPES
PRIVATE void compileMonOp(IN OPCODE opcode,IN CONST applyData *ad,IN Boolean mayraise)
#else
  PRIVATE void compileMonOp(opcode,ad,mayraise)
IN OPCODE opcode;
IN CONST applyData *ad;
IN Boolean mayraise;
#endif
{
  compileExpr(ad->f,ad->arg_env,ad->arg);
  if (mayraise)
    compileCodePointOp(opcode,ad->f);
  else
    bcPut(opcode,ad->f);
  if (ad->wrongp!=NULL) /* this is a test */
    compileGoto(ad->f,BC_IF,ad->wrongp,1);
  else if (!ad->push) /* this is a statement */
    compileDrop(ad->f,1);
}

#ifdef HAS_PROTOTYPES
PRIVATE void compileBinOp(IN OPCODE opcode,IN CONST applyData *ad,IN Boolean mayraise)
#else
PRIVATE void compileBinOp(opcode,ad,mayraise)
IN OPCODE opcode;
IN CONST applyData *ad;
IN Boolean mayraise;
#endif
{
  compileTwoArgs(ad->f,ad->arg_env,ad->arg,false);
  if (mayraise)
    compileCodePointOp(opcode,ad->f);
  else
    bcPut(opcode,ad->f);
  bcPush(-1,ad->f);
  if (ad->wrongp!=NULL) /* this is a test */
    compileGoto(ad->f,BC_IF,ad->wrongp,1);
  else if (!ad->push) /* this is a statement */
    compileDrop(ad->f,1);
}

#ifdef HAS_PROTOTYPES
PRIVATE void compileTerOp(IN OPCODE opcode,IN CONST applyData *ad,IN Boolean mayraise)
#else
PRIVATE void compileTerOp(opcode,ad,mayraise)
IN OPCODE opcode;
IN CONST applyData *ad;
IN Boolean mayraise;
#endif
{
  compileThreeArgs(ad->f,ad->arg_env,ad->arg);
  if (mayraise)
    compileCodePointOp(opcode,ad->f);
  else
    bcPut(opcode,ad->f);
  bcPush(-2,ad->f);
  if (ad->wrongp!=NULL) /* this is a test */
    compileGoto(ad->f,BC_IF,ad->wrongp,1);
  else if (!ad->push) /* this is a statement */
    compileDrop(ad->f,1);
}

#ifdef HAS_PROTOTYPES
PRIVATE void compileTerOpStmt(IN OPCODE opcode,IN CONST applyData *ad,IN Boolean mayraise)
#else
PRIVATE void compileTerOpStmt(opcode,ad,mayraise)
IN OPCODE opcode;
IN CONST applyData *ad;
IN Boolean mayraise;
#endif
{
  compileThreeArgs(ad->f,ad->arg_env,ad->arg);
  if (mayraise)
    compileCodePointOp(opcode, ad->f);
  else
    bcPut(opcode,ad->f);
  bcPush(-3,ad->f);
  if (ad->wrongp!=NULL) /* this is a test */
    ; /* should not happen. // compileGoto(ad->f,BC_IF,ad->wrongp,1); */
  else if (!ad->push) /* this is a statement */
    ;
  else compileConstant(ad->f,tuple0);
}

#define DEF_MONOP(name,opcode,mayraise) DEF_ACTION_1ARG(name)	\
  { compileMonOp(opcode,APPLYDATA,mayraise); return AD_OTHERS; } ACTION_END
#define DEF_BINOP(name,opcode,mayraise) DEF_ACTION_1ARG(name) \
  { compileBinOp(opcode,APPLYDATA,mayraise); return AD_OTHERS; } ACTION_END
#define DEF_TEROP(name,opcode,mayraise) DEF_ACTION_1ARG(name) \
{ compileTerOp(opcode,APPLYDATA,mayraise); return AD_OTHERS; } ACTION_END
#define DEF_TEROP_STMT(name,opcode,mayraise) DEF_ACTION_1ARG(name) \
{ compileTerOpStmt(opcode,APPLYDATA,mayraise); return AD_OTHERS; } ACTION_END

DEF_MONOP(bc_iabs,BC_IABS,false)
DEF_MONOP(bc_ineg,BC_INEG,false)
DEF_MONOP(bc_isqr,BC_ISQR,false)
DEF_MONOP(bc_inot,BC_INOT,false)

DEF_BINOP(bc_iplus,BC_IADD,false)
DEF_BINOP(bc_iminus,BC_ISUB,false)
DEF_BINOP(bc_itimes,BC_IMUL,false)

DEF_ACTION_1ARG(bc_overwrite)
{
  if (compileOneArgAndMaplet(AD_F,AD_ARG_ENV,AD_ARG))
    {
      bcPut (BC_MAPADD2,AD_F);
      bcPush (-2,AD_F);
    }
  else
    {
      bcPut (BC_OVERWRITE,AD_F);
      bcPush (-1,AD_F);
    }
  if (AD_WRONGP!=NULL) /* this is a test */
    compileGoto(AD_F,BC_IF,AD_WRONGP,1);
  else if (!AD_PUSH) /* this is a statement */
    compileDrop(AD_F,1);
  return AD_OTHERS;
}
ACTION_END

/*DEF_BINOP(bc_overwrite,BC_OVERWRITE,false)*/
DEF_BINOP(bc_domrestrto,BC_MAPDOMRESTRTO,false)
DEF_BINOP(bc_domrestrby,BC_MAPDOMRESTRBY,false)
DEF_BINOP(bc_diff,BC_DIFF,false)
DEF_BINOP(bc_delta,BC_DELTA,false)

#ifdef HAS_PROTOTYPES
PRIVATE void compileDivOp(IN OPCODE opcode,IN CONST applyData *ad)
#else
PRIVATE void compileDivOp(opcode,ad)
IN OPCODE opcode;
IN CONST applyData *ad;
#endif
{
  ExprPtr zerop,finp;
  COUNTER level;

  level = ad->f->stackDepth;
  compileTwoArgs(ad->f,ad->arg_env,ad->arg,false); /* --- a b */
  zerop = beginForward(ad->f);
  finp = beginForward(ad->f);
  compileGoto(ad->f,BC_ISNZERO,zerop,0);
  bcPut(opcode,ad->f);
  bcPush(-1,ad->f); /* --- a op b */
  compileGoto(ad->f,BC_GOTO,finp,0);
  setStackDepth(ad->f,level); /* --- */
  endForward(ad->f,zerop);
  compileRaiseImm(ad->f,ad->env,I_ARITHEXC);
  if (ad->wrongp!=NULL) /* this is a test */
    compileGoto(ad->f,BC_IF,ad->wrongp,1);
  else if (!ad->push) /* this is a statement */
    compileDrop(ad->f,1);
  endForward(ad->f,finp); /* --- res (if push); or ---- */
}

DEF_ACTION_1ARG(bc_idiv) { compileDivOp(BC_IDIV,APPLYDATA); return AD_OTHERS; } ACTION_END
DEF_ACTION_1ARG(bc_imod) { compileDivOp(BC_IMOD,APPLYDATA); return AD_OTHERS; } ACTION_END

DEF_BINOP(bc_ior,BC_IOR,false)
DEF_BINOP(bc_ixor,BC_IXOR,false)
DEF_BINOP(bc_iand,BC_IAND,false)
DEF_BINOP(bc_ilsl,BC_ILSL,false)
DEF_BINOP(bc_ilsr,BC_ILSR,false)
DEF_BINOP(bc_iasr,BC_IASR,false)

PRIVATE void testToBool(f,wrongp)
IN BCFILE *f;
INOUT ExprPtr wrongp;
{
  ExprPtr finp;
  COUNTER level;

  level = f->stackDepth;
  finp = beginForward(f);
  /* if test is true, return trueValue: */
  compileConstant(f,trueValue);
  compileGoto(f,BC_GOTO,finp,0);
  /* otherwise, return falseValue: */
  setStackDepth(f,level);
  endForward(f,wrongp);
  compileConstant(f,falseValue);
  endForward(f,finp);
}

#ifdef HAS_PROTOTYPES
PRIVATE void compileCmpOp(IN OPCODE opcode,IN applyData *ad,IN Boolean mayraise)
#else
PRIVATE void compileCmpOp(opcode,ad,mayraise)
IN OPCODE opcode;
IN applyData *ad;
IN Boolean mayraise;
#endif
{
  compileTwoArgs(ad->f,ad->arg_env,ad->arg,false); /* --- a b */
  if (mayraise)
    compileCodePoint(ad->f);
  if (ad->wrongp==NULL)
    {
      ExprPtr wrongp;

      if (!ad->push && !mayraise) /* this is a statement which cannot raise any exception:
				     result is ignored */
	compileDrop(ad->f,2);
      else /* this is an expression, generate true or false according to result of test */
	{
	  wrongp = beginForward(ad->f);
	  compileGoto(ad->f,opcode,wrongp,2); /* --- */
	  testToBool(ad->f,wrongp);
	}
    }
  else
    compileGoto(ad->f,opcode,ad->wrongp,2); /* --- */
}

#define DEF_CMPOP(name,opcode,mayraise)						\
  DEF_ACTION_1ARG(name) { compileCmpOp(opcode,APPLYDATA,mayraise); return AD_OTHERS; } \
  ACTION_END

DEF_CMPOP(bc_ilt,BC_ILT,false)
DEF_CMPOP(bc_ile,BC_ILE,false)
DEF_CMPOP(bc_igt,BC_IGT,false)
DEF_CMPOP(bc_ige,BC_IGE,false)

DEF_BINOP(bc_imin,BC_IMIN,false)
DEF_BINOP(bc_imax,BC_IMAX,false)

DEF_CMPOP(bc_rlt,BC_RLT,true)
DEF_CMPOP(bc_rle,BC_RLE,true)
DEF_CMPOP(bc_rgt,BC_RGT,true)
DEF_CMPOP(bc_rge,BC_RGE,true)

DEF_BINOP(bc_rmin,BC_RMIN,true)
DEF_BINOP(bc_rmax,BC_RMAX,true)

DEF_MONOP(bc_floor,BC_FLOOR,true)
DEF_MONOP(bc_ceil,BC_CEIL,true)

DEF_BINOP(bc_aref,BC_AREF,true)
DEF_TEROP_STMT(bc_aset,BC_ASET,true)
DEF_BINOP(bc_iaref,BC_IAREF,true)
DEF_TEROP_STMT(bc_iaset,BC_IASET,true)

#ifdef HAS_PROTOTYPES
PRIVATE void compileHdOp(IN OPCODE opcode,IN CONST applyData *ad)
#else
PRIVATE void compileHdOp(opcode,ad)
IN OPCODE opcode;
IN CONST applyData *ad;
#endif
{
  ExprPtr wrongp,finp;
  COUNTER level;
  Boolean push;

  level = ad->f->stackDepth;
  push = ad->push || ad->otherargs!=NIL; /* if there are remaining arguments, then we must push the
					    result: it will be applied to the remaining arguments. */
  compileExpr(ad->f,ad->arg_env,ad->arg);
  wrongp = beginForward(ad->f);
  finp = beginForward(ad->f);
  compileGoto(ad->f,BC_ISCONS,wrongp,0);
  if (push || ad->wrongp!=NULL) /* only compute result if expression or a test */
    bcPut(opcode,ad->f); /* --- res */
  compileGoto(ad->f,BC_GOTO,finp,0);
  setStackDepth(ad->f,level); /* --- */
  endForward(ad->f,wrongp);
  compileRaiseImm(ad->f,ad->env,I_MATCHEXC);
  setStackDepth(ad->f,level+1);
  endForward(ad->f,finp);
  if (!push)
    {
      if (ad->wrongp==NULL) /* a statement */
	compileDrop(ad->f,1);
      else compileGoto(ad->f,BC_IF,ad->wrongp,1);
    }
}

#define DEF_HDOP(name,opcode) DEF_ACTION_1ARG(name) \
  { compileHdOp(opcode,APPLYDATA); return AD_OTHERS; } ACTION_END

DEF_HDOP(bc_hd,BC_CAR)
DEF_HDOP(bc_tl,BC_CDR)

DEF_ACTION_1ARG(bc_int)
{
  ExprPtr wrongp,finp;
  COUNTER level;

  level = AD_F->stackDepth;
  compileExpr(AD_F,AD_ARG_ENV,AD_ARG);
  wrongp = beginForward(AD_F);
  finp = beginForward(AD_F);
  compileGoto(AD_F,BC_HASTAG+E_REAL,wrongp,0);
  if (AD_PUSH)
    bcPut(BC_NUMINT,AD_F); /* --- res */
  compileGoto(AD_F,BC_GOTO,finp,0);
  setStackDepth(AD_F,level); /* --- */
  endForward(AD_F,wrongp);
  compileRaiseImm(AD_F,AD_ENV,I_INCOMPEXC);
  setStackDepth(AD_F,level+1); /* --- res */
  endForward(AD_F,finp);
  if (!AD_PUSH)
    compileDrop(AD_F,1);
  /* cannot be a test, only a statement or an expression */
  return AD_OTHERS;
}
ACTION_END

DEF_MONOP(bc_num,BC_INTNUM,false)

#ifdef HAS_PROTOTYPES
PRIVATE void compileTstOp(IN OPCODE opcode,IN applyData *ad)
#else
PRIVATE void compileTstOp(opcode,ad)
IN OPCODE opcode;
IN applyData *ad;
#endif
{
  if (ad->wrongp==NULL)
    {
      ExprPtr wrongp;

      if (!ad->push) /* this is a statement: result is ignored, so don't even test */
	(*ad->compile)(ad->f,ad->arg_env,ad->arg,ad->wrongp);
      else /* this is an expression, generate true or false according to result of test */
	{
	  compileExpr(ad->f,ad->arg_env,ad->arg); /* --- a */
	  wrongp = beginForward(ad->f);
	  compileGoto(ad->f,opcode,wrongp,0); /* --- a */
	  compileDrop(ad->f,1);
	  testToBool(ad->f,wrongp);
	}
    }
  else
    {
      compileExpr(ad->f,ad->arg_env,ad->arg); /* --- a */
      compileGoto(ad->f,opcode,ad->wrongp,0); /* --- a */
      compileDrop(ad->f,1);
    }
}

#define DEF_TSTOP(name,opcode) \
  DEF_ACTION_1ARG(name) { compileTstOp(opcode,APPLYDATA); return AD_OTHERS; } ACTION_END

DEF_TSTOP(bc_empty,BC_ISEMPTY)
DEF_TSTOP(bc_null,BC_ISNIL)
DEF_ACTION_1ARG(bc_before)
{
  ExprPtr e;

  e = forgetExtra(AD_ARG);
  if (e->kind==(CHAR)E_TUPLE_2 || e->kind==(CHAR)E_UNSHARED_TUPLE_2)
    {
      if (AD_OTHERS!=NIL) /* there are remaining arguments: compile as an expression */
	{
	  compileExpr(AD_F,AD_ARG_ENV,CAR(e));
	  compileStmt(AD_F,AD_ARG_ENV,CDR(e));
	}
      else if (AD_WRONGP==NULL) /* expression or statement, and there are no remaining arguments */
	{
	  (*AD_COMPILE)(AD_F,AD_ARG_ENV,CAR(e),AD_WRONGP);
	  compileStmt(AD_F,AD_ARG_ENV,CDR(e));
	}
      else /* test, and there are no remaining arguments */
	{
	  compileExpr(AD_F,AD_ARG_ENV,CAR(e));
	  compileStmt(AD_F,AD_ARG_ENV,CDR(e));
	  compileGoto(AD_F,BC_IF,AD_WRONGP,1);
	}
    }
  else
    {
      if (AD_OTHERS!=NIL) /* there are remaining arguments: compile as an expression */
	goto compile_as_expr;
      if (AD_WRONGP==NULL)
	{
	  if (AD_PUSH)
	    { /* expression */
	    compile_as_expr:
	      compileExpr(AD_F,AD_ARG_ENV,e);
	      bcPut(BC_CAR,AD_F);
	    }
	  else compileStmt(AD_F,AD_ARG_ENV,e); /* statement, and there are no remaining arguments */
	}
      else
	{ /* test and there are no remaining arguments */
	  compileExpr(AD_F,AD_ARG_ENV,e);
	  compileGoto(AD_F,BC_IF,AD_WRONGP,1);
	}
    }
  return AD_OTHERS;
}
ACTION_END

DEF_MONOP(bc_size,BC_SIZE,false)

DEF_ACTION_1ARG(bc_inset)
{
  if (AD_WRONGP==NULL)
    {
      if (!AD_PUSH)
	(*AD_COMPILE)(AD_F,AD_ARG_ENV,AD_ARG,AD_WRONGP); /* result is ignored, so don't even compile
							    the test. */
      else
	{
	  ExprPtr wrongp;

	  compileTwoArgs(AD_F,AD_ARG_ENV,AD_ARG,true); /* --- m x */
	  wrongp = beginForward(AD_F);
	  compileGoto(AD_F,BC_MAPGET,wrongp,1); /* --- y */
	  compileDrop(AD_F,1);
	  testToBool(AD_F,wrongp);
	}
    }
  else
    {
      compileTwoArgs(AD_F,AD_ARG_ENV,AD_ARG,true); /* --- m x */
      compileGoto(AD_F,BC_MAPGET,AD_WRONGP,1); /* --- y */
      compileDrop(AD_F,1);
    }
  return AD_OTHERS;
}
ACTION_END

DEF_ACTION_1ARG(bc_compose)
{ /* (f o g) e1 e2 ... en = f (g (e1)) e2 ... en, i.e.
   'o' (f,g) e1 e2 ... en = f (g e1) e2 ... en */
  ExprPtr e,ge1,arglist;

  if (AD_OTHERS==NIL)
    return tuple0; /* don't do anything special when n==0 */
  e = forgetExtra(AD_ARG);
  if (e->kind==(CHAR)E_TUPLE_2 || e->kind==(CHAR)E_UNSHARED_TUPLE_2)
    {
      ge1 = arglistCons (CDR (e), AD_ENV, arglistCons (arglistArg (AD_OTHERS),
						       arglistEnv (AD_OTHERS),
						       NIL));
      /* ge1 is g applied to e1, coded as an arglist. */
      arglist = arglistCons (ge1, NULL, arglistNext (AD_OTHERS));
      /* NULL as environment is here to signal that argument is an arglist
	 representing an application. */
      compileApply1 (AD_F, AD_ENV, CAR(e), arglist, AD_TAIL, AD_COMPILE, AD_PUSH, AD_WRONGP);
      return NIL;
    }
  return tuple0;
}
ACTION_END

DEF_MONOP(bc_ord,BC_ORD,true)
DEF_BINOP(bc_ordof,BC_ORDOF,true)
DEF_TEROP(bc_substr,BC_SUBSTR,true)

PRIVATE ExprPtr compilePrimitive(f,env,fun,arg_env,arg,otherargs,tail,compile,push,wrongp)
IN BCFILE *f;
IN CONST compEnv *env, *arg_env; /* env is the compEnv in which we must call fun,
				    arg_env the one in which we must compile arg.
				    */
IN CONST ExprPtr fun,arg,otherargs;
IN Boolean tail;
IN CONST compileFunc compile;
IN Boolean push;
INOUT ExprPtr wrongp;
{
  COUNTER level;

  level = env->level - VARLEVEL(fun);
  if (env->globenvs & (1L << level)) /* if we are calling a global function */
    {
      ExprPtr rank,ff;
      Boolean found;

      rank = Kint(VARRANK(fun));
      ff = MapGet(rank,primCompileMap,&found);
      if (found)
	{
	  applyData data;

	  data.f = f;
	  data.env = env;
	  data.fun = fun;
	  data.arg_env = arg_env;
	  data.arg = arg;
	  data.otherargs = otherargs;
	  data.tail = tail;
	  data.compile = compile;
	  data.push = push;
	  data.wrongp = wrongp;
	  return (*ACTION(ff))(ENV_CONS(QENV(ff), 1), (ExprPtr)&data);
	}
    }
  return tuple0;
}

PRIVATE void compileApply1(f,env,fun,arglist,tail,compile,push,wrongp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr fun,arglist;
IN Boolean tail;
IN CONST compileFunc compile;
IN Boolean push;
INOUT ExprPtr wrongp;
{
  ExprPtr arg;
  ENTER;

  if (arglist==NIL)
    {
      (*compile)(f,env,fun,wrongp);
      return;
    }
  /* identify special cases of applications: */
reapply:
  switch (fun->kind)
    {
    case E_APPL: case E_GOTO:
      /*case E_N_APPL:*/
      arglist = arglistCons (APPLARG(fun), env, arglist);
      fun = APPLFUN(fun);
      goto reapply;
    case E_PAT_AS:
      {
	compEnv *newenv = (compEnv *) alloctmp (sizeof (compEnv));

	*newenv = *env;
	if (!tail)
	  defineMe (newenv, NIL); /* disallow tail recursions from inlined body
				     if inlined function is not itself in tail position. */
	env = newenv;
	fun = CDR(fun);
	goto reapply;
      }
    case E_PROFILE:
      {
	ExprPtr saveprof;

	saveprof = CE_PROF(env);
	CE_PROF(env) = CDR(fun);
	fun = CAR(fun);
	compileApply1(f,env,fun,arglist,tail,compile,push,wrongp);
	CE_PROF(env) = saveprof;
      }
      break;
    case E_DEBUGZONE:
      fun = DZCODE(fun);
      goto reapply;
    case E_POLYFUN: /* case <expr> of ... */
      {
	ExprPtr e;
	COUNTER nargs, nformals;

	nargs = tryNargs (arglistArg(arglist));
	nformals = tryNformals (fun);
	if (nformals!=nargs)
	  nargs = 1; /* if function and argument do not agree
			on number of arguments, then allocate tuples. */
	{
	  struct Var saveme;

	  if (!tail)
	    {
	      saveme = env->me;
	      defineMe (env, NIL);
	    }
	  if (nargs==1)
	    {
	      compileExprOrAppl (f, arglistEnv (arglist), arglistArg (arglist));
	      arglist = arglistNext (arglist);
	      arglist = compilePolyFun1(f,env,fun,1,compile,(push?1:0),wrongp,arglist);
	    }
	  else
	    {
	      compileExprOrApplNargs (f, arglistEnv (arglist), arglistArg (arglist),
				      nargs);
	      /* we evaluate the argument in its associated environment,
		 not env.  Otherwise references to variables
		 run a chance of being wrong. */
	      arglist = arglistNext (arglist);
	      arglist = compilePolyFun1(f,env,fun,nargs,compile,(push?1:0),wrongp,arglist);
	    }
	  if (!tail)
	    env->me = saveme;
	}
	if (arglist!=NIL)
	  goto args;
	break;
      }
    case E_MEMO_FN: /* a memo function as immediate value in the code does not
		       memoize (or it would memoize into a closure that would
		       be immediately garbageable) */
      fun = MEMOFUN(fun);
      goto reapply;
    case E_DELAY: /* same as for memo functions */
      fun = DCODE(fun);
      goto reapply;
    case E_DATACON:
      if (!push)
	{ /* then evaluate first arg for its side-effects only. */
	  compileStmtOrAppl (f, arglistEnv (arglist), arglistArg (arglist));
	}
      else
	{
	  ExprPtr val;

	  val = COUNTEDVAL(fun);
	  if (val->kind==(CHAR)E_DATA) /* datatype constructor
					  (not generative) */
	    {
	      INT i;

	      compileExprOrAppl (f, arglistEnv (arglist), arglistArg (arglist));
	      bcPut((val->mark & MARK_SHARED)?BC_KDATA:BC_DATA,f);
	      bcPut((OPCODE)val->magic,f);
	      i = getConstant(f,CAR(val));
	      bcPut((OPCODE)i,f);
	    }
	  else /* exception constructor, generative or not. */
	    {
	      if (COUNTEDN(fun) & 2) /* generative */
		compileExpr(f,env,val);
	      else
		compileConstant(f,val);
	      compileExprOrAppl (f, arglistEnv (arglist), arglistArg (arglist));
	      bcPut(BC_CONS,f);
	      bcPush(-1,f);
	    }
	}
      /* Now, because of the types, arglistNext (arglist) must be NIL. */
      break;
    case E_TUPLE_2:
      if (CAR(fun)->kind==(CHAR)E_BYTECODE)
	{
	callbind:
	  /* Always applied to NIL, so as to install a few new definitions. */
	  compileCurIEnv(f,env);
	  compileExprOrAppl (f, arglistEnv (arglist), arglistArg (arglist));
	  bcPut(BC_CALLBIND,f);
	  bcPut((OPCODE)getConstant(f,fun),f);
	  bcPush(-1,f);
	  {
	    ExprPtr normp;

	    normp = beginForward(f);
	    compileGoto(f,BC_IF,normp,1);
	    compileRaiseImm(f,env,I_BINDEXC); /* raise Bind */
	    compileDrop(f,1);
	    endForward(f,normp);
	    if (push)
	      compileConstant(f,NIL); /* dummy */
	  }
	  /* Now, because of the way modules are built, arglistNext (arglist) is NIL. */
	  break;
	}
      goto normal;
    case E_COMPILED_CLOSURE:
      goto callbind;
    case E_VARREF:
      if (tail && isMe(env,fun) && arglistNext (arglist)==NIL)
	{ /* deal with tail recursion: */
	  COUNTER savelevel;

	  savelevel = f->stackDepth;
	  compileExprOrApplNargs (f, arglistEnv (arglist),
				  arglistArg (arglist), f->nargs);
	  /* --- env old-arg1 ... old-argn <...> new-arg1 ... new-argn */
	  switch (f->nargs)
	    {
	    case 0: /* n==0 */
	      compileDrop (f, f->stackDepth-1); /* --- env */
	      break;
	    case 1: /* n==1 */
	      compileSwDrop(f,f->stackDepth-1); /* --- env new-arg1 */
	      break;
	    default: /* n>=2 */
	      if (f->stackDepth>1+f->nargs)
		{
		  bcPut(BC_NSWDROP,f);
		  bcPut((OPCODE)(-f->stackDepth+1), f);
		  bcPut((OPCODE)(-f->nargs), f);
		}
	      break;
	    }
	  bcPut(BC_GOTO_0,f);
	  bcPut((OPCODE)0,f);
	  bcPut((OPCODE)(1-bcTell(f)),f); /* and goto start of function */
	  f->dead = 1;
	  f->stackDepth = (push?(savelevel+1):savelevel);
	  break;
	}
      /* identify some special cases for known primitives. */
      {
	ExprPtr remaining;
	  
	remaining = compilePrimitive(f,env,fun,
				     arglistEnv (arglist), arglistArg (arglist),
				     arglistNext (arglist),tail,compile,push,wrongp);
	if (remaining!=tuple0)
	  {
	    if (remaining==NIL)
	      break;
	    arglist = remaining;
	    goto args;
	  }
      }
      /*fallthrough*/
normal:
    default:
      compileExpr(f,env,fun);
args:
      while (arglist!=NIL)
	{
	  COUNTER nargs;

	  nargs = tryNargs (arglistArg (arglist));
	  if (nargs==1 || nargs>MAX_ARITY) /* ordinary call, or call with arity too high. */
	    {
	      compileExprOrAppl (f, arglistEnv (arglist), arglistArg (arglist));
	      compileDebugZone(true,f);
	      compileCodePointOp(BC_CALL,f);
	      bcPush(-1,f);
	    }
	  else
	    {
	      compileExprOrApplNargs (f, arglistEnv (arglist), arglistArg (arglist), nargs);
	      compileDebugZone (true, f);
	      compileCodePointOp(BC_N_CALL,f);
	      bcPut((OPCODE)nargs,f);
	      bcPush (-nargs, f);
	    }
	  arglist = arglistNext (arglist);
	}
      /* tail calls that are not tail-recursive do not bring much;
	 on the other hand, they are more complicated than ordinary
	 calls to interpret, and they are impossible to compile
	 to C.  So we don't use the BC_TAILCALL opcode. */
      if (!push)
	{
	  if (wrongp!=NULL)
	    compileGoto(f,BC_IF,wrongp,1);
	  else
	    compileDrop(f,1);
	}
      break;
    }
  QUIT;
}

PRIVATE void compileApply(f,env,fun,arg,tail,compile,push,wrongp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr fun,arg;
IN Boolean tail;
IN CONST compileFunc compile;
IN Boolean push;
INOUT ExprPtr wrongp;
{
  ExprPtr arglist;
  ENTER;

  arglist = arglistCons (arg, env, NIL);
  compileApply1(f,env,fun,arglist,tail,compile,push,wrongp);
  QUIT;
}

PRIVATE void compileMapGet(f,env,m,x)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr m,x;
{
  ExprPtr errp,finp;
  COUNTER level;

  level = f->stackDepth;
  compileExpr(f,env,m);
  compileExpr(f,env,x);
  errp = beginForward(f);
  compileGoto(f,BC_MAPGET,errp,1);
  finp = beginForward(f);
  compileGoto(f,BC_GOTO,finp,0);
  setStackDepth(f,level);
  endForward(f,errp);
  compileRaiseImm(f,env,I_MAPGETEXC);
  endForward(f,finp);
}

PRIVATE void compileTupleSelect(f,index)
IN BCFILE *f;
IN CONST ExprPtr index;
{
  ExprPtr t2pp,finp;

  finp = beginForward(f);
  if (IVAL(index)<=2)
    {
      /* if index>2, then val must be an E_TUPLE_GNRL, because of typing.
       Otherwise, it must be a E_TUPLE_2, since there are no E_TUPLE_1s. */
      t2pp = beginForward(f);
      compileGoto(f,BC_ISTUPLE2,t2pp,0);
      /* if val->kind==(CHAR)E_TUPLE_2: */
      bcPut((index==ione)?BC_CAR:BC_CDR,f);
      compileGoto(f,BC_GOTO,finp,0);
      endForward(f,t2pp);
    }
  /* otherwise, if val is an E_TUPLE_GNRL: */
  compileNth(f,IVAL(index)-1);
  if (isForward(finp))
    {
      endForward(f,finp);
    }
}

PRIVATE void compileRecordSelect(f,env,lbl)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr lbl;
{
  ExprPtr r1pp,finp;

  finp = beginForward(f);
  if (lbl==ione || lbl->kind!=(CHAR)E_INT)
    { /* then record may be a E_RECORD_1 */
      r1pp = beginForward(f);
      compileGoto(f,BC_ISRECORD1,r1pp,0);
      bcPut(BC_CDR,f);
      compileGoto(f,BC_GOTO,finp,0);
      endForward(f,r1pp);
    }
  /* otherwise, val is a E_RECORD_GNRL: */
  compileSelectGnrl(f,env,lbl);
  if (isForward(finp))
    {
      endForward(f,finp);
    }
}

PRIVATE void compileMu1(f,env,rec,modlab,modval,share)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr rec,modlab,modval;
IN Boolean share;
{
  compileExpr(f,env,rec);
  compileExpr(f,env,modval);
  bcPut((share?BC_KMU1:BC_MU1),f);
  bcPut((OPCODE)getConstant(f,modlab),f);
  bcPush(-1,f);
}

PRIVATE void compileMuGnrl(f,env,rec,mod,share)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr rec,mod;
IN Boolean share;
{
  compileExpr(f,env,rec);
  compileExpr(f,env,mod);
  bcPut(share?BC_KMUGNRL:BC_MUGNRL,f);
  bcPush(-1,f);
}

#define compileNeutralNone ((compileNeutralFunc)0L)
PRIVATE void compileCombine1None(f,env)
IN BCFILE *f;
IN CONST compEnv *env;
{
  return;
}
PRIVATE void compileCombineNone(f,env,l1)
IN BCFILE *f;
IN CONST compEnv *env;
IN COUNTER l1;
{
  return;
}
PRIVATE void compileFinalizeNone(f,env)
IN BCFILE *f;
IN CONST compEnv *env;
{
  return;
}

#define compileNeutralExists compileNeutralNone
#define compileCombine1Exists compileCombine1None
#define compileCombineExists compileCombineNone
PRIVATE void compileEvalPromoteExists(f,env,body,errlevel,errmaplevel,errp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr body;
IN COUNTER errlevel,errmaplevel;
INOUT ExprPtr errp;
{
  ExprPtr contp;

  contp = beginForward(f);
  compileTest(f,env,body,contp);
  compileMapDrop(f,errmaplevel);
  compileGoto(f,BC_GOTO,errp,0); /* if true, then exit */
  endForward(f,contp); /* otherwise, go on searching */
}
#define compileFinalizeExists compileFinalizeNone

#define compileNeutralAll compileNeutralNone
#define compileCombine1All compileCombine1None
#define compileCombineAll compileCombineNone
PRIVATE void compileEvalPromoteAll(f,env,body,errlevel,errmaplevel,errp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr body;
IN COUNTER errlevel,errmaplevel;
INOUT ExprPtr errp;
{
  if (f->mapStackDepth==errmaplevel)
    { /* the simple case: we have nothing on the mapstack to drop */
      compileTest(f,env,body,errp); /* if false, then exit */
    }
  else
    {
      ExprPtr exitp,contp;

      exitp = beginForward(f);
      compileTest(f,env,body,exitp);
      contp = beginForward(f);
      compileGoto(f,BC_GOTO,contp,0);
      endForward(f,exitp);
      compileMapDrop(f,errmaplevel);
      compileGoto(f,BC_GOTO,errp,0);
      endForward(f,contp);
    }
  /* otherwise, go on searching */
}
#define compileFinalizeAll compileFinalizeNone

PRIVATE void compileNeutralSome(f,env)
IN BCFILE *f;
IN CONST compEnv *env;
{
  compileConstant(f,none);
}
#define compileCombine1Some compileCombine1None
#define compileCombineSome compileCombineNone
PRIVATE void compileEvalPromoteSome(f,env,body,errlevel,errmaplevel,errp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr body;
IN COUNTER errlevel,errmaplevel;
INOUT ExprPtr errp;
{
  compileExpr(f,env,body);
  bcPut(BC_KDATA,f);
  bcPut((OPCODE)0,f); /* magic=0 for SOME */
  bcPut((OPCODE)getConstant(f,optttm),f);
  compilePushDrop(f,errlevel,true);
  compileMapDrop(f,errmaplevel);
  compileGoto(f,BC_GOTO,errp,0); /* exit the loop, with SOME(result) on top of stack */
}
#define compileFinalizeSome compileFinalizeNone

#define compileNeutralUnsharedSome compileNeutralSome
#define compileCombine1UnsharedSome compileCombine1None
#define compileCombineUnsharedSome compileCombineNone
PRIVATE void compileEvalPromoteUnsharedSome(f,env,body,errlevel,errmaplevel,errp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr body;
IN COUNTER errlevel,errmaplevel;
INOUT ExprPtr errp;
{
  compileExpr(f,env,body);
  bcPut(BC_DATA,f);
  bcPut((OPCODE)0,f); /* magic=0 for SOME */
  bcPut((OPCODE)getConstant(f,optttm),f);
  compilePushDrop(f,errlevel,true);
  compileMapDrop(f,errmaplevel);
  compileGoto(f,BC_GOTO,errp,0); /* exit the loop, with SOME(result) on top of stack */
}
#define compileFinalizeUnsharedSome compileFinalizeNone

PRIVATE void compileExprErr(f,env,body,errlevel,errmaplevel,errp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr body;
IN COUNTER errlevel,errmaplevel;
INOUT ExprPtr errp;
{
  compileExpr(f,env,body);
}

#define compileNeutralProg compileNeutralNone
#define compileCombine1Prog compileCombine1None
#define compileCombineProg compileCombineNone
PRIVATE void compileEvalPromoteProg(f,env,body,errlevel,errmaplevel,errp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr body;
IN COUNTER errlevel,errmaplevel;
INOUT ExprPtr errp;
{
  compileStmt(f,env,body);
}
#define compileFinalizeProg compileFinalizeNone

PRIVATE void compileNeutralList(f,env)
IN BCFILE *f;
IN CONST compEnv *env;
{
  compileConstant(f,NIL);
}
#define compileCombine1List compileCombine1None
PRIVATE void compileCombineList(f,env,l1)
IN BCFILE *f;
IN CONST compEnv *env;
IN COUNTER l1;
{ /* builds a queue of of all elements in the final list in order. */
  compileFetchOp(f,BC_CONCATQUEUE,0,l1);
  bcPush(-1,f);
}
PRIVATE void compileEvalPromoteList(f,env,body,errlevel,errmaplevel,errp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr body;
IN COUNTER errlevel,errmaplevel;
INOUT ExprPtr errp;
{
  compileExpr(f,env,body);
  /* builds a queue with only one element in it */
  bcPut(BC_NEW1QUEUE,f);
}
PRIVATE void compileFinalizeList(f,env)
IN BCFILE *f;
IN CONST compEnv *env;
{
  bcPut(BC_QUEUELIST,f);
}

#define compileNeutralUnsharedList compileNeutralList
#define compileCombine1UnsharedList compileCombine1List
#define compileCombineUnsharedList compileCombineList
#define compileEvalPromoteUnsharedList compileEvalPromoteList
PRIVATE void compileFinalizeUnsharedList(f,env)
IN BCFILE *f;
IN CONST compEnv *env;
{
  bcPut(BC_QUEUEUNSHAREDLIST,f);
}

PRIVATE void compileNeutralSet(f,env)
IN BCFILE *f;
IN CONST compEnv *env;
{
  compileConstant(f,NewMap());
}
#define compileCombine1Set compileCombine1None
PRIVATE void compileCombineSet(f,env,l1)
IN BCFILE *f;
IN CONST compEnv *env;
IN COUNTER l1;
{
  compileFetch(f,l1);
  bcPut(BC_UNDERWRITE,f);
  bcPush(-1,f);
}
PRIVATE void compileEvalPromoteSet(f,env,body,errlevel,errmaplevel,errp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr body;
IN COUNTER errlevel,errmaplevel;
INOUT ExprPtr errp;
{
  compileExpr(f,env,body);
  compileConstant(f,tuple0);
  bcPut(BC_MAPLET,f);
  bcPush(-1,f);
}
#define compileFinalizeSet compileFinalizeNone

#define compileNeutralAppend compileNeutralList
#define compileCombine1Append compileCombine1List
#define compileCombineAppend compileCombineList
PRIVATE void compileEvalPromoteAppend(f,env,body,errlevel,errmaplevel,errp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr body;
IN COUNTER errlevel,errmaplevel;
INOUT ExprPtr errp;
{
  compileExpr(f,env,body);
  bcPut(BC_LISTQUEUE,f);
}
#define compileFinalizeAppend compileFinalizeList

#define compileNeutralUnsharedAppend compileNeutralList
#define compileCombine1UnsharedAppend compileCombine1List
#define compileCombineUnsharedAppend compileCombineList
#define compileEvalPromoteUnsharedAppend compileEvalPromoteAppend
#define compileFinalizeUnsharedAppend compileFinalizeUnsharedList

#define compileNeutralUnion compileNeutralSet
#define compileCombine1Union compileCombine1None
#define compileCombineUnion compileCombineSet
#define compileEvalPromoteUnion compileExprErr
#define compileFinalizeUnion compileFinalizeNone

#define compileNeutralUnionUnder compileNeutralSet
#define compileCombine1UnionUnder compileCombine1None
PRIVATE void compileCombineUnionUnder(f,env,l1)
IN BCFILE *f;
IN CONST compEnv *env;
IN COUNTER l1;
{
  compileFetch(f,l1);
  bcPut(BC_OVERWRITE,f);
  bcPush(-1,f);
}
#define compileEvalPromoteUnionUnder compileEvalPromoteUnion
#define compileFinalizeUnionUnder compileFinalizeNone

PRIVATE void compileNeutralInter(f,env)
IN BCFILE *f;
IN CONST compEnv *env;
{
  compileRaiseImm(f,env,I_EMPTYEXC);
  /* we can do this because we never compute distributed intersections
     by intersecting all sets with the common neutral, but actually
     return the neutral only if there are no sets to intersect
     (i.e, data->simplep must be false). */
}
#define compileCombine1Inter compileCombine1None
PRIVATE void compileCombineInter(f,env,l1)
IN BCFILE *f;
IN CONST compEnv *env;
IN COUNTER l1;
{
  compileFetch(f,l1);
  bcPut(BC_MAPINTER,f);
  bcPush(-1,f);
  /* we do not need to check that e1 is NIL or not, as in eval.c,
     since we never intersect with something representing the whole
     universe of sets (we do not use the neutral element when there
     is at least one set to intersect). */
}
#define compileEvalPromoteInter compileExprErr
#define compileFinalizeInter compileFinalizeNone

#define compileNeutralMap compileNeutralSet
#define compileCombine1Map compileCombine1None
#define compileCombineMap compileCombineSet
#define compileEvalPromoteMap compileExprErr
#define compileFinalizeMap compileFinalizeNone

#define compileNeutralMapUnder compileNeutralSet
#define compileCombine1MapUnder compileCombine1None
#define compileCombineMapUnder compileCombineUnionUnder
#define compileEvalPromoteMapUnder compileExprErr
#define compileFinalizeMapUnder compileFinalizeNone

PRIVATE void compileSimpleOf P((IN BCFILE *f,IN CONST compEnv *env,
				IN struct compilesimpleof *data,
				INOUT ExprPtr failp,
				IN COUNTER errlevel,IN COUNTER errmaplevel,
				INOUT ExprPtr errp,IN COUNTER reslevel));

PRIVATE void compileOneOf(f,env,_data,failp,errlevel,errmaplevel,errp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CHAR *_data;
INOUT ExprPtr failp; /* merge node of path that comes out of a subiteration
			that did nothing (pattern failed, test failed, ...) */
IN COUNTER errlevel,errmaplevel;
INOUT ExprPtr errp; /* merge node of path that comes out of the whole iteration. */
{
  COUNTER level;
  struct compilesimpleof *data;

  data = (struct compilesimpleof *)_data;
  level = f->stackDepth;
  if (data->mapsweep)
    (void) compileMapletPattern(f,env,data->pat,failp,NIL);
  else
    {
      if (nontrivMatch(data->pat))
	{
	  if (level!=f->stackDepth)
	    {
	      (void) compilePatternAt(f,env,data->pat,level-1,failp,NIL,false);
	      /*compileFetch(f,level-1);*/
	    }
	  else
	    (void) compilePattern(f,env,data->pat,failp,NIL,false);
	}
    }
  /* choicelast should now be NIL again */
  if (data->domains==NIL)
    {
      Boolean saveflags;
#if DEBUGGING!=0
      INT dstart;
      bcdState st;
#endif

#if DEBUGGING!=0
      dstart = beginDebug(f,&st);
#endif
      saveflags = env->flags;
      if ((env->flags & CEF_NEEDENV) && !(needEnvBindings(data->locals,env)
					  || needEnv(data->pred,env)
					  || needEnv(data->body,env)))
	env->flags &= ~CEF_NEEDENV;
      if (env->flags & CEF_NEEDENV)
	{
	  env->level--;
	  compileCurEnv(f,env); /* --- <...> env (if needenv); --- <...> (if !needenv) */
	  env->level++;
	  bcPut(BC_CONSENV,f);
	  bcPut((OPCODE)data->nlocals,f);
	  newCurEnv(env,f->stackDepth-1);
	  dumpLetStores(f,f->stackDepth-1,CE_LETDELAY(env));
	}
      /* --- <...> env (if needenv); --- <...> (if !needenv) */
      CE_LETDELAY(env) = NewMap();
      /* This is what we have called someplace else 'letconsenv trickery':
	 we only allocate the environment after all patterns have been
	 matched, and then we fill it in with the values we have gotten
	 (and only if such an environment is needed). */
      flushDbgNames(f,env);

      compileBindings(f,env,data->locals,failp);
      compileTest(f,env,data->pred,failp);
      (*data->evalpromote)(f,env,data->body,errlevel,errmaplevel,errp);
      env->flags = saveflags;
#if DEBUGGING!=0
      endDebug(dstart,f,&st);
#endif
    }
  else
    compileSimpleOf(f,env,data,failp,errlevel,errmaplevel,errp,level);
}

PRIVATE void startSweep(f,kind,list,swip)
IN BCFILE *f;
IN INT kind;
IN INT list;
OUT struct sweepinfo *swip;
{
  switch (kind)
    {
    case E_PAT_RECORD_GNRL_CONS: /* in list domain descriptor */
      bcPut(BC_STARTLISTSWEEP,f);
      bcPut((OPCODE)(list-f->stackDepth),f);
      swip->listref = f->stackDepth;
      bcPush(1,f);
      break;
    case E_PAT_REVRECORD_GNRL_CONS: /* in map domain descriptor */
      bcPut(BC_STARTMAPSWEEP,f);
      bcPut((OPCODE)(list-f->stackDepth),f);
      swip->listref = f->stackDepth;
      swip->auxref = f->mapStackDepth;
      bcPush(1+MAP_CHAIN_LENGTH,f);
      bcMapPush(2,f);
      break;
    case E_PAT_RECORD_MAPADD: /* sub map domain descriptor */
      bcPut(BC_STARTSUBMAPSWEEP,f);
      bcPut((OPCODE)(list-f->stackDepth),f);
      swip->listref = f->stackDepth;
      swip->auxref = f->stackDepth+1;
      bcPush(2,f);
      break;
    }
}

PRIVATE void getSweep(f,kind,swip)
IN BCFILE *f;
IN INT kind;
IN struct sweepinfo *swip;
{
  switch (kind)
    {
    case E_PAT_RECORD_GNRL_CONS: /* in list domain descriptor */
      bcPut(BC_GETLISTSWEEP,f);
      break;
    case E_PAT_REVRECORD_GNRL_CONS: /* in map domain descriptor */
      bcPut(BC_GETMAPSWEEP,f);
      break;
    case E_PAT_RECORD_MAPADD: /* sub map domain descriptor */
      bcPut(BC_GETSUBMAPSWEEP,f);
      break;
    }
  bcPut((OPCODE)(swip->listref-f->stackDepth),f);
  bcPush(1,f);
}

#ifdef OBSOLETE
PRIVATE void getCarCdr(f,kind,xp,yp)
IN BCFILE *f;
IN INT kind;
OUT COUNTER *xp,*yp;
{
  switch (kind)
    {
    case E_PAT_RECORD_GNRL_CONS: /* in list domain descriptor */
    case E_PAT_RECORD_MAPADD: /* sub map domain descriptor */
      *xp = f->stackDepth-1;
      *yp = -1; /* dummy */
      break;
    case E_PAT_REVRECORD_GNRL_CONS: /* in map domain descriptor */
      bcPut(BC_CDRCAR,f);
      bcPush(1,f);
      *xp = f->stackDepth-1;
      *yp = f->stackDepth-2;
      break;
    }
}
#endif

PRIVATE void incSweep(f,kind,swip,loopp)
IN BCFILE *f;
IN INT kind;
IN CONST struct sweepinfo *swip;
INOUT ExprPtr loopp;
{
  switch (kind)
    {
    case E_PAT_RECORD_GNRL_CONS: /* in list domain descriptor */
      compileGoto(f,BC_INCLISTSWEEP,loopp,0);
      bcPut((OPCODE)(swip->listref-f->stackDepth),f);
      break;
    case E_PAT_REVRECORD_GNRL_CONS: /* in map domain descriptor */
      compileGoto(f,BC_INCMAPSWEEP,loopp,0);
      bcPut((OPCODE)(swip->listref-f->stackDepth),f);
      bcPut((OPCODE)(f->mapStackDepth-swip->auxref),f);
      break;
    case E_PAT_RECORD_MAPADD: /* sub map domain descriptor */
      compileGoto(f,BC_INCSUBMAPSWEEP,loopp,0);
      bcPut((OPCODE)(swip->listref-f->stackDepth),f);
      break;
    }
}

PRIVATE void compileSweep(f,env,tag,list,combine,combine1,neutral,body,
			  data,failp,errlevel,errmaplevel,
			  errp,reslevel,resclass,simplep)
IN BCFILE *f;
IN CONST compEnv *env;
IN INT tag;
IN INT list;
IN compileCombineFunc combine;
IN compileCombine1Func combine1;
IN compileNeutralFunc neutral; /* only used is simplep is true */
IN compileBodyFunc body;
IN CHAR *data;
INOUT ExprPtr failp;
IN COUNTER errlevel,errmaplevel;
INOUT ExprPtr errp;
IN COUNTER reslevel;
IN COUNTER resclass;
IN Boolean simplep;
{
  ExprPtr outp;
  struct sweepinfo swi;

  {
    COUNTER level;

    level = f->stackDepth;
    outp = beginForward(f);
    /* first, test whether list is NIL */
    if (list!=f->stackDepth-1)
      compileFetch(f,list); /* --- list */
    switch (tag)
      {
      case E_PAT_RECORD_GNRL_CONS: /* in list domain descriptor */
	{
	  ExprPtr contp;

	  contp = beginForward(f);
	  compileGoto(f,BC_ISNIL,contp,0);
	  compileGoto(f,BC_GOTO,failp,0);
	  setStackDepth(f,level);
	  endForward(f,contp); /* --- */
	}
	break;
      case E_PAT_REVRECORD_GNRL_CONS: /* in map domain descriptor */
	{
	  ExprPtr contp;

	  contp = beginForward(f);
	  compileGoto(f,BC_ISEMPTY,contp,0);
	  compileGoto(f,BC_GOTO,failp,0);
	  setStackDepth(f,level);
	  endForward(f,contp); /* --- */
	}
	break;
      case E_PAT_RECORD_MAPADD: /* sub map domain descriptor; never empty */
	{
	  ExprPtr elsep;

	  elsep = beginForward(f);
	  compileGoto(f,BC_ISEMPTY,elsep,0);
	  /* if list is empty, just compute body on empty set */
	  /* compute func on empty set */
	  /* --- empty */
	  (*body)(f,env,data,failp,errlevel,errmaplevel,errp);
	  /* --- <...> res (if !quantifier); --- <...> (if quantifier) */
	  /* otherwise, return result */
	  (*combine1)(f,env);
	  compilePushDrop(f,reslevel,(resclass==0));
	  compileGoto(f,BC_GOTO,outp,0);
	  
	  setStackDepth(f,level);
	  endForward(f,elsep); /* --- */
	}
	break;
      }
  }
  {
    COUNTER looplevel;

    startSweep(f,tag,list,&swi); /* --- <loop-data> (may take more than one stack slot) */
    looplevel = f->stackDepth;
    if (simplep)
      {
	if (neutral==NULL)
	  ;
	else
	  (*neutral) (f, env); /* --- <loop-data> [res] */
      }
    else
      /* compute function on first element of list that satisfies the predicate: */
      {
	ExprPtr startp,nonep;

	startp = beginBackward(f); /* start loop; --- <loop-data> */
	getSweep(f,tag,&swi); /* --- <loop-data> elt */
	nonep = beginForward(f);
	(*body)(f,env,data,nonep,errlevel,errmaplevel,errp);
	(*combine1)(f,env); /* --- <loop-data> <...> res (if !quantifier);
			       --- <loop-data> <...> (if quantifier) */
	compilePushDrop(f,looplevel,(resclass==0)); /* --- <loop-data> [res] */
	if (isForward(nonep))
	  { /* If pattern fails to match, or predicate returns false, then loop
	       until pattern matches and predicate returns true. */
	    ExprPtr contp;
	    COUNTER savelevel;

	    savelevel = f->stackDepth;
	    contp = beginForward(f);
	    compileGoto(f,BC_GOTO,contp,0);
	    setStackDepth(f,looplevel);
	    endForward(f,nonep); /* --- <loop-data> */
	    if (tag==(CHAR)E_PAT_REVRECORD_GNRL_CONS) /* in map */
	      { /* then if incSweep fails, we first have to drop the element on top of mapstack */
		ExprPtr fail2p;

		fail2p = beginForward(f);
		incSweep(f,tag,&swi,fail2p); /* --- <loop-data> */
		compileGoto(f,BC_GOTO,startp,0);
		endForward(f,fail2p);
		compileMapDrop(f,f->mapStackDepth-2);
		compileGoto(f,BC_GOTO,failp,0);
	      }
	    else
	      {
		incSweep(f,tag,&swi,failp); /* --- <loop-data> */
		compileGoto(f,BC_GOTO,startp,0);
	      }
	    setStackDepth(f,savelevel);
	    endForward(f,contp); /* --- <loop-data> [res] (res is there iff !quantifier) */
	  }
	/* --- <loop-data> [first-res] */
	/* where <loop-data> is described by swi, and may take more than
	   one slot on the stack. looplevel is the level at which first-res is found. */
      }
    { /* Now, compile the real while loop. */
      ExprPtr startp,exitp,enterp = NOEXPR;

      if (simplep)
	{
	  Boolean olddead;

	  enterp = beginForward(f);
	  olddead = f->dead;
	  compileGoto(f,BC_GOTO,enterp,0);
	  f->dead = olddead;
	}
      exitp = beginForward(f);
      startp = beginBackward(f); /* start loop; --- <loop-data> [old-res] */
      incSweep(f,tag,&swi,exitp);
      if (simplep)
	endForward(f,enterp);

      getSweep(f,tag,&swi); /* --- <loop-data> [old-res] elt */
      (*body)(f,env,data,startp,errlevel,errmaplevel,errp);
      /* --- <loop-data> [old-res] <...> [res] */
      (*combine)(f,env,looplevel);
      /* --- <loop-data> [old-res] <...> [new-res] */
      compilePushDrop(f,looplevel,(resclass==0));
      /* --- <loop-data> [new-res] */
      compileGoto(f,BC_GOTO,startp,0);

      setStackDepth(f,looplevel+(resclass==0));
      endForward(f,exitp);
      /* --- <loop-data> [res] */

    }
    if (tag==(CHAR)E_PAT_REVRECORD_GNRL_CONS) /* in map domain descriptor */
      { /* then drop the elements on top of mapstack */
	compileMapDrop(f,f->mapStackDepth-2);
	bcMapPush(-2,f);
      }
  }
  /* And get the final result: */
  if (isForward(outp))
    {
      compilePushDrop(f,reslevel,(resclass==0)); /* --- [res] */
      endForward(f,outp);
    } /* otherwise, return --- <...> [res] */
}

DEF_PRIVATE_ACTION_1ARG(compilemapaddpattern)
{
  struct compilemapaddpattern *data = (struct compilemapaddpattern *)ENV_UNDERLYING(env);

  return compileMapAddPatternGnrl1(data->f,data->env,data->pat,
				   data->nomatchp,data->choicelist,
				   data->emptyp);
}
ACTION_END

PRIVATE ExprPtr compileMapAddPatternGnrlk(f,env,pat,nomatchp,choicelist,emptyp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr pat;
INOUT ExprPtr nomatchp;
IN CONST ExprPtr choicelist;
INOUT ExprPtr emptyp;
{
  struct compilemapaddpattern *data = (struct compilemapaddpattern *)
    alloctmp(sizeof(struct compilemapaddpattern));

  data->f = f;
  data->env = env;
  data->pat = pat;
  data->nomatchp = nomatchp;
  data->choicelist = choicelist;
  data->emptyp = emptyp;
  return CallCC(ACTION_NAME(compilemapaddpattern),(ExprPtr)data,true);
}

PRIVATE ExprPtr compileMapAddPatternGnrl1(f,env,pat,nomatchp,choicelist,emptyp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr pat;
INOUT ExprPtr nomatchp; /* where to go in case the pattern does not match, but more:
			   there is no hope that backtracking will make it match either. */
IN CONST ExprPtr choicelist;
INOUT ExprPtr emptyp; /* where to go in case the pattern does not match. */
{
  /* now, pat = {x => y} ++ rest: */

  if (STKFULL(f))
    return compileMapAddPatternGnrlk(f,env,pat,nomatchp,choicelist,emptyp);
again:
  switch (pat->kind)
    {
    case E_EMPTY:
      compileGoto(f,BC_ISEMPTY,nomatchp,0);
      break;
    case E_PAT_MAPADD:
      if (PATMAPADDREST(pat)==NewMap())
	{
	  pat = PATMAPADDXY(pat);
	  compileGoto(f,BC_ISMAPLET,nomatchp,0);
	  if (nontrivMatch(CAR(pat)) || nontrivMatch(CDR(pat)))
	    {
	      compileFetch(f,f->stackDepth-1);
	      choicelist = compileMapletPattern(f,env,pat,emptyp,choicelist);
	    }
	}
      else
	{
	  ExprPtr contp,failp,errp,successp;
	  struct sweepinfo swi;
	  ExprPtr xypat;
	  COUNTER level,looplevel,endlevel;
	  ExprPtr val;

	  /* --- map */
	  level = f->stackDepth-1;
	  xypat = PATMAPADDXY(pat);
	  if (cstPattern(CAR(xypat),&val))
	    {
	      compileFetch(f,level); /* --- map map */
	      compileConstant(f,val); /* --- map map cst */
	      compileGoto(f,BC_MAPGET,emptyp,1); /* --- map y */
	      choicelist = compilePattern(f,env,CDR(xypat),emptyp,choicelist,false);
	      if (nontrivMatch(PATMAPADDREST(pat)))
		{
		  compileConstant(f,val);
		  compileFetch(f,level);
		  bcPut(BC_MAPREMOVE,f);
		  bcPush(-1,f);
		  pat = PATMAPADDREST(pat);
		  goto again;
		}
	    }
	  else
	    {
	      contp = beginForward(f);
	      compileGoto(f,BC_ISEMPTY,contp,0);
	      compileGoto(f,BC_GOTO,nomatchp,0);
	      endForward(f,contp);
	      /* --- map */
	      startSweep(f,E_PAT_REVRECORD_GNRL_CONS,level,&swi);
	      /* --- map <loop-data> */
	      contp = beginBackward(f);
	      looplevel = f->stackDepth;
	      failp = beginForward(f);
	      errp = beginForward(f);
	      successp = beginForward(f);
	      if (nontrivMatch(CAR(xypat)) || nontrivMatch(CDR(xypat)))
		{
		  getSweep(f,E_PAT_REVRECORD_GNRL_CONS,&swi); /* --- map <loop-data> m */
		  (void) compileMapletPattern(f,env,xypat,errp,choicelist);
		  /* Because we don't share PAT_MAPADD patterns, choicelist
		     does not change in the whole function. */
		}
	      if (nontrivMatch(PATMAPADDREST(pat)))
		{
		  getSweep(f,E_PAT_REVRECORD_GNRL_CONS,&swi);
		  bcPut(BC_CAR,f);
		  compileFetch(f,level);
		  bcPut(BC_MAPREMOVE,f);
		  bcPush(-1,f); /* --- map <loop-data> <...> map\{x} */
		  (void) compileMapAddPatternGnrl1(f,env,PATMAPADDREST(pat),failp,choicelist,
						   errp);
		}
	      endlevel = f->stackDepth;
	      compileGoto(f,BC_GOTO,successp,0);
	      if (isForward(errp))
		{
		  setStackDepth(f,looplevel);
		  endForward(f,errp);
		}
	      errp = beginForward(f);
	      incSweep(f,E_PAT_REVRECORD_GNRL_CONS,&swi,errp);
	      compileGoto(f,BC_GOTO,contp,0);
	      setStackDepth(f,level+1);
	      if (isForward(failp))
		{
		  endForward(f,failp);
		  compileMapDrop(f,f->mapStackDepth-2);
		  compileGoto(f,BC_GOTO,nomatchp,0);
		}
	      if (isForward(errp))
		{
		  endForward(f,errp);
		  compileMapDrop(f,f->mapStackDepth-2);
		  compileGoto(f,BC_GOTO,emptyp,0);
		}
	      if (isForward(successp))
		{
		  setStackDepth(f,endlevel);
		  endForward(f,successp);
		  compileMapDrop(f,f->mapStackDepth-2);
		   /* this is slightly incorrect:
		      it is correct to drop the map stack, however,
		      we cannot drop the value stack as well, as it
		      should be the case when we drop the map stack.
		      But it's better than not dropping the map stack
		      at all. */
		}
	      bcMapPush(-2,f);
	    }
	}
      break;
    default:
      choicelist = compilePattern(f,env,pat,nomatchp,choicelist,false);
      /* this pattern cannot fail; in fact, it must be a variable pattern or an ellipsis. */
      break;
    }
  return choicelist;
}

DEF_PRIVATE_ACTION_1ARG(compilesimpleofk)
{
  struct compilesimpleofk *data;

  data = (struct compilesimpleofk *)ENV_UNDERLYING(env);
  compileSimpleOf(data->f,data->env,data->data,
		  data->failp,data->errlevel,data->errmaplevel,data->errp,
		  data->reslevel);
  return NIL;
}
ACTION_END

PRIVATE ExprPtr compileSimpleOf1(f,env,data,failp,errlevel,errmaplevel,errp,reslevel)
IN BCFILE *f;
IN CONST compEnv *env;
IN struct compilesimpleof *data;
INOUT ExprPtr failp;
IN COUNTER errlevel,errmaplevel;
INOUT ExprPtr errp;
IN COUNTER reslevel;
{
  struct compilesimpleofk *d =
    (struct compilesimpleofk *)alloctmp(sizeof(struct compilesimpleofk));

  d->f = f;
  d->env = env;
  d->data = data;
  d->failp = failp;
  d->errlevel = errlevel;
  d->errmaplevel = errmaplevel;
  d->errp = errp;
  d->reslevel = reslevel;
  return CallCC(ACTION_NAME(compilesimpleofk),(ExprPtr)d,true);
}

PRIVATE void compileSimpleOf(f,env,data,failp,errlevel,errmaplevel,errp,reslevel)
IN BCFILE *f;
IN CONST compEnv *env;
IN struct compilesimpleof *data;
INOUT ExprPtr failp;
IN COUNTER errlevel,errmaplevel;
INOUT ExprPtr errp;
IN COUNTER reslevel;
{
  ExprPtr domains,dompats,savepat,list;
  Boolean savemapsweep;
  ENTER;

  if (STKFULL(savemapsweep))
    {
      (void) compileSimpleOf1(f,env,data,failp,errlevel,errmaplevel,errp,reslevel);
      return;
    }
  domains = data->domains; /* list (with special kinds) of variables (as indexes on the
			      run-time stack) containing the different domains */
  dompats = data->dompats;
  if (domains==NIL)
    berror(message(621));
  else
    {
      list = CAR(domains);
      data->domains = CDR(domains);
      savepat = data->pat;
      savemapsweep = data->mapsweep;
      data->dompats = CDR(dompats);
      data->pat = CAR(CAR(dompats));
      data->mapsweep = (dompats->kind==(CHAR)E_PAT_REVRECORD_GNRL_CONS);
      compileSweep(f,env,
		   (INT)dompats->kind,IVAL(list),
		   data->combine,data->combine1,data->neutral,
		   compileOneOf,(CHAR *)data,failp,
		   errlevel,errmaplevel,errp,
		   reslevel,data->resclass,data->simplep);
      /* restore data */
      data->mapsweep = savemapsweep;
      data->pat = savepat;
      data->domains = domains;
      data->dompats = dompats;
    }
  QUIT;
}

PRIVATE void compileMapBodyPred(f,env,data,swis,failp,errlevel,errmaplevel,errp)
IN BCFILE *f;
IN CONST compEnv *env;
IN struct compilesimpleof *data;
IN CONST struct sweepinfo *swis;
INOUT ExprPtr failp;
IN COUNTER errlevel,errmaplevel;
INOUT ExprPtr errp;
{
  ExprPtr pats,pat,doms;
  struct sweepinfo *swip;
  COUNTER level;

  for (pats = data->dompats, doms = data->domains, swip = swis; doms!=NIL;
       pats = CDR(pats), doms = CDR(doms), swip++)
    {
      getSweep(f,pats->kind,swip);
      level = f->stackDepth-1;
      pat = CAR(CAR(pats));
      switch (pats->kind)
	{
	case E_PAT_RECORD_GNRL_CONS: /* in list domain descriptor */
	case E_PAT_RECORD_MAPADD: /* sub map domain descriptor */
	  (void) compilePattern(f,env,pat,failp,NIL,false);
	  break;
	case E_PAT_REVRECORD_GNRL_CONS: /* in map domain descriptor */
	  if (pat->kind==(CHAR)E_CONS)
	    {
	      if (nontrivMatch(CAR(pat)))
		{
		  compileFetchCar(f,level);
		  (void) compilePattern(f,env,CAR(pat),failp,NIL,false);
		}
	      if (nontrivMatch(CDR(pat)))
		{
		  if (level!=f->stackDepth-1)
		    compileFetchCdr(f,level);
		  else
		    bcPut(BC_CDR,f);
		  (void) compilePattern(f,env,CDR(data->pat),failp,NIL,false);
		}
	    }
	  else
	    {
	      if (nontrivMatch(CDR(pat)))
		{
		  compileFetchCdr(f,level);
		  (void) compilePattern(f,env,CDR(pat),failp,NIL,false);
		}
	      if (nontrivMatch(CAR(pat)))
		{
		  if (level!=f->stackDepth-1)
		    compileFetchCar(f,level);
		  else
		    bcPut(BC_CAR,f);
		  (void) compilePattern(f,env,CAR(data->pat),failp,NIL,false);
		}
	    }
	}
    }
  {
    Boolean saveflags;
#if DEBUGGING!=0
    INT dstart;
    bcdState st;
#endif

#if DEBUGGING!=0
    dstart = beginDebug(f,&st);
#endif
    saveflags = env->flags;
    if ((env->flags & CEF_NEEDENV) && !(needEnvBindings(data->locals,env)
					|| needEnv(data->pred,env)
					|| needEnv(data->body,env)))
      env->flags &= ~CEF_NEEDENV;
    if (env->flags & CEF_NEEDENV)
      {
	env->level--;
	compileCurEnv(f,env); /* --- <...> env (if needenv); --- <...> (if !needenv) */
	env->level++;
	bcPut(BC_CONSENV,f);
	bcPut((OPCODE)data->nlocals,f);
	newCurEnv(env,f->stackDepth-1);
	dumpLetStores(f,f->stackDepth-1,CE_LETDELAY(env));
      }
    /* --- <...> env (if needenv); --- <...> (if !needenv) */
    CE_LETDELAY(env) = NewMap();
    /* This is what we have called someplace else 'letconsenv trickery':
       we only allocate the environment after all patterns have been
       matched, and then we fill it in with the values we have gotten
       (and only if such an environment is needed). */
    flushDbgNames(f,env);

    compileBindings(f,env,data->locals,failp);
    compileTest(f,env,data->pred,failp);
    (*data->evalpromote)(f,env,data->body,errlevel,errmaplevel,errp);
    env->flags = saveflags;
#if DEBUGGING!=0
    endDebug(dstart,f,&st);
#endif
  }
}

PRIVATE void compileMapInc(f,env,data,swis,exitp,failp)
IN BCFILE *f;
IN CONST compEnv *env;
IN struct compilesimpleof *data;
IN CONST struct sweepinfo *swis;
INOUT ExprPtr exitp;
INOUT ExprPtr failp;
{
  ExprPtr pats,doms;
  struct sweepinfo *swip;

  pats = data->dompats;
  doms = data->domains;
  swip = swis;
  incSweep(f,pats->kind,swip,exitp); /* if first list is empty, goto exitp */
  /* otherwise, increment all other lists: */
  for (; pats = CDR(pats), doms = CDR(doms), swip++, doms!=NIL; )
    {
      incSweep(f,pats->kind,swip,failp);
      /* otherwise, if some non-first list is empty (while first list is not), then fail. */
    }
}

PRIVATE void compileSimpleMap(f,env,data,neutralp,errlevel,errmaplevel,errp,outp)
IN BCFILE *f;
IN CONST compEnv *env;
IN struct compilesimpleof *data;
INOUT ExprPtr neutralp;
IN COUNTER errlevel,errmaplevel;
INOUT ExprPtr errp;
INOUT ExprPtr outp;
{
  INT list;
  COUNTER level;
  INOUT ExprPtr failp; /* contrarily to compileSimpleOf(), failp is the merge node
			  we have to point to when the lists to be traversed are
			  not of the same length (leads to raising ParSweep). */

  level = f->stackDepth; /* --- */
  list = IVAL(CAR(data->domains));
  failp = beginForward(f);
  /* first, deal with the case where all domains are empty */
  {
    ExprPtr nonemptyp;

    nonemptyp = beginForward(f);
    switch (data->dompats->kind)
      {
      case E_PAT_RECORD_GNRL_CONS: /* in list domain descriptor */
	if (list!=f->stackDepth-1)
	  compileFetch(f,list);
	compileGoto(f,BC_ISNIL,nonemptyp,0);
empty: /* --- list1 */
	/* if first domain is empty, check that all the other domains are empty,
	   too; if this fails, raise ParSweep (i.e., goto failp) */
	{
	  ExprPtr doms,pats;

	  for (doms = data->domains, pats = data->dompats;
	       pats = CDR(pats), doms = CDR(doms), doms!=NIL; )
	    {
	      list = IVAL(CAR(doms));
	      switch (pats->kind)
	      {
	      case E_PAT_RECORD_GNRL_CONS: /* in list domain descriptor */
		if (list!=f->stackDepth-1)
		  compileFetch(f,list);
		compileGoto(f,BC_ISNIL,failp,0);
		break;
	      case E_PAT_REVRECORD_GNRL_CONS: /* in map domain descriptor */
		if (list!=f->stackDepth-1)
		  compileFetch(f,list);
		compileGoto(f,BC_ISEMPTY,failp,0);
		break;
	      case E_PAT_RECORD_MAPADD: /* sub map domain descriptor; never empty */
		break;
	      }
	    }
	}
	compileGoto(f,BC_GOTO,neutralp,0);
	setStackDepth(f,level); /* --- */
	endForward(f,nonemptyp);
	break;
      case E_PAT_REVRECORD_GNRL_CONS: /* in map domain descriptor */
	if (list!=f->stackDepth-1)
	  compileFetch(f,list);
	compileGoto(f,BC_ISEMPTY,nonemptyp,0);
	goto empty;
	break;
      case E_PAT_RECORD_MAPADD: /* sub map domain descriptor; never empty */
	break;
      }
  } /* --- */
  { /* Now, first domain is non-empty.  We still have
       to check that the others are non-empty, too.
       Otherwise, we should raise ParSweep (goto failp).
    */
    ExprPtr doms, pats;
    ExprPtr sequelp;

    for (doms = data->domains, pats = data->dompats;
	 pats = CDR(pats), doms = CDR(doms), doms!=NIL; )
      {
	list = IVAL(CAR(doms));
	switch (pats->kind)
	  {
	  case E_PAT_RECORD_GNRL_CONS: /* in list domain descriptor */
	    if (list!=f->stackDepth-1)
	      compileFetch(f,list);
	    compileGoto(f,BC_ISCONS,failp,0);
	    break;
	  case E_PAT_REVRECORD_GNRL_CONS: /* in map domain descriptor */
	    sequelp = beginForward (f);
	    if (list!=f->stackDepth-1)
	      compileFetch(f,list);
	    compileGoto(f,BC_ISEMPTY,sequelp,0);
	    compileGoto(f,BC_GOTO,failp,0);
	    endForward(f,sequelp);
	    break;
	  case E_PAT_RECORD_MAPADD: /* sub map domain descriptor; never empty */
	    break;
	  }
      }
    compileDrop(f,f->stackDepth-level);
  }
  {
    struct sweepinfo *swis;
    COUNTER looplevel;

    /* now, compute function on first element */
    {
      INT n = ListLength(data->domains);

      swis = (struct sweepinfo *)alloctmp(n*sizeof(struct sweepinfo));
    }
    /* first, initialize all local references to parts of the objects to sweep through */
    {
      ExprPtr doms,pats;
      struct sweepinfo *swip;

      for (doms = data->domains, pats = data->dompats, swip = swis; doms!=NIL;
	   doms = CDR(doms), pats = CDR(pats), swip++)
	{
#ifdef BUGGY
	  if (pats->kind==(CHAR)E_PAT_RECORD_GNRL_CONS) /* in list domain descriptor */
	    { /* then don't copy list: we can modify it where it is */
	      swip->listref = IVAL(CAR(doms));
	    }
	  else
	    /* Buggy, because not copying the list means that it will be
	       assigned to by the INCLISTSWEEP instructions, so that
	       if the list we are ranging over is a variable, then
	       references to this variable after the comprehension
	       will point to NIL instead.  This is not what we want. */
#endif
	    startSweep(f,pats->kind,IVAL(CAR(doms)),swip);
	}
    }
    looplevel = f->stackDepth; /* --- <loop-data1> ... <loop-datan> */
    if (data->simplep)
      {
	if (data->neutral==NULL)
	  ;
	else
	  (*data->neutral) (f, env); /* --- <loop-data1> ... <loop-datan> [res] */
      }
    else
      {/* second, evaluate predicate and function, if necessary, on data,
	  until some element verifying the predicate is found */
	ExprPtr startp,nonep;

	startp = beginBackward(f); /* start loop; --- <loop-data1> ... <loop-datan> */
	nonep = beginForward(f);
	compileMapBodyPred(f,env,data,swis,nonep,errlevel,errmaplevel,errp);
	compilePushDrop(f,looplevel,(data->resclass==0));
	/* --- <loop-data1> ... <loop-datan> [res] */
	if (isForward(nonep))
	  { /* If pattern fails to match, or predicate returns false, then loop
	       until pattern matches and predicate returns true. */
	    ExprPtr contp,exitp;
	    COUNTER savelevel;

	    savelevel = f->stackDepth;
	    contp = beginForward(f);
	    compileGoto(f,BC_GOTO,contp,0);
	    setStackDepth(f,looplevel);
	    endForward(f,nonep); /* --- <loop-data1> ... <loop-datan> */

	    exitp = beginForward(f);
	    if (f->mapStackDepth==errmaplevel)
	      {
		compileMapInc(f,env,data,swis,exitp,failp);
		compileGoto(f,BC_GOTO,startp,0);
	      }
	    else
	      {
		ExprPtr fail2p;

		fail2p = beginForward(f);
		compileMapInc(f,env,data,swis,exitp,fail2p);
		compileGoto(f,BC_GOTO,startp,0);
		if (isForward(fail2p))
		  {
		    endForward(f,fail2p);
		    compileMapDrop(f,errmaplevel);
		    compileGoto(f,BC_GOTO,failp,0);
		  }
	      }

	    setStackDepth(f,savelevel);
	    endForward(f,exitp);
	    compileMapDrop(f,errmaplevel);

	    { /* check that iteration has ended at the same time on all domains: */
	      ExprPtr pats,doms;
	      struct sweepinfo *swip;
	      ExprPtr emptyp;

	      for (doms = data->domains, pats = data->dompats, swip = swis;
		   pats = CDR(pats), doms = CDR(doms), swip++, doms!=NIL; )
		{
		  if (CDR(doms)==NIL)
		    emptyp = contp;
		  else
		    emptyp = beginForward(f);
		  incSweep(f,pats->kind,swip,emptyp);
		  compileGoto(f,BC_GOTO,failp,0);
		  endForward(f,emptyp);
		}
	    }
	    /* --- <loop-data1> ... <loop-datan> [res] */
	  }
	(*data->combine1)(f,env); /* --- <loop-data1> ... <loop-datan> [res] */
      }
    { /* now, compile the real while loop */
      COUNTER innerlevel;
      ExprPtr startp,exitp,fail2p,enterp = NOEXPR;

      innerlevel = f->stackDepth;
      if (data->simplep)
	{
	  Boolean olddead;

	  enterp = beginForward(f);
	  olddead = f->dead;
	  compileGoto(f,BC_GOTO,enterp,0);
	  f->dead = olddead;
	}
      startp = beginBackward(f); /* start loop; --- <loop-data1> ... <loop-datan> [old-res] */
      exitp = beginForward(f);
      if (f->mapStackDepth==errmaplevel)
	fail2p = failp;
      else fail2p = beginForward(f);
      compileMapInc(f,env,data,swis,exitp,fail2p);
      if (data->simplep)
	endForward(f,enterp);

      compileMapBodyPred(f,env,data,swis,startp,errlevel,errmaplevel,errp);
      /* --- <loop-data1> ... <loop-datan> [old-res] <...> [res] */
      (*data->combine)(f,env,innerlevel-1);
      /* --- <loop-data1> ... <loop-datan> [old-res] <...> [new-res] */
      compilePushDrop(f,looplevel,(data->resclass==0));
      /* --- <loop-data1> ... <loop-datan> [new-res] */
      compileGoto(f,BC_GOTO,startp,0);

      if (f->mapStackDepth!=errmaplevel && isForward(fail2p))
	{
	  endForward(f,fail2p);
	  compileMapDrop(f,errmaplevel);
	  compileGoto(f,BC_GOTO,failp,0);
	}

      setStackDepth(f,innerlevel);
      endForward(f,exitp); /* --- <loop-data1> ... <loop-datan> [res] */
      {
	ExprPtr pats,doms;
	struct sweepinfo *swip;
	ExprPtr emptyp;

	for (doms = data->domains, pats = data->dompats, swip = swis;
	     pats = CDR(pats), doms = CDR(doms), swip++, doms!=NIL; )
	  {
	    emptyp = beginForward(f);
	    incSweep(f,pats->kind,swip,emptyp);
	    if (CDR(doms)!=NIL)
	      compileGoto(f,BC_GOTO,failp,0);
	    else
	      {
		COUNTER savelevel;

		savelevel = f->stackDepth;
		compileDrop(f,f->stackDepth-errlevel);
		switch (data->resclass)
		  {
		  case 0: /* return expression */
		    endForward(f,failp);
		    compileRaiseImm(f,env,I_MAPEXC);
		    compileGoto(f,BC_GOTO,outp,0);
		    break;
		  case 1: /* existential test */
		    endForward(f,failp);
		    compileRaiseImm(f,env,I_MAPEXC);
		    compileGoto(f,BC_IF,outp,1);
		    compileGoto(f,BC_GOTO,errp,0);
		    break;
		  case 2: /* universal test */
		    endForward(f,failp);
		    compileRaiseImm(f,env,I_MAPEXC);
		    compileGoto(f,BC_IF,errp,1);
		    compileGoto(f,BC_GOTO,outp,0);
		    break;
		  default: /* case 3: statement */
		    endForward(f,failp);
		    compileRaiseImm(f,env,I_MAPEXC);
		    compileGoto(f,BC_GOTO,outp,0);
		    break;
		  }
		setStackDepth(f,savelevel);
	      }
	    endForward(f,emptyp);
	  }
      }
    }
  }
  compileMapDrop(f,errmaplevel);
  f->mapStackDepth = errmaplevel;
  /* And get the final result: */
  compilePushDrop(f,level,(data->resclass==0));
  /* --- [res] */
}

PRIVATE ExprPtr compileDomainsSimple1 P((IN BCFILE *f,IN CONST compEnv *env,
					 IN CONST ExprPtr domains,INOUT ExprPtr *locals));

DEF_PRIVATE_ACTION_1ARG(compiledomainssimplek)
{
  struct compiledomainssimplek *data;

  data = (struct compiledomainssimplek *)ENV_UNDERLYING(env);
  return compileDomainsSimple1(data->f,data->env,data->domains,&data->locals);
}
ACTION_END

PRIVATE ExprPtr compileDomainsSimple1k(f,env,domains,locals)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr domains;
INOUT ExprPtr *locals;
{
  struct compiledomainssimplek *data = (struct compiledomainssimplek *)
    alloctmp(sizeof(struct compiledomainssimplek));
  ExprPtr res;

  data->f = f;
  data->env = env;
  data->domains = domains;
  data->locals = *locals;
  res = CallCC(ACTION_NAME(compiledomainssimplek),(ExprPtr)data,true);
  *locals = data->locals;
  return res;
}

PRIVATE INT getExprIndex(f,env,e)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr e;
{
  ExprPtr i_l,index;
  COUNTER level;
  Boolean found;
  ENTER;

  e = forgetExtra(e);
  if (e->kind==(CHAR)E_VARREF)
    {
      i_l = KINT(VARRANK(e));
      level = env->level - VARLEVEL(e);
      if (level!=0)
	i_l = KONS(i_l,KINT(level));
      index = MapGet(i_l,CE_VARS(env),&found);
      QUIT;
      if (found)
	return IVAL(index);
    }
  compileExpr(f,env,e);
  return f->stackDepth-1;
}

PRIVATE ExprPtr compileDomainsSimple1(f,env,domains,locals)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr domains;
INOUT ExprPtr *locals;
/* return list (with tags E_PAT*RECORD*) of levels (INTs) in the run-time stack;
   these levels are the places where we shall find the evaluated domains on
   the run-time stack. */
{
  ExprPtr rest,res;
  ExprPtr level;

  if (STKFULL(res))
    return compileDomainsSimple1k(f,env,domains,locals);
  if (domains==NIL)
    {
      *locals = NIL;
      return NIL;
    }
  switch (domains->kind)
    {
    case E_PAT_RECORD_GNRL_CONS:
    case E_PAT_REVRECORD_GNRL_CONS:
    case E_PAT_RECORD_MAPADD:
      /* in map, in list, sub map domain descriptors */
      level = KINT(getExprIndex(f,env,CDR(CAR(domains))));
      rest = compileDomainsSimple1(f,env,CDR(domains),locals);
      MAKEPAIR(domains->kind,level,rest,res);
      break;
    case E_TUPLE_1: /* local declarations */
      *locals = REFVAL(domains);
      return NIL;
    }
  return KEEP(res);
}

PRIVATE ExprPtr compileDomainsSimple(f,env,domains,locals)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr domains;
OUT ExprPtr *locals;
{
  ExprPtr doms;
  ENTER;

  doms = compileDomainsSimple1(f,env,domains,locals);
  KEEPONLY(doms); /* no need to keep *locals, because it is in domains */
  return doms;
}

PRIVATE void fillData(data,ik)
INOUT struct compilesimpleof *data;
IN IterKind ik;
{
  switch (ik)
    {
    case C_PROG:
      data->neutral = compileNeutralProg;
      data->combine = compileCombineProg;
      data->combine1 = compileCombine1Prog;
      data->evalpromote = compileEvalPromoteProg;
      data->finalize = compileFinalizeProg;
      data->resclass = 3; /* statement, does not return a result */
      data->simplep = true;
      break;
    case C_LIST:
      data->neutral = compileNeutralList;
      data->combine = compileCombineList;
      data->combine1 = compileCombine1List;
      data->evalpromote = compileEvalPromoteList;
      data->finalize = compileFinalizeList;
      data->resclass = 0; /* expression, returns a result as an ExprPtr */
      data->simplep = true;
      break;
    case C_UNSHARED_LIST:
      data->neutral = compileNeutralUnsharedList;
      data->combine = compileCombineUnsharedList;
      data->combine1 = compileCombine1UnsharedList;
      data->evalpromote = compileEvalPromoteUnsharedList;
      data->finalize = compileFinalizeUnsharedList;
      data->resclass = 0; /* expression, returns a result as an ExprPtr */
      data->simplep = true;
      break;
    case C_SET:
      data->neutral = compileNeutralSet;
      data->combine = compileCombineSet;
      data->combine1 = compileCombine1Set;
      data->evalpromote = compileEvalPromoteSet;
      data->finalize = compileFinalizeSet;
      data->resclass = 0; /* expression, returns a result as an ExprPtr */
      data->simplep = true;
      break;
    case C_APPEND:
      data->neutral = compileNeutralAppend;
      data->combine = compileCombineAppend;
      data->combine1 = compileCombine1Append;
      data->evalpromote = compileEvalPromoteAppend;
      data->finalize = compileFinalizeAppend;
      data->resclass = 0; /* expression, returns a result as an ExprPtr */
      data->simplep = true;
      break;
    case C_UNSHARED_APPEND:
      data->neutral = compileNeutralUnsharedAppend;
      data->combine = compileCombineUnsharedAppend;
      data->combine1 = compileCombine1UnsharedAppend;
      data->evalpromote = compileEvalPromoteUnsharedAppend;
      data->finalize = compileFinalizeUnsharedAppend;
      data->resclass = 0; /* expression, returns a result as an ExprPtr */
      data->simplep = true;
      break;
    case C_UNION:
      data->neutral = compileNeutralUnion;
      data->combine = compileCombineUnion;
      data->combine1 = compileCombine1Union;
      data->evalpromote = compileEvalPromoteUnion;
      data->finalize = compileFinalizeUnion;
      data->resclass = 0; /* expression, returns a result as an ExprPtr */
      data->simplep = true;
      break;
    case C_UNION_UNDER:
      data->neutral = compileNeutralUnionUnder;
      data->combine = compileCombineUnionUnder;
      data->combine1 = compileCombine1UnionUnder;
      data->evalpromote = compileEvalPromoteUnionUnder;
      data->finalize = compileFinalizeUnionUnder;
      data->resclass = 0; /* expression, returns a result as an ExprPtr */
      data->simplep = true;
      break;
    case C_INTER:
      data->neutral = compileNeutralInter;
      data->combine = compileCombineInter;
      data->combine1 = compileCombine1Inter;
      data->evalpromote = compileEvalPromoteInter;
      data->finalize = compileFinalizeInter;
      data->resclass = 0; /* expression, returns a result as an ExprPtr */
      data->simplep = false; /* the methods above require simplep==false */
      break;
    case C_MAP:
      data->neutral = compileNeutralMap;
      data->combine = compileCombineMap;
      data->combine1 = compileCombine1Map;
      data->evalpromote = compileEvalPromoteMap;
      data->finalize = compileFinalizeMap;
      data->resclass = 0; /* expression, returns a result as an ExprPtr */
      data->simplep = true;
      break;
    case C_MAP_UNDER:
      data->neutral = compileNeutralMapUnder;
      data->combine = compileCombineMapUnder;
      data->combine1 = compileCombine1MapUnder;
      data->evalpromote = compileEvalPromoteMapUnder;
      data->finalize = compileFinalizeMapUnder;
      data->resclass = 0; /* expression, returns a result as an ExprPtr */
      data->simplep = true;
      break;
    case C_ALL:
      data->neutral = compileNeutralAll;
      data->combine = compileCombineAll;
      data->combine1 = compileCombine1All;
      data->evalpromote = compileEvalPromoteAll;
      data->finalize = compileFinalizeAll;
      data->resclass = 2; /* universal test expression, branches wherever appropriate */
      data->simplep = true;
      break;
    case C_EXISTS:
      data->neutral = compileNeutralExists;
      data->combine = compileCombineExists;
      data->combine1 = compileCombine1Exists;
      data->evalpromote = compileEvalPromoteExists;
      data->finalize = compileFinalizeExists;
      data->resclass = 1; /* existential test expression, branches wherever appropriate */
      data->simplep = true;
      break;
    case C_SOME:
      data->neutral = compileNeutralSome;
      data->combine = compileCombineSome;
      data->combine1 = compileCombine1Some;
      data->evalpromote = compileEvalPromoteSome;
      data->finalize = compileFinalizeSome;
      data->resclass = 0; /* expression, returns a result as an ExprPtr */
      data->simplep = false; /* the methods above require simplep==false */
      break;
    case C_UNSHARED_SOME:
      data->neutral = compileNeutralUnsharedSome;
      data->combine = compileCombineUnsharedSome;
      data->combine1 = compileCombine1UnsharedSome;
      data->evalpromote = compileEvalPromoteUnsharedSome;
      data->finalize = compileFinalizeUnsharedSome;
      data->resclass = 0; /* expression, returns a result as an ExprPtr */
      data->simplep = false; /* the methods above require simplep==false */
      break;
    }
}

PRIVATE void compileFinalOf(f,env,data,errlevel,errmaplevel,neutralp,finp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST struct compilesimpleof *data;
IN COUNTER errlevel,errmaplevel;
INOUT ExprPtr neutralp;
INOUT ExprPtr finp;
{
  (*data->finalize)(f,env);
  if (isForward(neutralp))
    {
      if (data->neutral==NULL) /* case where neutral does nothing (quantifier is then true) */
	{
	  compileDrop(f,f->stackDepth-errlevel);
	  endForward(f,neutralp);
	  if (finp!=NIL)
	    endForward(f,finp);
	}
      else
	{
	  compilePushDrop(f,errlevel,(data->resclass==0));
	  if (finp==NIL)
	    finp = beginForward(f);
	  compileGoto(f,BC_GOTO,finp,0);
	  setStackDepth(f,errlevel);
	  endForward(f,neutralp);
	  (*data->neutral)(f,env);
	  endForward(f,finp);
	}
    }
}

PRIVATE void compileOf1(f,env,body,pred,domains,ik,errlevel,errmaplevel,errp,nlocals)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr body,pred,domains;
IN IterKind ik;
IN COUNTER errlevel,errmaplevel;
INOUT ExprPtr errp;
IN INT nlocals;
{
  ExprPtr neutralp;
  struct compilesimpleof *data =
    (struct compilesimpleof *)alloctmp(sizeof(struct compilesimpleof));
  ENTER;

  env->level--;
  data->nlocals = nlocals;
  data->dompats = domains;
  data->domains = compileDomainsSimple(f,env,data->dompats,&data->locals);
#ifdef OBSOLETE
  if (env->flags & CEF_NEEDENV)
    {
      compileCurEnv(f,env); /* --- oldenv (if needenv); --- (if !needenv) */
      env->level = env->ilevel;
      bcPut(BC_CONSENV,f);
      bcPut((OPCODE)nlocals,f);
      newCurEnv(env,f->stackDepth-1);
    }
  /* --- env (if needenv); --- (if !needenv) */
#endif

  env->level = env->ilevel;
  data->body = body;
  data->pred = pred;
  fillData(data,ik);
  neutralp = beginForward(f);
  compileSimpleOf(f,env,data,neutralp,errlevel,errmaplevel,errp,errlevel);
  /* Notice that we call compileSimpleOf() with failp=neutralp: this way,
     if no pattern matches, it will jump directly to neutralp. */
  compilePushDrop(f,errlevel,(data->resclass==0));
  compileFinalOf(f,env,data,errlevel,errmaplevel,neutralp,NIL);
  QUIT;
}

PRIVATE void compileMap1(f,env,body,pred,domains,ik,errlevel,errmaplevel,errp,nlocals)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr body,pred,domains;
IN IterKind ik;
IN COUNTER errlevel,errmaplevel;
INOUT ExprPtr errp;
IN INT nlocals;
{
  ExprPtr neutralp,outp;
  struct compilesimpleof *data =
    (struct compilesimpleof *)alloctmp(sizeof(struct compilesimpleof));
  ENTER;

  env->level--;
  data->nlocals = nlocals;
  data->dompats = domains;
  data->domains = compileDomainsSimple(f,env,data->dompats,&data->locals);
#ifdef OBSOLETE
  if (env->flags & CEF_NEEDENV)
    {
      compileCurEnv(f,env); /* --- oldenv (if needenv); --- (if !needenv) */
      env->level = env->ilevel;
      bcPut(BC_CONSENV,f);
      bcPut((OPCODE)nlocals,f);
      newCurEnv(env,f->stackDepth-1);
    }
  /* --- env (if needenv); --- (if !needenv) */
#endif

  env->level = env->ilevel;
  data->body = body;
  data->pred = pred;
  fillData(data,ik);
  neutralp = beginForward(f);
  outp = beginForward(f);
  compileSimpleMap(f,env,data,neutralp,errlevel,errmaplevel,errp,outp);
  compileFinalOf(f,env,data,errlevel,errmaplevel,neutralp,outp);
  QUIT;
}

PRIVATE void compileOf(f,env,body,pred,domains,ik,errlevel,errmaplevel,errp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr body,pred;
IN CONST ExprPtr domains;
IN IterKind ik;
IN COUNTER errlevel,errmaplevel;
INOUT ExprPtr errp;
{
  compEnv *newenv = (compEnv *)alloctmp(sizeof(compEnv));
  INT nlocals;

  *newenv = *env;
  newenv->_comp.next = theHeapFrames;
  CE_LETDELAY(newenv) = NewMap();
  resetDbgNames(newenv);
  theHeapFrames = (gcStackView *)&newenv->_comp;

  nlocals = IVAL(CAR(domains));
  newenv->ilevel = ++newenv->level;
  if ((newenv->flags & CEF_NEEDENV) && !(needEnvDoms(CDR(domains),env)
					 || needEnv(pred,newenv)
					 || needEnv(body,newenv)))
    newenv->flags &= ~CEF_NEEDENV;
  /* Note that needEnv() is applied to newenv, but needEnvDoms() is applied
     to env (the old one), which has a level lower by 1. */

  compileOf1(f,newenv,body,pred,CDR(domains),ik,errlevel,errmaplevel,errp,nlocals);

  theHeapFrames = newenv->_comp.next;
}

PRIVATE void compileMap(f,env,body,pred,domains,ik,errlevel,errmaplevel,errp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr body,pred;
IN CONST ExprPtr domains;
IN IterKind ik;
IN COUNTER errlevel,errmaplevel;
INOUT ExprPtr errp;
{
  compEnv *newenv = (compEnv *)alloctmp(sizeof(compEnv));
  INT nlocals;

  *newenv = *env;
  newenv->_comp.next = theHeapFrames;
  CE_LETDELAY(newenv) = NewMap();
  resetDbgNames (newenv);
  theHeapFrames = (gcStackView *)&newenv->_comp;

  nlocals = IVAL(CAR(domains));
  newenv->ilevel = ++newenv->level;
  if ((newenv->flags & CEF_NEEDENV) && !(needEnvDoms(CDR(domains),env)
					 || needEnv(pred,newenv)
					 || needEnv(body,newenv)))
    newenv->flags &= ~CEF_NEEDENV;
  /* Note that needEnv() is applied to newenv, but needEnvDoms() is applied
     to env (the old one), which has a level lower by 1. */

  compileMap1(f,newenv,body,pred,CDR(domains),ik,errlevel,errmaplevel,errp,nlocals);

  theHeapFrames = newenv->_comp.next;
}

PRIVATE void compileOfNoFuss(f,env,e,ik)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr e;
IN IterKind ik;
{
  ExprPtr errp;

  errp = beginForward(f);
  compileOf(f,env,CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),ik,f->stackDepth,f->mapStackDepth,errp);
  /* isForward(errp) should not hold here. */
}

PRIVATE void compileMapNoFuss(f,env,e,ik)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr e;
IN IterKind ik;
{
  ExprPtr errp;

  errp = beginForward(f);
  compileMap(f,env,CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),ik,f->stackDepth,f->mapStackDepth,errp);
  /* isForward(errp) should not hold here. */
}

PRIVATE void compileSomeOf(f,env,e,ik)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr e;
IN IterKind ik;
{
  ExprPtr foundp;

  foundp = beginForward(f);
  compileOf(f,env,CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),ik,f->stackDepth,f->mapStackDepth,foundp);
  if (isForward(foundp))
    {
      endForward(f,foundp);
    }
}

PRIVATE void compileSomeMap(f,env,e,ik)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr e;
IN IterKind ik;
{
  ExprPtr foundp;

  foundp = beginForward(f);
  compileMap(f,env,CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),ik,f->stackDepth,f->mapStackDepth,foundp);
  if (isForward(foundp))
    {
      endForward(f,foundp);
    }
}

PRIVATE void compileWhile(f,env,cond,suchthat,body,ik,errlevel,errmaplevel,errp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr cond,suchthat,body;
IN IterKind ik;
IN COUNTER errlevel,errmaplevel;
INOUT ExprPtr errp;
{
  struct compilesimpleof *data
    = (struct compilesimpleof *)alloctmp(sizeof(struct compilesimpleof));
  ExprPtr neutralp,outp;
  COUNTER level;

  level = f->stackDepth;
  fillData(data,ik);
  data->nlocals = 0;
  neutralp = beginForward(f);
  {
    ExprPtr startp;

    startp = beginBackward(f); /* --- */
    compileTest(f,env,cond,neutralp); /* if test is false, return neutral */
    compileTest(f,env,suchthat,startp); /* if suchthat is false, then loop back */
    /* otherwise, compute first value of body */
    (*data->evalpromote)(f,env,body,errlevel,errmaplevel,errp); /* --- <...> [res] */
    (*data->combine1)(f,env); /* --- <...> [res] */
    compilePushDrop(f,level,(data->resclass==0)); /* --- [res] */
  }
  /* --- [first-res] */
  outp = beginForward(f);
  {
    /* now, loop and combine */
    ExprPtr startp;

    startp = beginBackward(f); /* --- [old-res] */
    compileTest(f,env,cond,outp); /* if cond is false, then exit of the loop */
    compileTest(f,env,suchthat,startp); /* if suchthat is false, then loop back */
    (*data->evalpromote)(f,env,body,errlevel,errmaplevel,errp);
    /* --- [old-res] <...> [new-res] */
    (*data->combine)(f,env,level); /* --- [old-res] <...> [res] */
    if (data->resclass==0)
      compileSwDrop(f,f->stackDepth-level); /* --- res */
    /* we don't use compilePushDrop(), as an optimization */
    /* --- [res] */
    compileGoto(f,BC_GOTO,startp,0);
  }
  if (isForward(outp))
    {
      setStackDepth(f,level+(data->resclass==0));
      endForward(f,outp);
    }
  compileFinalOf(f,env,data,errlevel,errmaplevel,neutralp,NIL);
}

PRIVATE void compileWhileNoFuss(f,env,e,ik)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr e;
IN IterKind ik;
{
  ExprPtr errp;

  errp = beginForward(f);
  compileWhile(f,env,CAR(WCOND(e)),CDR(WCOND(e)),WBODY(e),ik,
	       f->stackDepth,f->mapStackDepth,errp);
  /* isForward(errp) should not hold here. */
}

PRIVATE void compileSomeWhile(f,env,e,ik)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr e;
IN IterKind ik;
{
  ExprPtr foundp;

  foundp = beginForward(f);
  compileWhile(f,env,CAR(WCOND(e)),CDR(WCOND(e)),WBODY(e),ik,
	       f->stackDepth,f->mapStackDepth,foundp);
  if (isForward(foundp))
    {
      endForward(f,foundp);
    }
}

PRIVATE void compileHandle(f,env,body,handlers,compile,push,wrongp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr body,handlers;
IN CONST compileFunc compile;
IN Boolean push;
INOUT ExprPtr wrongp;
{
  ExprPtr handlep,outp,wp;
  COUNTER level;

  level = f->stackDepth;
  outp = beginForward(f);
  handlep = beginForward(f);
  wp = beginForward(f);
  bcPush(1,f);
  compileGoto(f,BC_HANDLE,handlep,0);
  bcPush(-1,f);

  {
    compEnv *newenv = (compEnv *)alloctmp(sizeof(compEnv));

    *newenv = *env;
    CE_LETDELAY(newenv) = NewMap();
    resetDbgNames (newenv);
    newenv->_comp.next = theHeapFrames;
    theHeapFrames = (gcStackView *)&newenv->_comp;

    {
      ExprPtr aux;
      ENTER;

      aux = MakeCounted(E_COUNTED, handlep, (INT)level);
      (void) KEEP (aux);
      aux = MakeCounted(E_COUNTED, aux, (INT)f->mapStackDepth);
      (void) KEEP (aux);
      CE_HANDLERS(newenv) = MakePair(E_CONS,aux,CE_HANDLERS(newenv));
      QUIT;
    }

    (*compile)(f,newenv,body,wp);

    theHeapFrames = newenv->_comp.next;
  }

  compileDebugZone (true, f);
  compileCodePointOp(BC_NEXTHANDLE,f);
  compileGoto(f,BC_GOTO,outp,0);

  if (isForward(wp))
    {
      endForward(f,wp);
      compileCodePointOp(BC_NEXTHANDLE,f);
      compileGoto(f,BC_GOTO,wrongp,0);
    }

  setStackDepth(f,level+1); /* --- exn */
  endForward(f,handlep); /* raise has just been triggered */
  compileCodePointOp(BC_NEXTHANDLE,f);
  {
    ExprPtr globerrp;

    globerrp = beginForward(f);
    (void) compilePolyFun0(f,env,handlers,1,compile,(push?1:0),wrongp,globerrp,outp,NIL);
    compilePushDrop(f,level,push);
    if (isForward(globerrp))
      {
	compileGoto(f,BC_GOTO,outp,0);
	setStackDepth(f,level+1);
	endForward(f,globerrp);
	compileRaise(f,env);
	compilePushDrop(f,level,push);
      }
  }

  setStackDepth(f,level+(push?1:0));
  endForward(f,outp);
}

DEF_PRIVATE_ACTION_1ARG(compiletest)
{
  struct compiletest *data;

  data = (struct compiletest *)ENV_UNDERLYING(env);
  compileTest(data->f,data->env,data->e,data->wrongp);
  return NIL;
}
ACTION_END

PRIVATE ExprPtr compileTest1(f,env,e,wrongp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr e;
INOUT ExprPtr wrongp;
{
  struct compiletest *data = (struct compiletest *)alloctmp(sizeof(struct compiletest));

  data->f = f;
  data->env = env;
  data->e = e;
  data->wrongp = wrongp;
  return CallCC(ACTION_NAME(compiletest),(ExprPtr)data,true);
}

PRIVATE void compileTest(f,env,test,wrongp)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr test;
INOUT ExprPtr wrongp;
{
  if (STKFULL(f))
    {
      (void) compileTest1(f,env,test,wrongp);
      return;
    }
again:
  switch (test->kind)
    {
    case E_AND:
      compileTest(f,env,PROGDO(test),wrongp);
      test = PROGTHEN(test);
      goto again;
    case E_OR:
      {
	ExprPtr secondchancep,finp;
	COUNTER level;

	level = f->stackDepth;
	secondchancep = beginForward(f);
	compileTest(f,env,PROGDO(test),secondchancep);
	finp = beginForward(f);
	compileGoto(f,BC_GOTO,finp,0);
	setStackDepth(f,level);
	endForward(f,secondchancep);
	compileTest(f,env,PROGTHEN(test),wrongp);
	setStackDepth(f,level);
	endForward(f,finp);
	break;
      }
    case E_NOT:
      {
	ExprPtr rightp;
	COUNTER level;

	level = f->stackDepth;
	rightp = beginForward(f);
	compileTest(f,env,REFVAL(test),rightp);
	compileGoto(f,BC_GOTO,wrongp,0);
	setStackDepth(f,level);
	if (isForward(rightp))
	  {
	    endForward(f,rightp);
	  }
	break;
      }
    case E_EQUAL:
      compileExpr(f,env,CAR(test));
      compileExpr(f,env,CDR(test));
      compileGoto(f,BC_ISEQUAL,wrongp,2);
      break;
    case E_NEQUAL:
      {
	ExprPtr rightp;
	COUNTER level;

	level = f->stackDepth;
	rightp = beginForward(f);
	compileExpr(f,env,CAR(test));
	compileExpr(f,env,CDR(test));
	compileGoto(f,BC_ISEQUAL,rightp,2);
	compileGoto(f,BC_GOTO,wrongp,0);
	setStackDepth(f,level);
	endForward(f,rightp);
	break;
      }
    case E_ALL_OF:
      compileOf(f,env,CAR(OFFP(test)),CDR(OFFP(test)),OFDOM(test),C_ALL,
		f->stackDepth,f->mapStackDepth,wrongp);
      break;
    case E_EXISTS_OF:
      {
	ExprPtr rightp;
	COUNTER level;

	level = f->stackDepth;
	rightp = beginForward(f);
	compileOf(f,env,CAR(OFFP(test)),CDR(OFFP(test)),OFDOM(test),C_EXISTS,
		  f->stackDepth,f->mapStackDepth,rightp);
	compileGoto(f,BC_GOTO,wrongp,0);
	setStackDepth(f,level);
	endForward(f,rightp);
	break;
      }
    case E_ALL_MAP:
      compileMap(f,env,CAR(OFFP(test)),CDR(OFFP(test)),OFDOM(test),C_ALL,
		 f->stackDepth,f->mapStackDepth,wrongp);
      break;
    case E_EXISTS_MAP:
      {
	ExprPtr rightp;
	COUNTER level;

	level = f->stackDepth;
	rightp = beginForward(f);
	compileMap(f,env,CAR(OFFP(test)),CDR(OFFP(test)),OFDOM(test),C_EXISTS,
		   level,f->mapStackDepth,rightp);
	compileGoto(f,BC_GOTO,wrongp,0);
	setStackDepth(f,level);
	endForward(f,rightp);
	break;
      }
    case E_ALL_WHILE:
      compileWhile(f,env,CAR(WCOND(test)),CDR(WCOND(test)),WBODY(test),C_ALL,
		   f->stackDepth,f->mapStackDepth,wrongp);
      break;
    case E_EXISTS_WHILE:
      {
	ExprPtr rightp;
	COUNTER level;

	level = f->stackDepth;
	rightp = beginForward(f);
	compileWhile(f,env,CAR(WCOND(test)),CDR(WCOND(test)),WBODY(test),C_EXISTS,
		     level,f->mapStackDepth,rightp);
	compileGoto(f,BC_GOTO,wrongp,0);
	setStackDepth(f,level);
	endForward(f,rightp);
	break;
      }
    case E_HANDLE:
      compileHandle(f,env,HBODY(test),HHANDLERS(test),compileTest,false,wrongp);
      break;
    case E_APPL: case E_GOTO:
      /* case E_N_APPL:*/
      compileApply(f,env,APPLFUN(test),APPLARG(test),
		   (test->kind==(CHAR)E_GOTO),
		   compileTest,false,wrongp);
      break;
    case E_DEBUGZONE:
#if DEBUGGING!=0
      if (debugEnabled)
	{
	  ExprPtr savedz = currentDebugZone;

	  currentDebugZone = DZINFO(test);
	  compileTest(f,env,DZCODE(test),wrongp);
	  currentDebugZone = savedz;
	  break;
	}
#endif
      test = DZCODE(test);
      goto again;
    case E_PROFILE:
      {
	ExprPtr saveprof;

	saveprof = CE_PROF(env);
	CE_PROF(env) = CDR(test);
	test = CAR(test);
	compileTest (f,env,test,wrongp);
	CE_PROF(env) = saveprof;
      }
      break;
      /*compilexpr:*/
    default:
      if (test->kind==(CHAR)E_QUOTE && REFVAL(test)==trueValue)
	;
      else if (test->kind==(CHAR)E_QUOTE && REFVAL(test)==falseValue)
	compileGoto(f,BC_GOTO,wrongp,0);
      else
	{
	  compileExpr(f,env,test);
	  compileGoto(f,BC_IF,wrongp,1);
	}
      break;
    }
}

PRIVATE void compileTestExpr(f,env,test)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr test;
{
  ExprPtr wrongp;

  if (test->kind==(CHAR)E_QUOTE && (REFVAL(test)==trueValue ||
				    REFVAL(test)==falseValue))
    compileConstant(f,test);
  else
    {
      wrongp = beginForward(f);
      compileTest(f,env,test,wrongp);
      testToBool(f,wrongp);
    }
}

PRIVATE void compileIf(f,env,test,right,wrong)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr test,right,wrong;
{
  ExprPtr wrongp,finp;
  COUNTER level;

  if (test->kind==(CHAR)E_QUOTE && REFVAL(test)==trueValue)
    compileExpr(f,env,right);
  else if (test->kind==(CHAR)E_QUOTE && REFVAL(test)==falseValue)
    compileExpr(f,env,wrong);
  else
    {
      level = f->stackDepth;
      wrongp = beginForward(f);
      compileTest(f,env,test,wrongp);
      finp = beginForward(f);
      /* if test returns true: */
      setStackDepth(f,level);
      compileExpr(f,env,right);
      compileGoto(f,BC_GOTO,finp,0);
      /* otherwise: */
      setStackDepth(f,level);
      endForward(f,wrongp);
      compileExpr(f,env,wrong);
      setStackDepth(f,level+1);
      endForward(f,finp);
    }
}

DEF_PRIVATE_ACTION_1ARG(compilestmt)
{
  struct compilestmt *data;

  data = (struct compilestmt *)ENV_UNDERLYING(env);
  compileStmt(data->f,data->env,data->e);
  return NIL;
}
ACTION_END

PRIVATE void compileStmt1(f,env,e)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr e;
{
  struct compilestmt *data = (struct compilestmt *)alloctmp(sizeof(struct compilestmt));

  data->f = f;
  data->env = env;
  data->e = e;
  (void) CallCC(ACTION_NAME(compilestmt),(ExprPtr)data,true);
}

PRIVATE void compileStmt(f,env,e)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr e;
{
  if (STKFULL(e))
    {
      compileStmt1(f,env,e);
      return;
    }
again:
  switch (e->kind)
    {
    case E_NIL: case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
    case E_EMPTY: case E_TUPLE_0: case E_RECORD_0:
      break;
    case E_PROG:
      compileStmt(f,env,PROGDO(e));
      e = PROGTHEN(e);
      goto again;
    case E_SET:
      compileExpr(f,env,SETREF(e));
      compileExpr(f,env,SETVAL(e));
      bcPut(BC_SETREF,f);
      bcPush(-2,f);
      break;
    case E_PROG_OF:
      compileOfNoFuss(f,env,e,C_PROG);
      break;
    case E_PROG_MAP:
      compileMapNoFuss(f,env,e,C_PROG);
      break;
    case E_PROG_WHILE:
      compileWhileNoFuss(f,env,e,C_PROG);
      break;
    case E_HANDLE:
      compileHandle(f,env,HBODY(e),HHANDLERS(e),compileStmtFun,false,NULL);
      break;
    case E_APPL: case E_GOTO:
      /*case E_N_APPL:*/
      compileApply(f,env,APPLFUN(e),APPLARG(e),
		   (e->kind==(CHAR)E_GOTO),
		   compileStmtFun,false,NULL);
      break;
    case E_DEBUGZONE:
#if DEBUGGING!=0
      if (debugEnabled)
	{
	  ExprPtr savedz = currentDebugZone;

	  currentDebugZone = DZINFO(e);
	  compileStmt(f,env,DZCODE(e));
	  currentDebugZone = savedz;
	  break;
	}
#endif
      e = DZCODE(e);
      goto again;
    case E_PROFILE:
      {
	ExprPtr saveprof;

	saveprof = CE_PROF(env);
	CE_PROF(env) = CDR(e);
	e = CAR(e);
	compileStmt (f,env,e);
	CE_PROF(env) = saveprof;
      }
    break;
    /*compilexpr:*/
    default:
      compileExpr(f,env,e);
      compileDrop(f,1);
      break;
    }
}

PRIVATE void compileStmtExpr(f,env,e)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr e;
{
  compileStmt(f,env,e);
  compileConstant(f,tuple0);
}

DEF_PRIVATE_ACTION_1ARG(compileExpr)
{
  struct compilestmt *data;

  data = (struct compilestmt *)ENV_UNDERLYING(env);
  compileExpr(data->f,data->env,data->e);
  return NIL;
}
ACTION_END

PRIVATE ExprPtr compileExpr1(f,env,e)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr e;
{
  struct compilestmt *data = (struct compilestmt *)alloctmp(sizeof(struct compilestmt));

  data->f = f;
  data->env = env;
  data->e = e;
  return CallCC(ACTION_NAME(compileExpr),(ExprPtr)data,true);
}

PRIVATE void compileExpr(f,env,e)
IN BCFILE *f;
IN CONST compEnv *env;
IN CONST ExprPtr e;
{
  ENTER;

  if (STKFULL(f))
    {
      (void) compileExpr1(f,env,e);
      QUIT;
      return;
    }
again:
  (void) alloctmp(0L);
  switch (e->kind)
    {
    case E_NIL: case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
    case E_EMPTY: case E_TUPLE_0: case E_RECORD_0:
      compileConstant(f,e);
      break;
    case E_DATACON:
      compileDataCon(f,env,COUNTEDN(e),COUNTEDVAL(e));
      break;
    case E_REF:
      compileUnary(f,env,BC_REF,REFVAL(e));
      break;
    case E_TUPLE_1:
      compileUnary(f,env,BC_KTUPLE1,REFVAL(e));
      break;
    case E_UNSHARED_TUPLE_1:
      compileUnary(f,env,BC_MAKETUPLE1,REFVAL(e));
      break;
    case E_RECORD_1:
      compileRecord1(f,env,BC_KRECORD1,e);
      break;
    case E_UNSHARED_RECORD_1:
      compileRecord1(f,env,BC_MAKERECORD1,e);
      break;
    case E_TUPLE_2:
      compileBinary(f,env,BC_KTUPLE2,CAR(e),CDR(e));
      break;
    case E_UNSHARED_TUPLE_2:
      compileBinary(f,env,BC_MAKETUPLE2,CAR(e),CDR(e));
      break;
    case E_CONS:
      compileBinary(f,env,BC_KONS,CAR(e),CDR(e));
      break;
    case E_UNSHARED_CONS:
      compileBinary(f,env,BC_CONS,CAR(e),CDR(e));
      break;
    case E_TUPLE_GNRL:
      compileTupleGnrl(f,env,BC_KTUPLEGNRL,e);
      break;
    case E_UNSHARED_TUPLE_GNRL:
      compileTupleGnrl(f,env,BC_MAKETUPLEGNRL,e);
      break;
    case E_RECORD_GNRL:
      compileRecordGnrl(f,env,BC_KRECORDGNRL,e);
      break;
    case E_UNSHARED_RECORD_GNRL:
      compileRecordGnrl(f,env,BC_MAKERECORDGNRL,e);
      break;
    case E_MAPLET:
      compileBinary(f,env,BC_MAPLET,MAPLETX(e),MAPLETY(e));
      break;
    case E_OVERWRITE:
      compileBinary(f,env,BC_OVERWRITE,CAR(e),CDR(e));
      break;
    case E_UNDERWRITE:
      compileBinary(f,env,BC_UNDERWRITE,CAR(e),CDR(e));
      break;
    case E_QUOTE:
      compileConstant(f,REFVAL(e));
      break;
    case E_VARREF:
      compileVarref(f,env,VARLEVEL(e),VARRANK(e));
      break;
    case E_LET:
      compileLet(f,env,LETBIND(e),LETBODY(e));
      break;
    case E_PAT_AS:
      e = CAR(e);
      goto again;
    case E_POLYFUN:
      /*case E_N_LAMBDA:*/
      compilePolyFun(f,env,e,false);
      break;
    case E_MEMO_FN:
      compileMemoFn(f,env,MEMOFUN(e),MEMOREM(e));
      break;
    case E_APPL: case E_GOTO:
      /*case E_N_APPL:*/
      compileApply(f,env,APPLFUN(e),APPLARG(e),
		   (e->kind==(CHAR)E_GOTO),
		   compileExprFun,true,NULL);
      break;
    case E_GET:
    case E_TSEL1:
      compileUnary(f,env,BC_REFVAL,GETREF(e));
      break;
    case E_MAPGET:
      compileMapGet(f,env,CAR(e),CDR(e));
      break;
    case E_TSELECT:
      compileExpr(f,env,CDR(e));
      compileTupleSelect(f,CAR(e));
      break;
    case E_TSEL2:
      compileExpr(f,env,CDR(e));
      bcPut((CAR(e)==ione)?BC_CAR:BC_CDR,f);
      break;
    case E_TSELGNRL:
      compileExpr(f,env,CDR(e));
      compileNth(f,IVAL(CAR(e))-1);
      break;
    case E_RSELECT:
      compileExpr(f,env,CDR(e));
      compileRecordSelect(f,env,CAR(e));
      break;
    case E_RSEL1:
      compileExpr(f,env,CDR(e));
      bcPut(BC_CDR,f);
      break;
    case E_RSELGNRL:
      compileExpr(f,env,CDR(e));
      compileSelectGnrl(f,env,CAR(e));
      break;
    case E_RMU1:
      compileMu1(f,env,CAR(e),CAR(CDR(e)),CDR(CDR(e)),true);
      break;
    case E_UNSHARED_RMU1:
      compileMu1(f,env,CAR(e),CAR(CDR(e)),CDR(CDR(e)),false);
      break;
    case E_RMUX:
      compileMuGnrl(f,env,CAR(e),CDR(e),true);
      break;
    case E_UNSHARED_RMUX:
      compileMuGnrl(f,env,CAR(e),CDR(e),false);
      break;
    case E_IF:
      compileIf(f,env,TESTIF(e),CAR(TESTFORK(e)),CDR(TESTFORK(e)));
      break;
    case E_PROG:
      compileStmt(f,env,PROGDO(e));
      e = PROGTHEN(e);
      goto again;
    case E_SET: case E_PROG_OF: case E_PROG_MAP: case E_PROG_WHILE:
      compileStmtExpr(f,env,e);
      break;
    case E_AND: case E_OR: case E_NOT: case E_EQUAL: case E_NEQUAL:
    case E_ALL_OF: case E_EXISTS_OF:
    case E_ALL_MAP: case E_EXISTS_MAP:
    case E_ALL_WHILE: case E_EXISTS_WHILE:
      compileTestExpr(f,env,e);
      break;
    case E_LIST_OF:
      compileOfNoFuss(f,env,e,C_LIST);
      break;
    case E_UNSHARED_LIST_OF:
      compileOfNoFuss(f,env,e,C_UNSHARED_LIST);
      break;
    case E_SET_OF:
      compileOfNoFuss(f,env,e,C_SET);
      break;
    case E_APPEND_OF:
      compileOfNoFuss(f,env,e,C_APPEND);
      break;
    case E_UNSHARED_APPEND_OF:
      compileOfNoFuss(f,env,e,C_UNSHARED_APPEND);
      break;
    case E_UNION_OF:
      compileOfNoFuss(f,env,e,C_UNION);
      break;
    case E_UNION_UNDER_OF:
      compileOfNoFuss(f,env,e,C_UNION_UNDER);
      break;
    case E_INTER_OF:
      compileOfNoFuss(f,env,e,C_INTER);
      break;
    case E_MAP_OF:
      compileOfNoFuss(f,env,e,C_MAP);
      break;
    case E_MAP_UNDER_OF:
      compileOfNoFuss(f,env,e,C_MAP_UNDER);
      break;
    case E_SOME_OF:
      compileSomeOf(f,env,e,C_SOME);
      break;
    case E_UNSHARED_SOME_OF:
      compileSomeOf(f,env,e,C_UNSHARED_SOME);
      break;
    case E_LIST_MAP:
      compileMapNoFuss(f,env,e,C_LIST);
      break;
    case E_UNSHARED_LIST_MAP:
      compileMapNoFuss(f,env,e,C_UNSHARED_LIST);
      break;
    case E_SET_MAP:
      compileMapNoFuss(f,env,e,C_SET);
      break;
    case E_APPEND_MAP:
      compileMapNoFuss(f,env,e,C_APPEND);
      break;
    case E_UNSHARED_APPEND_MAP:
      compileMapNoFuss(f,env,e,C_UNSHARED_APPEND);
      break;
    case E_UNION_MAP:
      compileMapNoFuss(f,env,e,C_UNION);
      break;
    case E_UNION_UNDER_MAP:
      compileMapNoFuss(f,env,e,C_UNION_UNDER);
      break;
    case E_INTER_MAP:
      compileMapNoFuss(f,env,e,C_INTER);
      break;
    case E_MAP_MAP:
      compileMapNoFuss(f,env,e,C_MAP);
      break;
    case E_MAP_UNDER_MAP:
      compileMapNoFuss(f,env,e,C_MAP_UNDER);
      break;
    case E_SOME_MAP:
      compileSomeMap(f,env,e,C_SOME);
      break;
    case E_UNSHARED_SOME_MAP:
      compileSomeMap(f,env,e,C_UNSHARED_SOME);
      break;
    case E_LIST_WHILE:
      compileWhileNoFuss(f,env,e,C_LIST);
      break;
    case E_UNSHARED_LIST_WHILE:
      compileWhileNoFuss(f,env,e,C_UNSHARED_LIST);
      break;
    case E_SET_WHILE:
      compileWhileNoFuss(f,env,e,C_SET);
      break;
    case E_APPEND_WHILE:
      compileWhileNoFuss(f,env,e,C_APPEND);
      break;
    case E_UNSHARED_APPEND_WHILE:
      compileWhileNoFuss(f,env,e,C_UNSHARED_APPEND);
      break;
    case E_UNION_WHILE:
      compileWhileNoFuss(f,env,e,C_UNION);
      break;
    case E_UNION_UNDER_WHILE:
      compileWhileNoFuss(f,env,e,C_UNION_UNDER);
      break;
    case E_INTER_WHILE:
      compileWhileNoFuss(f,env,e,C_INTER);
      break;
    case E_MAP_WHILE:
      compileWhileNoFuss(f,env,e,C_MAP);
      break;
    case E_MAP_UNDER_WHILE:
      compileWhileNoFuss(f,env,e,C_MAP_UNDER);
      break;
    case E_SOME_WHILE:
      compileSomeWhile(f,env,e,C_SOME);
      break;
    case E_UNSHARED_SOME_WHILE:
      compileSomeWhile(f,env,e,C_UNSHARED_SOME);
      break;
    case E_HANDLE:
      compileHandle(f,env,HBODY(e),HHANDLERS(e),compileExprFun,true,NULL);
      break;
    case E_RAISE:
      compileExpr(f,env,RBODY(e));
      compileRaise(f,env);
      break;
    case E_EXCEPTION:
      if (e->mark & MARK_SHARED)
	compileConstant(f,e); /* extern exceptions are not generative */
      else
	compileUnary(f,env,(COUNTEDN(e) & EXC_BENIGN)?BC_EXCEPTION_BENIGN:BC_EXCEPTION,EXCVAL(e));
      break;
    case E_DELAY:
      compileUnary(f,env,BC_DELAY,DCODE(e));
      break;
    case E_PACK:
      compileExpr(f,env,PACKVAL(e));
      compileConstant(f,PACKTYPE(e));
      bcPut(BC_PACK,f);
      bcPush(-1,f);
      break;
    case E_DEBUGZONE:
#if DEBUGGING!=0
      if (debugEnabled)
	{
	  ExprPtr savedz = currentDebugZone;

	  currentDebugZone = DZINFO(e);
	  compileExpr(f,env,DZCODE(e));
	  currentDebugZone = savedz;
	  break;
	}
#endif
      e = DZCODE(e);
      goto again;
    case E_PROFILE:
      {
	ExprPtr saveprof;

	saveprof = CE_PROF(env);
	CE_PROF(env) = CDR(e);
	e = CAR(e);
	compileExpr (f,env,e);
	CE_PROF(env) = saveprof;
      }
    break;
    default:
      kinderror(message(623),e->kind);
      /*NOTREACHED*/
      break;
    }
  QUIT;
}

PUBLIC ExprPtr Compile(e,level,bindp)
IN CONST ExprPtr e;
IN COUNTER level;
IN Boolean bindp;
{
  BCFILE *f;
  ExprPtr errp;
  compEnv *env = (compEnv *)alloctmp(sizeof(compEnv));

#if DEBUGGING!=0
  currentDebugZone = NULL;
#endif
  env->level = env->ilevel = level;
  env->globenvs = 0x1L; /* only level 0 holds GLOBENV */
  env->flags = CEF_NEEDENV | CEF_SPARESTACK;
  env->_comp.n = CE_N;
  env->_comp.next = theHeapFrames;
  CE_VARS(env) = NewMap();
  CE_INLINES(env) = NewMap();
  CE_HANDLERS(env) = NIL;
  CE_LETDELAY(env) = NewMap();
  CE_DZ(env) = NIL;
  CE_PROF(env) = defaultProf;
  CE_FV(env) = NewMap();
  CE_DBGNAMES(env) = NewMap();
  CE_PARENT_DBGNAMES(env) = NewMap();
  defineMe(env,NIL);
  theHeapFrames = (gcStackView *)&env->_comp;

  f = bcOpenD(env);

  newCurEnv(env,0); /* environment will be at position 0 of the stack, initially */
  errp = beginForward(f);
  if (bindp)
    {
      compileBindings(f,env,e,errp);
      compileConstant(f,falseValue); /* everything ok */
      if (isForward(errp))
	{
	  ExprPtr finp;

	  finp = beginForward(f);
	  compileGoto(f,BC_GOTO,finp,0);
	  setStackDepth(f,f->stackDepth-1);
	  endForward(f,errp);
	  compileConstant(f,trueValue);
	  endForward(f,finp);
	}
    }
  else
    compileExpr(f,env,e);
  compileDebugZone(false,f);
  compileCodePointOp(BC_RETURN,f);
  bcPut((OPCODE)(f->stackDepth-1),f);

  theHeapFrames = env->_comp.next;
  return bcCloseD (f);
}
#endif
